//
//The Lua for Mario's Powerup Abilities
//

local CAPESPINSTATE = S_MARIO_CAPESPIN //cape.lua

//Mega Mario 1up Bar HUD
local OFFSETMAX = 28
local OFFSETFACTOR = 1

local MarioHUD = {[0] = {}, [1] = {}}

MarioHUD[0].yoffset = 0
MarioHUD[1].yoffset = 0

MarioHUD[0].combo = 0
MarioHUD[1].combo = 0

hud.add(function(v, player, cam)
	if G_IsSpecialStage() or MarioScoreGametype() return end
	local mo = player.mo
	local id = player == secondarydisplayplayer and 1 or 0
	local mrhud = MarioHUD[id]
	local alive = (mo and mo.valid and mo.health and not player.playerstate and player.mariomega)
	
	if player.spectator or not IsMario(player) or not alive
		mrhud.yoffset = max($ - 1, 0)
	elseif mrhud.yoffset < OFFSETMAX
		mrhud.yoffset = min($ + 1, OFFSETMAX)
	end
	
	local realcombo = player.mariomegacombo or 0
	local combo = mrhud.combo
	local totalcombo = combo/21
	
	if mrhud.yoffset
		if combo != realcombo
			if combo < realcombo
				mrhud.combo = $ + 1
			else
				mrhud.combo = $ - 1
			end
		end
		
		local scale = FRACUNIT
		local x = 160 * FRACUNIT
		local y = mrhud.yoffset*OFFSETFACTOR * FRACUNIT 
		local flags = V_HUDTRANS|V_SNAPTOTOP|V_PERPLAYER
		
		v.drawScaled(x, y, scale, v.cachePatch("MRMEGABAR"), flags)
		
		for i = 1, 105
			local color = SKINCOLOR_WHITE
			local colormap = TC_DEFAULT
			if totalcombo > (i - 1)/21
				color = SKINCOLOR_EMERALD
				if i % 21 != 11
					colormap = TC_RAINBOW
				end
			elseif i % 21 == 11
				color = SKINCOLOR_RED
			elseif combo >= i
				color = SKINCOLOR_ORANGE
			end
			
			v.drawScaled(x, y, scale, v.cachePatch("MRPIECE"), flags,
			v.getColormap(colormap, color))
			
			if not (i % 21)
				x = $ + 3*FRACUNIT
			else
				x = $ + FRACUNIT
			end
		end
	else
		mrhud.combo = 0
	end
end)

states[freeslot("S_MARIO_FIREEXPLODE")] = {freeslot("SPR_MRPT"), FF_FULLBRIGHT|FF_ANIMATE|X, 6, nil, 5, 1, S_NULL}
states[freeslot("S_MARIO_SUPERBALL")] = {SPR_MRPT, FF_ANIMATE|B, 6, nil, 1, 3, S_MARIO_SUPERBALL}
mobjinfo[freeslot("MT_MARIO_FIREBALL")] = {
	spawnstate = S_SMALLFIREBAR1,
	spawnhealth = 1000,
	reactiontime = 8,
	//deathsound = sfx_mario1,
	deathstate = S_MARIO_FIREEXPLODE,
	speed = 24*FRACUNIT,
	radius = 24*FRACUNIT,
	height = 48*FRACUNIT,
	mass = DMG_FIRE,
	damage = 1,
	flags = MF_MISSILE|MF_NOCLIP|MF_GRENADEBOUNCE|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_PUSHABLE,
}

local MAXGRAVITY = 32
local GRAVITYFACTOR = MAXGRAVITY/3
rawset(_G, "MARIO_FIREDELAY", TICRATE*4/5)

rawset(_G, "MARIO_GOLDFIRECOLOR", SKINCOLOR_SUPERGOLD4)
rawset(_G, "MARIO_ICEFIRECOLOR", SKINCOLOR_ICY)
rawset(_G, "MARIO_BALLFIRECOLOR", SKINCOLOR_MOSS)
rawset(_G, "MARIO_SILVERFIRECOLOR", SKINCOLOR_DUSK)
rawset(_G, "MARIO_GREENFIRECOLOR", SKINCOLOR_MINT)

//Set up the fireball type
local function MarioFireballSpawn(mobj, after)
	mobj.hit_sound = sfx_mario2
	if mobj.info.raisestate == 1 //Gold
		mobj.color = MARIO_GOLDFIRECOLOR
		mobj.colorized = true
		mobj.name = "gold fireball"
	elseif mobj.info.raisestate == 2 //Ice
		mobj.color = MARIO_ICEFIRECOLOR
		mobj.colorized = true
		mobj.hit_sound = sfx_mrfrze
		mobj.name = "iceball"
	elseif mobj.info.raisestate == 3 //Super ball
		mobj.state = S_MARIO_SUPERBALL
		mobj.color = MARIO_BALLFIRECOLOR
		mobj.name = "super ball"
	elseif mobj.info.raisestate == 4 //Silver
		mobj.color = MARIO_SILVERFIRECOLOR
		mobj.colorized = true
		mobj.name = "silver fireball"
	elseif mobj.info.raisestate == 5 //Green
		mobj.color = MARIO_GREENFIRECOLOR
		mobj.colorized = true
		mobj.name = "green fireball"
	else
		mobj.name = "fireball"
	end
	
	mobj.cantouchteam = false
	mobj.blockable = 1
	mobj.block_stun = 6
	mobj.block_sound = sfx_mario1
	mobj.block_hthrust = 7
	mobj.block_vthrust = 5
	mobj.spawnfire = true
	
	mobj.info.raisestate = 0
end
addHook("MobjSpawn", MarioFireballSpawn, MT_MARIO_FIREBALL)

//Fireball Thinker
local function MarioFireballThinker(mobj)
	local mobjgravity = -gravity * P_MobjFlip(mobj)
	if mobj.target and mobj.target.valid then mobjgravity = FixedMul($, mobj.target.scale) end
	if mobj.color == MARIO_BALLFIRECOLOR then mobjgravity = $/2 end
	local dumbgravity = mobjgravity * MAXGRAVITY
	
	//Fake MobjSpawn hook for after the object has been spawned
	if not mobj.threshold
		if mobj.target and mobj.target.valid
			P_TryMove(mobj, mobj.x + mobj.target.momx,
			mobj.y + mobj.target.momy, true)
			if P_MobjFlip(mobj.target) == -1 and P_MobjFlip(mobj) != -1
				mobj.flags2 = $ | MF2_OBJECTFLIP
				mobjgravity = -$
				dumbgravity = -$
			end
		end
		
		mobj.scalespeed = mobj.scale/6
		P_SetScale(mobj, mobj.scale/2)

		mobj.momz = mobjgravity * MAXGRAVITY
		mobj.z = $ + mobj.height/2*P_MobjFlip(mobj)
		
		//Fun little hack for water splashing
		mobj.flags = $ &~ MF_PUSHABLE
		if not mobj.fuse then mobj.fuse = TICRATE*8 end
		
		if mobj.color != MARIO_BALLFIRECOLOR
			mobj.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
			mobj.blendmode = AST_ADD
		end
		
		//Different properties
		if mobj.color == MARIO_GOLDFIRECOLOR or mobj.color == MARIO_SILVERFIRECOLOR //Gold
			mobj.destscale = $*3/2
		elseif mobj.color == MARIO_BALLFIRECOLOR //Super ball
			mobj.flags = $ | MF_BOUNCE
			mobj.destscale = $*3/2
		end
		
		mobj.flags = $ &~ MF_NOCLIP
		mobj.threshold = 1
	end
		
	//You're dead! Don't go any further
	if not mobj.health or not (mobj.flags & MF_MISSILE)
		mobj.momx = 0
		mobj.momy = 0
		mobj.momz = 0
		
		mobj.destscale = mobj.scale
		mobj.spritexscale = FRACUNIT*3/2
		mobj.spriteyscale = FRACUNIT*3/2
		
		//First time stuff
		if mobj.threshold != 69 then mobj.threshold = 69 else return end
		
		//Gold explosion
		if mobj.color == MARIO_GOLDFIRECOLOR or mobj.color == MARIO_SILVERFIRECOLOR
			P_RadiusAttack(mobj, mobj.target, mobj.info.radius*6, mobj.info.mass, false)
			
			mobj.state = S_CACOFIRE_EXPLODE1
			
			for i = 1, 4
				local xpl = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_THOK)
				
				xpl.color = mobj.color
				xpl.colorized = mobj.colorized
				
				xpl.blendmode = mobj.blendmode
				xpl.renderflags = mobj.renderflags
				
				xpl.state = S_CACOFIRE_EXPLODE1
				xpl.tics = $ * (i + 1)
				
				xpl.fuse = 0
				
				xpl.scalespeed = i * mobj.scale/INT8_MAX
				xpl.destscale = mobj.scale * 8
			end
			
			local xpl = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_THOK)
			S_StartSound(xpl, sfx_mrglxp)
			xpl.state = S_INVISIBLE
			xpl.tics = TICRATE
			
		elseif mobj.color == MARIO_ICEFIRECOLOR
			S_StartSound(mobj, sfx_mricbr)
		elseif mobj.color == MARIO_BALLFIRECOLOR
			S_StartSound(mobj, sfx_mario1)
			mobj.colorized = true
			mobj.spritexscale = FRACUNIT
			mobj.spriteyscale = FRACUNIT
			mobj.state = S_SPINDUST1
		else
			S_StartSound(mobj, sfx_mrfire)
		end
		
		return
	end
	
	//Horizontal Movement
	mobj.friction = FRACUNIT
	if mobj.target and mobj.target.valid
		local angle = mobj.angle
		if mobj.color == MARIO_BALLFIRECOLOR and mobj.momx and mobj.momy
			angle = R_PointToAngle2(0,0, mobj.momx, mobj.momy)
			mobj.angle = angle - ANGLE_180
		end
		
		P_InstaThrust(mobj, angle, FixedMul(mobj.info.speed, mobj.target.scale))
	end
	
	//Floor bounce
	if P_IsObjectOnGround(mobj)
	or (mobj.eflags & MFE_JUSTHITFLOOR and mobj.eflags & MFE_ONGROUND)
		mobj.momz = -dumbgravity
		
	//Gravity
	elseif mobj.color != MARIO_BALLFIRECOLOR
		dumbgravity = abs($)
		if mobj.greenflip then mobjgravity = -$ end
		
		mobj.momz = min(max($ + mobjgravity * GRAVITYFACTOR, -dumbgravity), dumbgravity)
		if ((mobj.target and mobj.target.skin == "luigi")
		or mobj.color == MARIO_SILVERFIRECOLOR or mobj.color == MARIO_GREENFIRECOLOR)
		and abs(mobj.momz) >= abs(dumbgravity)
			mobj.greenflip = not $
		end
		
	//Super ball hit ceiling or floor
	elseif mobj.z + mobj.height >= mobj.ceilingz
		mobj.momz = dumbgravity
	elseif mobj.z <= mobj.floorz
		mobj.momz = -dumbgravity
	end
	
	//Ghost trail
	if mobj.threshold <= 1
		mobj.threshold = 2
	else
		local ghost = P_SpawnGhostMobj(mobj)
		ghost.blendmode = mobj.blendmode
		ghost.scalespeed = ghost.scale/12
		ghost.renderflags = mobj.renderflags
		ghost.destscale = 1
	end
end
addHook("MobjThinker", MarioFireballThinker, MT_MARIO_FIREBALL)

//No monitors!
addHook("ShouldDamage", function(monitor, mobj)
	if monitor and monitor.valid and monitor.flags & MF_MONITOR
	and mobj and mobj.valid and mobj.type == MT_MARIO_FIREBALL
		return false
	end
end)

//Collect coins
addHook("MobjMoveCollide", function(mo, mobj)
	if mo.color == MARIO_BALLFIRECOLOR and mo.flags & MF_MISSILE
	and mo.target and mo.target.valid and mo.target.player
	and mobj.health and mobj.flags & (MF_SPECIAL|MF_SPRING)
	and not (mobj.mariomonitorcoin and mobj.fuse > MARIO_DESPAWNTIME - TICRATE)
	and (mobj.type == MT_COIN or mobj.type == MT_FLINGCOIN
	or mobj.type == MT_RING or mobj.type == MT_FLINGRING)
	and mobj.z - mobj.scale <= mo.z + mo.height
	and mobj.z + mobj.height + mobj.scale >= mo.z
		if mobj.info.deathsound == sfx_rnglft
		or mobj.info.deathsound == sfx_rngrgh
			mobj.info.deathsound = sfx_itemup
		end
		if mo.target.player == displayplayer or mo.target.player == secondarydisplayplayer
			S_StartSound(mo.target, mobj.info.deathsound)
		else
			S_StartSound(mobj, mobj.info.deathsound)
		end
		A_ForceStop(mobj)
		P_KillMobj(mobj)
		P_GivePlayerRings(mo.target.player, 1)
	end
end, MT_MARIO_FIREBALL)

//Break Walls
local firewallbreak = function(mobj, sectorthing)
	local nerds = false
	for fof in sectorthing.ffloors()
		if fof.valid and (fof.flags & FF_BUSTUP) and (fof.flags & FF_EXISTS)	
			if mobj.z + mobj.height < fof.bottomheight continue end -- Are we too low?
			if mobj.z > fof.topheight continue end -- Are we too high? 
				
			EV_CrumbleChain(nil, fof)
			if fof.master.flags & ML_EFFECT5
				P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, mobj, fof.target)
			end
			nerds = true
		end
	end
	return nerds
end
addHook("MobjLineCollide", function(mobj, line)
	local killone
	local killtwo
	if not (line and line.valid and mobj and mobj.valid) then return end
	if line.frontsector and line.frontsector.valid then killone = firewallbreak(mobj, line.frontsector) end
	if line.backsector and line.backsector.valid then killtwo = firewallbreak(mobj, line.backsector) end
	if killone or killtwo then return true end
end, MT_MARIO_FIREBALL)

//Replace SRB2's original fireballs
mobjinfo[MT_FIREBALL].flags = $ | MF_NOCLIP
addHook("MobjSpawn", function(mobj)
	mobj.type = MT_MARIO_FIREBALL
	mobj.state = mobj.info.spawnstate
	mobj.flags = mobj.info.flags
	mobj.scale = $
	
	mobj.info.raisestate = 0
	mobj.threshold = 0
	MarioFireballSpawn(mobj)
end, MT_FIREBALL)

//Mini Mario bouncing off enemies
//Look, I know all the other mini code is in mario.lua, but you're stupid
addHook("ShouldDamage", function(mobj, mo, src, dmg, dmgtype)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if not player or player.mariosize >= 0 or player.mariogroundpound
	or player.powers[pw_super] or (player.powers[pw_invulnerability]
	and not (mobj and mobj.valid and mobj.flags & MF_MONITOR))
	or mo.state == S_PLAY_MELEE or mo.state == S_PLAY_MELEE_FINISH
	or mo.state == S_PLAY_MELEE_LANDING or player.marioshoe
	or not (mobj and mobj.valid) or mobj.player or dmgtype
		return
	end
	
	if mobj.flags & MF_SOLID
		return false
	end
	
	local bottomheight = mo.z
	local topheight = mo.z + mo.height

	if (mo.eflags & MFE_VERTICALFLIP)
		local swap = bottomheight
		bottomheight = topheight
		topheight = swap
	end
	
	//From above
	if not P_IsObjectOnGround(mo) and not (player.pflags & PF_SPINNING)
	and P_MobjFlip(mo)*(bottomheight - mobj.z + mobj.height/2) > 0
		local jumpfactor = player.jumpfactor
		player.jumpfactor = $*5/4
		if player.mariospinjump
			player.jumpfactor = $*2/3
		end
		player.pflags = $ &~ PF_JUMPED &~ PF_NOJUMPDAMAGE
		P_DoJump(player, false)
		player.jumpfactor = jumpfactor
		
		if player.panim == PA_FALL
		and not player.mariolongjump and not player.mariocrouch
			mo.state = S_PLAY_JUMP
		end
	else
		local bumper = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_BUMPER)
		bumper.reactiontime = 0
		P_DoSpring(bumper, mo)
		P_RemoveMobj(bumper)
		
		mo.momx = $*2/3 + mobj.momx
		mo.momy = $*2/3 + mobj.momy
		mo.momz = $*2/3 + mobj.momz
	end
	
	local part = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_POPPARTICLEMAR)
	P_TeleportMove(part, mobj.x/2 + mo.x/2, mobj.y/2 + mo.y/2, mobj.z/2 + mo.z/2)
	part.destscale = 1
	part.scalespeed = FixedMul($, part.scale)
	
	player.powers[pw_flashing] = max($, 2)
	
	S_StartSound(mo, sfx_mrmin5)
	mo.eflags = $ &~ MFE_SPRUNG
	return false
end)

//The abilities themselves
rawset(_G, "MarioPowerupAbility", function(player, fire)
	local mo = player.mo
	
	if not fire and player.marioforcehammer then return end //Force Hammer
	
	//Fireballs
	if MARIOFIREPOWERS[player.powers[pw_shield]] or player.powers[pw_super]
		if player.mariofiredelay >= 0 and mo.state != S_PLAY_FIRE
			if player.mariofiredelay
				player.mariofiredelay = min(-$, MARIO_FIREDELAY/-2)
			else
				player.mariofiredelay = MARIO_FIREDELAY
			end
			
			mo.state = S_PLAY_STND
			mo.state = S_PLAY_FIRE
			
			local angle = mo.angle
			if player.pflags & PF_ANALOGMODE
				local enemy = P_LookForEnemies(player, false, true)
				if enemy and enemy.valid and not (enemy.flags & MF_MONITOR)
					angle = R_PointToAngle2(mo.x, mo.y, enemy.x, enemy.y)
				end
			end
			
			if player.powers[pw_shield] == SH_GOLDFLOWER
			or player.powers[pw_super]
				mobjinfo[MT_MARIO_FIREBALL].raisestate = mo.skin == "luigi" and 4 or 1
				S_StartSound(mo, sfx_mrglfr)
			elseif player.powers[pw_shield] == SH_ICEFLOWER
				mobjinfo[MT_MARIO_FIREBALL].raisestate = 2
				S_StartSound(mo, sfx_mriceb)
			elseif player.powers[pw_shield] == SH_SUPERBALL
				mobjinfo[MT_MARIO_FIREBALL].raisestate = 3
				S_StartSound(mo, sfx_mrbal5)
				
				player.mariofiredelay = -abs($)
			else
				mobjinfo[MT_MARIO_FIREBALL].raisestate = mo.skin == "luigi" and 5 or 0
				S_StartSound(mo, sfx_mario7)
			end
			
			player.drawangle = angle
			P_SPMAngle(mo, MT_MARIO_FIREBALL, angle, 0,0)
		end
		
		if not player.powers[pw_super]
			return true
		end
	end
	
	if player.powers[pw_shield] == SH_CAPE or player.powers[pw_super]
		mo.state = S_PLAY_STND
		mo.state = CAPESPINSTATE
		
		player.mariojumpdir = player.cmd.angleturn<<16 - ANGLE_22h - ANG60 * ((leveltime + #player) % 6)
		
		S_StartSound(mo, sfx_mrspin)
		player.panim = PA_DASH
		
		return true
	end
end)