//
//The Lua for Mario's Cape
//

local CAPEFALL = 4
local CAPESPINTIME = TICRATE*3/5
local CAPESPINSTATE = freeslot("S_MARIO_CAPESPIN")
states[S_MARIO_CAPESPIN] = {SPR_PLAY, SPR2_WALK, CAPESPINTIME, nil, 0,  0, S_PLAY_WALK}

local CAPESPRITE2 = {
	//Rolling
	[SPR2_ROLL] = {-18, -8},
	
	//Ground Pound
	[SPR2_TIRE] = {-18, -4},
	
	//Backflip
	[SPR2_TWIN] = {-8, -8},
	
	//Hammer
	[SPR2_MLEE] = {-8, -4},
	[SPR2_MLEL] = {-8, -4},
	
	//Stuff
	[SPR2_SPNG] = {-6, 0},
	[SPR2_SKID] = {-8, 0},
	
	//Dive/Swimming (Both used in former)
	[SPR2_SWIM] = {-12, 0},
	[SPR2_SPIN] = {-15, 7},
	
	//Crouching/Crawling
	[SPR2_LAND] = {-8, 0},
	[SPR2_CLMB] = {-11, 4},
	
	//Long jump
	[SPR2_TAL0] = {-10, 0},
	[SPR2_STUN] = {-12, 0},
	
	//Pain
	[SPR2_PAIN] = {-8, -4},
	
	//Edge
	[SPR2_EDGE] = {-4, 7},
	
	//Ride/Wall cling
	[SPR2_RIDE] = {-4, 2},
	[SPR2_TAL2] = {8, 0},
	
	//Shoe
	[SPR2_TAL1] = {8, 0},
	
	//Cape stuff (important)
	[SPR2_TAL7] = {-3, 0},
	[SPR2_GLID] = {0, 8},
}

local LCAPESPRITE2 = {
	//Main
	[SPR2_STND] = {10, 0},
	[SPR2_WALK] = {10, 0},
	[SPR2_BNCE] = {4, 0},
	
	//Running
	[SPR2_RUN] = {5, 0},
	[SPR2_TAL7] = {0, 0},
	[SPR2_SKID] = {-4, 0},
	
	//Jump
	[SPR2_JUMP] = {3, 0},
	[SPR2_FLT_] = {3, 0},
	[SPR2_TAL0] = {-8, 0},
	
	//Spring
	[SPR2_SPNG] = {2, 0},
	[SPR2_FALL] = {4, 0},
	
	//Hammer
	[SPR2_MLEE] = {-6, 0},
	[SPR2_MLEL] = {-6, 0},
	
	//Dive/Swimming (Both used in former)
	[SPR2_SWIM] = {-2, 0},
	[SPR2_SPIN] = {-20, 7},
	
	//Edge
	[SPR2_EDGE] = {4, 7},
	
	//Shoe
	[SPR2_TAL1] = {12, 0},
}

freeslot("SPR_MRCP")
mobjinfo[freeslot("MT_MARIO_CAPE")] = {
	spawnstate = S_INVISIBLE,
	deathstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = mobjinfo[MT_PLAYER].radius,
	height = mobjinfo[MT_PLAYER].height,
	dispoffset = 1,
	flags = MF_SCENERY|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
}

//Cape collision in BattleMod
addHook("ShouldDamage", function(mo,cape,owner)
	//Relegate this hook to interactions with players and cape projectiles
	if not(CBW_Battle) then return end
	if not(mo and mo.valid) then return end
	if not(mo.player) then return end
	if not(cape and cape.valid and cape.type == MT_MARIO_CAPE) then return end
	if not(cape.flags&MF_MISSILE) then return end //cape is dead
	if not(owner and owner.type == MT_PLAYER) then return end
	//Player vs tagged
	if mo.player.pflags&PF_TAGIT and not(owner.player.pflags&PF_TAGIT) then return false end
	//Players
	if CBW_Battle.PlayerCanBeDamaged(mo.player) and not(CBW_Battle.MyTeam(owner.player,mo.player))
		then
		if G_TagGametype()
			return true
		elseif mo.player.battle_def
			P_InstaThrust(mo,R_PointToAngle2(0,0,owner.momx,owner.momy),owner.scale*10)
			P_SetObjectMomZ(mo, 4*FRACUNIT, false)
			S_StartSound(mo,sfx_s3k7b)
			CBW_Battle.PlayerCreditPusher(mo.player,owner)
			return false
		else
			return true
		end
	end
	return nil
end, MT_PLAYER)

rawset(_G, "DoMarioCape", function(player, mo, cape)
	local cape = player.mariocape
	local hascape = (mo and mo.valid and mo.health
	and IsMario(mo) and (player.powers[pw_shield] == SH_CAPE
	or (player.mariopain and player.mariopain.shield == SH_CAPE)))
	
	if not (cape and cape.valid and cape.type == MT_MARIO_CAPE)
		if hascape
			player.mariocape = P_SpawnMobjFromMobj(mo, 0,0,0, MT_MARIO_CAPE)
			cape = player.mariocape
			cape.target = mo
			cape.tracer = mo
		else
			return
		end
	elseif not hascape
		P_RemoveMobj(cape)
		player.mariocape = nil
		return
	end
	if not cape.name then cape.name = "cape" end
	
	local sprite2 = mo.sprite2 &~ FF_SPR2SUPER &~ FF_SPR2ENDSTATE &~ FF_SPR2MIDSTART
	
	//Slow fall (because I don't feel like doing it elsewhere)
	local slowfall = false
	if player.cmd.buttons & BT_JUMP and not MarioCantMove(player)
	and not P_IsObjectOnGround(mo) and not P_IsObjectInGoop(mo)
	and not player.mariowallslide and not player.mariocapeflight
	and mo.momz * P_MobjFlip(mo) < -CAPEFALL*mo.scale
		P_SetObjectMomZ(mo, -CAPEFALL*FRACUNIT, false)
		slowfall = true
		player.mariogroundpoundtime = min($, 1)
		
		if mo.state == S_PLAY_WALK then mo.state = S_PLAY_FALL end
	end
	
	cape.state = cape.info.spawnstate
	cape.tics = 2
	cape.health = cape.info.spawnhealth
	if (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6)
	or (player.mariopain and not ((mo.tics/4) % 2) and not player.mariopain.mega)
	or sprite2 == SPR2_TAL5
		A_CapeChase(cape, 1, 0)
		return true
	end
	
	//Set it up
	cape.skin = mo.skin
	cape.color = mo.color
	cape.colorized = mo.colorized
	cape.scale = mo.scale
	cape.radius = mo.radius
	cape.height = mo.height
	cape.angle = player.drawangle
	
	cape.rollangle = mo.rollangle
	cape.mirrored = mo.mirrored
	cape.spritexscale = mo.spritexscale
	cape.spriteyscale = mo.spriteyscale
	cape.spritexoffset = mo.spritexoffset
	cape.spriteyoffset = mo.spriteyoffset
	cape.renderflags = mo.renderflags
	cape.blendmode = mo.blendmode
	cape.flags2 = mo.flags2 & (MF2_DONTDRAW|MF2_SHADOW|MF2_SPLAT)
	
	//Disappear in first person
	if not splitscreen and player == displayplayer
	and CV_FindVar("chasecam") and not CV_FindVar("chasecam").value
		cape.flags2 = $ | MF2_DONTDRAW
	end
	
	//Spin attack
	if player.mariospinjump or mo.state == CAPESPINSTATE
		cape.spritexscale = $*2
		
		//Hitbox
		for i = 1, 16
			local angle = ANGLE_22h*i
			local thrust = 64*mo.scale
			thrust = $ + FixedMul(player.mariosize, $)
			
			cape.flags = $ | MF_MISSILE &~ MF_NOCLIP &~ MF_NOCLIPTHING
			
			A_CapeChase(cape, 1, 0)
			P_TeleportMove(cape, cape.x + mo.momx + P_ReturnThrustX(cape, angle, thrust),
			cape.y + mo.momy + P_ReturnThrustY(cape, angle, thrust), cape.z + mo.momz)
			
			if not (cape and cape.valid) then return end
			
			local subsector = R_PointInSubsectorOrNil(cape.x, cape.y)
			local sector = subsector and subsector.valid and subsector.sector
			if sector and sector.valid
				for fof in sector.ffloors()
					if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
					if not (fof.flags & FF_BUSTUP) -- Is it bustable?
					and not (fof.flags & FF_MARIO and fof.flags & FF_SHATTERBOTTOM) continue end
					
					if cape.z - mo.momz + mo.height*2/3 < fof.bottomheight continue end -- Are we too low?
					if cape.z - mo.momz + mo.height/3 > fof.topheight continue end -- Are we too high?  
					
					EV_CrumbleChain(fof) -- Crumble
					if fof.master.flags & ML_EFFECT5
						P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, cape, fof.target)
					end
				end
			end
		end
		
		//Pulse effect
		if not ((player.panim == PA_DASH and mo.tics or (leveltime + #player)) % 4)
			local spark = P_SpawnMobjFromMobj(mo, 0,0, 32*(mo.spritexscale or FRACUNIT), MT_THOK)
			spark.state = S_CACOFIRE_EXPLODE4
			spark.tics = 5
			
			spark.momx = mo.momx
			spark.momy = mo.momy
			spark.momz = mo.momz
			
			spark.spritexscale = mo.spritexscale
			spark.spriteyscale = mo.spriteyscale
			
			spark.scalespeed = FixedMul($*3, spark.scale)
			P_SetScale(spark, spark.scale/4)
			spark.destscale = $*8
			
			spark.momz = $ + FixedMul(spark.scalespeed, spark.info.height/2)
			
			spark.color = SKINCOLOR_APPLE
			spark.colorized = true
			
			spark.blendmode = AST_ADD
			spark.flags2 = $ | MF2_SPLAT | MF2_SHADOW
			if P_MobjFlip(mo) == -1
				spark.z = $ + spark.height*15/4
			end
		end
	end
	
	//Handle sprites
	cape.sprite = SPR_MRCP
	cape.flags = $ &~ MF_MISSILE | MF_NOCLIP | MF_NOCLIPTHING
	
	//Gliding
	if sprite2 == SPR2_GLID
		if mo.skin == "luigi"
			cape.frame = min(O + (mo.frame & FF_FRAMEMASK), R)
		else
			cape.frame = min(K + (mo.frame & FF_FRAMEMASK), N)
		end
		cape.threshold = G
		
	//Spinning
	elseif (player.mariospinjump or (mo.state == CAPESPINSTATE and mo.tics > 0))
	and not slowfall
		cape.frame = ((leveltime/2 + #player) % 2 and C or E)
		cape.threshold = cape.frame
			
	//Other stuff
	elseif (mo.state == S_PLAY_WALK and mo.tics == 1)
	or (mo.state != S_PLAY_WALK and (leveltime + #player) % 3 == 0)
	or (player.mariogroundpound and mo.momz * P_MobjFlip(mo) < 0)
	or ((mo.state == S_PLAY_RUN or mo.state == S_TAILSOVERLAY_FLY) and leveltime & 1)
		local onground = P_IsObjectOnGround(mo)
		or player.powers[pw_carry] == CR_MINECART
		or player.powers[pw_carry] == CR_ROLLOUT
	
		//Falling
		if (not onground and mo.momz * P_MobjFlip(mo) < 0
		and not (player.pflags & PF_SPINNING or sprite2 == SPR2_ROLL or sprite2 == SPR2_TWIN))
		or slowfall or sprite2 == SPR2_SPIN or mo.state == CAPESPINSTATE
			cape.threshold = $ + 1
			if (cape.threshold > D and cape.threshold < G)
			or cape.threshold > J
				cape.threshold = G
			end
			
		//Running
		elseif ((onground and player.speed) or player.pflags & PF_SPINNING
		or sprite2 == SPR2_ROLL or sprite2 == SPR2_TWIN or sprite2 == SPR2_TAL7)
		and not player.skidtime and not player.mariogroundpound
			if cape.threshold < G
				cape.threshold = $ + 1
				if cape.threshold == G
					cape.threshold = C
				end
			elseif cape.threshold >= G
				cape.threshold = D
			end
			
		//Standing
		elseif cape.threshold
		or sprite2 == SPR2_SPNG or sprite2 == SPR2_LAND or sprite2 == SPR2_EDGE
		or player.skidtime or player.mariocrouch or player.mariogroundpound
			cape.threshold = max($ - 1, 0)
			if cape.threshold > D
				cape.threshold = D
			elseif cape.threshold == D or (cape.threshold < B
			and (sprite2 == SPR2_SPNG or sprite2 == SPR2_LAND or sprite2 == SPR2_EDGE
			or player.skidtime or player.mariocrouch or player.mariogroundpound))
				cape.threshold = B
			end
		end
	end
	
	//Set it!
	if not cape.frame then cape.frame = cape.threshold end
	
	//Super color
	if player.powers[pw_super] and not cape.colorized
		cape.color = SKINCOLOR_WHITE
		cape.colorized = true
		cape.frame = $ | FF_FULLBRIGHT
	elseif player.powers[pw_invulnerability]
		cape.frame = $ | FF_FULLBRIGHT
	end
	
	//Offset junk for specific animations
	if sprite2 == SPR2_GLID
		A_CapeChase(cape, 1, 0)
		return true
	end
	
	local table = mo.skin == "luigi" and LCAPESPRITE2[sprite2] or CAPESPRITE2[sprite2]
	local forward, height
	
	//Get the values
	if table
		height = table[1]
		forward = table[2]
	elseif (P_PlayerShouldUseSpinHeight(player) or player.mariocrouch)
	and not player.mariopain
		height = CAPESPRITE2[SPR2_LAND][1]
		forward = CAPESPRITE2[SPR2_LAND][2]
	end
	
	//Set the height offset
	if height
		cape.spriteyoffset = $ + height*FRACUNIT
	end
	
	//Set the foward offset
	if forward == nil
		forward = -8
	else
		forward = $ - 8
	end
	
	//Offset junk for specific animations
	local frame = mo.frame & FF_FRAMEMASK
	if sprite2 == SPR2_WALK
		if frame == A or frame == E
			cape.spriteyoffset = $ - FRACUNIT
		elseif frame == C or frame == G
			cape.spriteyoffset = $ + 3*FRACUNIT
		else
			cape.spriteyoffset = $ + FRACUNIT
		end
	elseif (sprite2 == SPR2_RUN or sprite2 == SPR2_CLMB or sprite2 == SPR2_EDGE)
	and not (frame % 2)
		cape.spriteyoffset = $ + FRACUNIT
		if sprite2 == SPR2_EDGE
			forward = $ - 1
		end
	end
	
	//Move it!
	forward = $ + FixedMul(player.mariosize, $)
	A_CapeChase(cape, 1, forward*FRACUNIT)

	return true
end)