//
//The Lua for Mario's Health System
//

if srb2p then return end

local CAPESPINSTATE = S_MARIO_CAPESPIN //cape.lua

local OFFSETMAX = 26
local OFFSETFACTOR = 2
local MAXDELAY = TICRATE
local PULSATETIME = TICRATE/4
local PULSATEFACTOR = 24
local LOWHEATHTIME = TICRATE*5

local MarioHUD = {[0] = {}, [1] = {}}

MarioHUD[0].healthyoffset = 0
MarioHUD[1].healthyoffset = 0

MarioHUD[0].delay = 0
MarioHUD[1].delay = 0

MarioHUD[0].oldhealth = 0
MarioHUD[1].oldhealth = 0

MarioHUD[0].pulsate = 0
MarioHUD[1].pulsate = 0

MarioHUD[0].lowhealth = 0
MarioHUD[1].lowhealth = 0

hud.add(function(v, player, cam)
	if player.mariohealth == nil or player.mariocoins == nil or G_IsSpecialStage() return end
	local mo = player.mo
	local id = player == secondarydisplayplayer and 1 or 0
	local mrhud = MarioHUD[id]
	local alive = (mo and mo.valid and mo.health and not player.playerstate)
	
	//Moving the health out of view
	if player.spectator or not IsMario(player)
	or (alive and player.mariohealth and player.mariohealth >= 3 and not MarioUltimateMode())
		if mrhud.delay
			mrhud.delay = alive and player.mariohealth and $ - 1 or 0
			if mrhud.healthyoffset < OFFSETMAX
				mrhud.healthyoffset = min($ + 4, OFFSETMAX)
			end
		elseif mrhud.healthyoffset
			mrhud.healthyoffset = $ - 1
		end
	elseif mrhud.healthyoffset < OFFSETMAX
		mrhud.healthyoffset = min($ + 4, OFFSETMAX)
		mrhud.delay = MAXDELAY
	end
	
	local newhealth = alive and player.mariocoins*10 + player.mariohealth or 0
	if mrhud.oldhealth != newhealth
		mrhud.pulsate = PULSATETIME
		
		if newhealth == 1
			mrhud.lowhealth = LOWHEATHTIME
		end
	elseif mrhud.pulsate
		mrhud.pulsate = $ - 1
	end
	mrhud.oldhealth = newhealth
	
	if mrhud.lowhealth and player.mariohealth == 1
		mrhud.lowhealth = $ - 1
		
		if not (mrhud.lowhealth%TICRATE)
			local factor = max((LOWHEATHTIME*4/5 - mrhud.lowhealth)/TICRATE, 1)
			mrhud.pulsate = max($, PULSATETIME/factor)
			S_StartSoundAtVolume(nil, sfx_mrlowh, 255/factor, player)
		end
	elseif mrhud.lowhealth
		mrhud.lowhealth = 0
	end
	
	if mrhud.healthyoffset
		local coins = alive and min(player.mariocoins, 3) or 0
		local health = alive and min(player.mariohealth, 3) or 0
		if health and MarioUltimateMode() then health = 4 end
		
		local scale = FRACUNIT + mrhud.pulsate*FRACUNIT/PULSATEFACTOR
		local x = 288 * FRACUNIT
		local y = (mrhud.healthyoffset*OFFSETFACTOR + mrhud.pulsate) * FRACUNIT 
		local flags = V_HUDTRANS|V_SNAPTOTOP|V_SNAPTORIGHT|V_PERPLAYER
		
		local patch = v.cachePatch("MRCOIN"..coins)
		local width = patch.width*scale
		local height = patch.height
		height = ($ + patch.topoffset - ($ - v.height()/v.dupy()))*scale
	
		v.drawCropped(x, y, scale, scale, patch, flags, nil, 0,0, width, height)
		v.drawCropped(x, y, scale, scale, v.cachePatch("MRBAR"..health),
		flags, nil, 0,0, width, height)
	end
end)

//Make it so the health icon can actually be seen after the title card is over
hud.add(function(v, player, ticker, endtime)
	if ticker >= endtime then return end
	
	if displayplayer and displayplayer.valid
	and displayplayer.mo and displayplayer.mo.valid
	and IsMario(displayplayer)
		MarioHUD[0].delay = max($, MAXDELAY)
	end
	
	if secondarydisplayplayer and secondarydisplayplayer.valid
	and secondarydisplayplayer.mo and secondarydisplayplayer.mo.valid
	and IsMario(secondarydisplayplayer)
		MarioHUD[1].delay = max($, MAXDELAY)
	end
end, "titlecard")

//Hooray for multiple PlayerSpawn hooks
addHook("PlayerSpawn", function(player)
	local mo = player.mo
	if player != displayplayer and player != secondarydisplayplayer then return end
	if mo and mo.valid and IsMario(mo)
		if player == displayplayer
			MarioHUD[0].delay = max($, MAXDELAY)
		else
			MarioHUD[1].delay = max($, MAXDELAY)
		end
	end
end)

//Special HP Damage Script
addHook("MobjDamage", function(mo, inf, src, dmg, dmgtype)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if player and not (dmgtype&DMG_DEATHMASK)
	and player.powers[pw_carry] != CR_NIGHTSMODE
	and not player.powers[pw_invulnerability]
	and not player.powers[pw_flashing]
	and not player.powers[pw_super]
	and not G_IsSpecialStage()
	and not player.ultdeath
	and (not player.guard or player.guard <= 0)
		//Shoe death
		if player.marioshoe and player.marioshoe.valid
			P_KillMobj(player.marioshoe)
			
			player.marioshoe.tracer = nil
			player.marioshoe = nil
			
			player.powers[pw_flashing] = max(flashingtics - TICRATE, TICRATE)
			mo.state = S_PLAY_FALL
			return true
			
		//Cape flight hit
		elseif player.mariocapeflight and not P_IsObjectOnGround(mo)
			player.powers[pw_flashing] = max(flashingtics - TICRATE, TICRATE)
			S_StartSound(mo, sfx_mrbtls)
			
			mo.state = CAPESPINSTATE
			player.panim = PA_ABILITY2
			mo.tics = -1
			return true
		
		//Die if we don't have any health left
		elseif ((player.mariohealth <= 1 or MarioUltimateMode())
		and not player.powers[pw_shield]) or player.powers[pw_shield] == SH_MINI
			P_KillMobj(mo, inf, src, dmgtype)
			
			P_PlayerEmeraldBurst(player)
			P_PlayerWeaponPanelOrAmmoBurst(player)
			P_PlayerFlagBurst(player)
			
			player.powers[pw_shield] = 0
			P_SpawnShieldOrb(player)
			
			mo.health = 0
			player.mariohealth = 0
			return true
			
		//Decrease the HP and give ourselves a temporary shield to bypass losing rings
		else
			local alreadyinpain = player.mariopain
			
			if not alreadyinpain
				player.mariopain = {
					momx = mo.momx,
					momy = mo.momy,
					momz = mo.momz,
					state = mo.state,
					frame = mo.frame,
					sprite2 = mo.sprite2,
					tics = mo.tics,
					flags = mo.flags,
					pflags = player.pflags,
					drawangle = player.drawangle,
					flashing = player.powers[pw_flashing],
					shield = player.powers[pw_shield],
					dmgtype = dmgtype,
					groundpound = player.mariogroundpound,
					spinjump = player.mariospinjump,
					spinsfx = player.mariospinsfx,
					gpjump = player.mariogpjump,
					crouch = player.mariocrouch,
					flipmove = player.marioflipmove,
					capeflight = player.mariocapeflight,
				}
			end
			
			if not player.powers[pw_shield]
				if not (gametyperules & GTR_RINGSLINGER)
					P_PlayerEmeraldBurst(player)
					P_PlayerWeaponPanelOrAmmoBurst(player)
					player.powers[pw_shield] = SH_PITY
				elseif player.rings <= 0
					player.powers[pw_shield] = SH_PITY
				end
				
				player.mariohealth = $ - 1
			end
			
			if dmgtype == DMG_FIRE and not inf and not src
				if not alreadyinpain
					player.mariopain.lava = true
				end
				
				MarioPlayVoice(player, "lava")
				S_StartSound(mo, sfx_s3k4b)
				S_StartSound(mo, sfx_s3k48)
			else
				MarioPlayVoice(player, "pain")
			end
			
			if not player.powers[pw_carry]
				P_ResetPlayer(player)
			end
			
			mo.flags = MF_NOCLIPTHING|MF_NOGRAVITY
			player.pflags = $|PF_FULLSTASIS|PF_INVIS
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			if P_IsObjectOnGround(mo) then mo.z = $ + P_MobjFlip(mo) end
			
			player.powers[pw_flashing] = flashingtics
			mo.state = S_PLAY_PAIN
			player.panim = PA_DASH
		end
	end
end, MT_PLAYER)