//
//Super Mario's main moveset and abilities
//

rawset(_G, "IsMario", function(mario)
	return mario and mario.valid and mario.skin
	and (skins[mario.skin].name == "mario" or skins[mario.skin].name == "luigi")
end)

local function SpawnMarioAfterimage(mo, tics)
	local player = mo and mo.valid and mo.player
	if not player then return end
	
	local ghost = P_SpawnMobjFromMobj(mo, 0,0,0, MT_OVERLAY)
	if ghost and ghost.valid
		ghost.state = S_INVISIBLE
		ghost.tics = tics
		
		ghost.target = mo
		ghost.tracer = mo
		ghost.flags2 = $|MF2_LINKDRAW
		
		ghost.skin = mo.skin
		ghost.sprite = mo.sprite
		ghost.sprite2 = mo.sprite2
		ghost.frame = mo.frame|FF_TRANS50
		
		ghost.spritexscale = mo.spritexscale
		ghost.spriteyscale = mo.spriteyscale
		
		if VERSIONSTRING == "v2.2.9" //Won't need this for 2.2.10 onwards
		or (srb2p and VERSION <= 103) //SRB2P check too
			ghost.movedir = mo.player.drawangle - mo.angle
		end
	end
end

local DOUBLEFALLSTATE
local DoSpin
function A_SetMarioPanim(mo, var1, var2)
	local player = mo and mo.valid and mo.player
	if not player then return end
	
	player.panim = var2
	
	if var1 and player.marioprevstate == mo.state
		if mo.frame & FF_FRAMEMASK == D
			if player.marioflipmove
				player.marioflipmove = $ + 1
				if player.marioflipmove > 4
					if mo.skin == "luigi"
						DoSpin(player, true)
					else
						mo.state = DOUBLEFALLSTATE
					end
					if player.pflags & PF_JUMPED
						player.pflags = $ | PF_NOJUMPDAMAGE
					end
					return
				end
			end
			S_StartSoundAtVolume(mo, sfx_mrtrfl, 128)
		end
		
		SpawnMarioAfterimage(mo, mo.tics + 2)
	end
end

//Constants - Please update the ones in bonus.lua accordingly!

local JUMPSTATE = freeslot("S_MARIO_JUMP1")
states[S_MARIO_JUMP1] = {SPR_PLAY, SPR2_TAL0, 3, A_SetMarioPanim, 0, PA_FALL, freeslot("S_MARIO_JUMP2")}
states[S_MARIO_JUMP2] = {SPR_PLAY, SPR2_JUMP, 1, A_SetMarioPanim, 0, PA_FALL, S_MARIO_JUMP2}
local FALLSTATE = freeslot("S_MARIO_FALL")
states[S_MARIO_FALL] = {SPR_PLAY, SPR2_FLT_, 2, A_SetMarioPanim, 0, PA_FALL, S_MARIO_FALL}

local DOUBLEJUMPSTATE = freeslot("S_MARIO_DOUBLE")
states[S_MARIO_DOUBLE] = {SPR_PLAY, SPR2_SPNG, 2, A_SetMarioPanim, 0, PA_JUMP, S_MARIO_DOUBLE}
DOUBLEFALLSTATE = freeslot("S_MARIO_DOUBLEFALL")
states[S_MARIO_DOUBLEFALL] = {SPR_PLAY, SPR2_FALL, 2, A_SetMarioPanim, 0, PA_FALL, S_MARIO_DOUBLEFALL}

local TRIPLEJUMPSTATE = freeslot("S_MARIO_TRIPLE")
states[S_MARIO_TRIPLE] = {SPR_PLAY, SPR2_ROLL, 3, A_SetMarioPanim, 1,  PA_FALL, S_MARIO_TRIPLE}
local TRIPLETIME = 5

local WALLJUMPSTATE = DOUBLEJUMPSTATE
local WALLSLIDESTATE = S_PLAY_CLING
local WALLJUMPFALL = 4
local WALLJUMPTHRUST = 12

local SWIMSTATE = S_PLAY_SWIM

local SPININPUTTIME = 10
local SPINSTATE = freeslot("S_MARIO_SPIN")
states[S_MARIO_SPIN] = {SPR_PLAY, SPR2_BNCE, 2, A_SetMarioPanim, 0, PA_ABILITY, S_MARIO_SPIN}
local GROUNDSPINTIME = TICRATE
local GROUNDSPINGRACE = GROUNDSPINTIME/2

local DIVESTATE = S_PLAY_SPINDASH
local DIVERECOVERYSTATE = freeslot("S_MARIO_RECOVER")
states[S_MARIO_RECOVER] = {SPR_PLAY, SPR2_ROLL|FF_ANIMATE, 8, nil, 3,  2, FALLSTATE}
local DIVESPEED = 32

local LONGJUMPSTATE = freeslot("S_MARIO_LONGJUMP")
states[S_MARIO_LONGJUMP] = {SPR_PLAY, SPR2_SPNG, 8, A_SetMarioPanim, 0, PA_FALL, freeslot("S_MARIO_LONGTRANS")}
states[S_MARIO_LONGTRANS] = {SPR_PLAY, SPR2_TAL0, 3, A_SetMarioPanim, 0, PA_FALL, freeslot("S_MARIO_LONGFALL")}
states[S_MARIO_LONGFALL] = {SPR_PLAY, SPR2_TAL9, 2, A_SetMarioPanim, 0, PA_FALL, S_MARIO_LONGFALL}
local LONGJUMPSPEED = 48

local BONKSTATE = freeslot("S_MARIO_BONK")
states[S_MARIO_BONK] = {SPR_PLAY, SPR2_STUN, TICRATE*5, A_SetMarioPanim, 0, PA_DASH, S_PLAY_WALK}
local BONKTIMETICS = states[S_MARIO_BONK].tics
local BONKLANDTICS = 15
local GLIDETIMETIME = 16

local GROUNDPOUNDSTART = 10
local GROUNDPOUNDEND = GROUNDPOUNDSTART*2
local GROUNDPOUNDENDSTATE = freeslot("S_MARIO_GROUNDPOUND")
states[S_MARIO_GROUNDPOUND] = {SPR_PLAY, SPR2_TIRE, -1, A_SetMarioPanim, 0, PA_ABILITY, S_PLAY_WALK}
local GROUNDPOUNDSTARTSTATE = freeslot("S_MARIO_POUNDSTART")
states[S_MARIO_POUNDSTART] = {SPR_PLAY, SPR2_ROLL|FF_SPR2ENDSTATE, 2, nil, GROUNDPOUNDENDSTATE, 0, S_MARIO_POUNDSTART}

local CROUCHSTATE = S_PLAY_GLIDE_LANDING
local CRAWLSTATE = S_PLAY_CLIMB
local ROLLSTATE = freeslot("S_MARIO_ROLL")
states[S_MARIO_ROLL] = {SPR_PLAY, SPR2_SWIM, 4, nil, 0, 0, freeslot("S_MARIO_ROLL2")}
states[S_MARIO_ROLL2] = {SPR_PLAY, SPR2_SPIN, 4, nil, 0, 0, S_PLAY_ROLL}

local BACKFLIPSTATE = freeslot("S_MARIO_BACKFLIP")
states[S_MARIO_BACKFLIP] = {SPR_PLAY, SPR2_TWIN, 3, A_SetMarioPanim, 2, PA_FALL, S_MARIO_BACKFLIP}
local CHARGETIME = 15
local BACKFLIPSPEED = 8

local SCUTTLESTATE = freeslot("S_MARIO_SCUTTLE")
states[S_MARIO_SCUTTLE] = {SPR_PLAY, SPR2_WALK, 1, A_SetMarioPanim, 0, PA_FALL, S_MARIO_SCUTTLE}

local SIDEFLIPSTATE = freeslot("S_MARIO_SIDEFLIP")
states[S_MARIO_SIDEFLIP] = {SPR_PLAY, SPR2_TWIN|FF_ANIMATE, 12, nil, 3,  2, S_PLAY_SPRING}
local SIDEFLIPSPEED = 16

local LEDGEGRABSTATE = S_TAILSOVERLAY_PLUS30DEGREES
local PULLUPSTATE = freeslot("S_MARIO_PULLUP")
local PULLUPSTATE2 = freeslot("S_MARIO_PULLUP2")
states[S_MARIO_PULLUP] = {SPR_PLAY, SPR2_SWIM, 4, nil, 0,  0, S_MARIO_PULLUP2}
states[S_MARIO_PULLUP2] = {SPR_PLAY, SPR2_CLMB|FF_ANIMATE, 4, nil, 1,  2, S_PLAY_EDGE}

local HAMMERTICS = 6
local HAMMERDIST = 38

local CAPESPINSTATE = S_MARIO_CAPESPIN //cape.lua
local PSPEEDTIME = TICRATE/2
local CAPEJUMPTIME = TICRATE*2/3
local CAPETURNFACTOR = 16
local CAPEFALLTIME = 8

local MINCAPEMOM = 5
local MAXCAPEMOM = 12
local CAPEMOMDIV = 3
local CAPEMOMFACTOR = 16

local CAPEFLYSTATE = freeslot("S_MARIO_CAPEFLY")
states[S_MARIO_CAPEFLY] = {SPR_PLAY, SPR2_GLID, -1, A_SetMarioPanim, 0, PA_DASH, S_MARIO_CAPEFLY}

local SQUISHTIME = 8
local SQUISHFACTOR = 32
local BALLOONSTATE = freeslot("S_MARIO_BALLOON")
states[S_MARIO_BALLOON] = {SPR_PLAY, SPR2_FLY_, -1, A_SetMarioPanim, 0, PA_DASH, S_MARIO_BALLOON}

local LAVASTATE = freeslot("S_MARIO_LAVA")
states[S_MARIO_LAVA] = {SPR_PLAY, SPR2_PAIN, -1, nil, 0, 0, S_MARIO_LAVA}

local RUNSPEED = 24
local SKIDSPEED = 4
local SKIDTIME = TICRATE*4
local SKIDCANCEL = SKIDTIME - TICRATE/4

//Get the player's input angle
rawset(_G, "MarioInputAngle", function(player)
	return player.marioinputangle
end)

//Spawns objects in a circle
rawset(_G, "MarioCircle", function(mobj, mobjtype, amount, ns, zoffset, scale, keepmom, color, vertical, vertangle)
	if not amount then return end
	local i = 0	while i < amount
		local fa = i*(ANGLE_180/(amount/2)) + (vertangle or 0)
		local mo = P_SpawnMobjFromMobj(mobj,0,0,zoffset,mobjtype)
		if mo and mo.valid
			local height = mo.height
			mo.scale = scale
			
			if mo.flags & MF_MISSILE
				mo.target = mobj
			end
			
			if mobjtype == MT_GHOST
				mo.state = S_INVISIBLE
				mo.sprite = SPR_MRPT
				mo.frame = FF_FULLBRIGHT
				mo.tics = 12
				
				mo.flags = mobjinfo[MT_THOK].flags
				mo.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
				mo.color = SKINCOLOR_SUPERGOLD5
				mo.blendmode = AST_ADD
				
				if not vertical
					if mobj.player and mobj.player.mariomega
					and mobj.player.mariopain and mobj.player.mariopain.mega
						mo.momz = 12*mo.scale*P_MobjFlip(mo)
					else
						mo.momz = 4*mo.scale*P_MobjFlip(mo)
					end
				end
			elseif mobjtype == MT_MINECARTSPARK
				mo.fuse = TICRATE/2
			end
			
			if mobj.eflags&MFE_VERTICALFLIP
				mo.z = $ - (mo.height-height)
			end
			
			if color
				mo.color = color
				if mobjtype != MT_GHOST
					mo.colorized = true
				end
			end
			
			if vertical
				mo.momz = FixedMul(sin(fa),ns)/2
				P_InstaThrust(mo, vertangle+ANGLE_90,
				FixedMul(cos(fa),ns)/2)
			else
				mo.momx = FixedMul(cos(fa),ns)/2
				mo.momy = FixedMul(sin(fa),ns)/2
				mo.angle = fa
			end
			
			if keepmom
				mo.momx = $ + mobj.momx/keepmom
				mo.momy = $ + mobj.momy/keepmom
				mo.momz = $ + mobj.momz/keepmom
			end
		end
		i = $ + 1
	end
end)

//Ya can't move!
rawset(_G, "MarioCantMove", function(player)
	local mo = player.mo
	return not (mo and mo.valid)
	or P_PlayerInPain(player)
	or player.pflags & PF_SLIDING or player.mariowassliding
	or player.powers[pw_carry]
	or player.spectator
	or player.playerstate
	or not mo.health
	or player.powers[pw_super] == 1
	or player.iseggrobo
	or player.isjettysyn
	or player.guard
	or player.actionstate
	or (player.airdodge and player.airdodge > 0)
	or player.tumble
	or player.powers[pw_nocontrol]
	or (player.isxmomentum and mo.state == S_PLAY_FACEPLANT)
	or player.mariopain
end)

//Reset Mario's variables
rawset(_G, "ResetMario", function(player, bonk, triple)
	local mo = player.mo
	
	if not triple
		player.mariotriple = 0
		player.mariotripletime = 0
	end
	
	if player.mariofiredelay == nil
		player.mariofiredelay = 0
		player.mariosize = 0
		player.mariomega = 0
		player.mariowassuper = false
	end
	
	player.mariojumpdir = nil
	
	player.mariowalljump = 0
	player.mariowallangle = player.drawangle
	player.mariowallslide = 0
	player.marioslidesfx = 0
	
	player.marioprevinput = {}
	player.mariodidcircle = 0
	player.mariospinsfx = 0
	player.mariospinjump = 0
	
	player.luigispin = 0
	player.luigiscuttle = false
	
	player.marioslide = nil
	
	player.mariogroundpound = 0
	player.mariogroundpoundtime = 0
	player.mariogpjump = 0
	
	player.mariocrouch = 0
	player.mariocharge = 0
	
	player.marioflipmove = 0
	
	player.mariopspeed = 0
	player.mariocapemom = 0
	player.mariocapeflight = 0
	
	if player.mariolongjump
		player.normalspeed = skins[player.skin].normalspeed
		player.mariolongjump = 0
	end
	
	if not bonk
		player.mariobouncedir = nil
		player.mariobouncespeed = nil
		player.mariobonkland = 0
		player.mariolongbonk = 0
	end
	
	if player.mariohealth == nil or player.playerstate
	or not (mo and mo.valid and mo.mariohealthset)
		player.mariohealth = 3
		player.mariocoins = 0
		
		if mo and mo.valid
			mo.mariohealthset = true
		end
	end
	
	if player.marioprevmomz == nil
		if mo and mo.valid
			player.marioprevz = mo.z
			player.marioprevmomz = mo.momz
		else
			player.marioprevz = 0
			player.marioprevmomz = 0
		end
	end
	
	//Debug power ups
	/*
	if leveltime <= 2
		player.marioshield = SH_WHIRLWIND
		//player.mariokeepshoe = true
	end
	*/
end)
addHook("PlayerSpawn", ResetMario)
for player in players.iterate
	ResetMario(player)
end

//Activate Mario's Dive
local function DoDive(player, groundpound, directional)
	local mo = player.mo
	mo.state = DIVESTATE
	mo.sprite2 = SPR2_SWIM
	mo.tics = 4
	
	if player.powers[pw_shield] == SH_BALLOON or (player.powers[pw_super]
	and groundpound and player.mariogroundpound == GROUNDPOUNDSTART)
		MarioCircle(mo, MT_MARIO_DUST, 8, 12*mo.scale,
		player.height/2, mo.scale/2, 0, nil, true, player.drawangle)
		
		S_StartSound(mo, sfx_mrpbds)
		if player.powers[pw_shield] == SH_BALLOON
			player.mariosquish = SQUISHTIME
		else
			mo.state = S_PLAY_ROLL
		end
	else
		S_StartSound(mo, sfx_mrdive)
		MarioPlayVoice(player, "dive")
	end
	
	player.mariowalljump = 0
	player.marioflipmove = 0
	player.pflags = $ | PF_SPINNING &~ PF_JUMPED &~ PF_THOKKED
	
	//Spindash Switch
	if mo.subsector and mo.subsector.valid
	and mo.subsector.sector and mo.subsector.sector.valid
		for fof in mo.subsector.sector.ffloors()
			if fof.valid and mo.z + mo.momz <= fof.topheight
			and mo.z + mo.momz + mo.height >= fof.bottomheight
			and fof.master and fof.master.valid and fof.master.flags&ML_NOCLIMB
			and (fof.master.special == 153 or fof.master.special == 180
			or (fof.master.special >= 150 and fof.master.special <= 152)
			or (fof.master.special >= 190 and fof.master.special <= 195)
			or fof.master.special == 176 or fof.master.special == 177)
				player.pflags = $|PF_STARTDASH
				P_InstaThrust(player.mo,0,0)
				return
			end
		end
	end
	
	//Horizontal momentum
	mo.momx = $/2
	mo.momy = $/2
	
	local speed = DIVESPEED*mo.scale
	if groundpound then speed = max($, LONGJUMPSPEED*mo.scale) end
	if mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_GOOWATER)
	and not player.mariowaterswitch
		speed = $/2
	end
	
	//Limit the influence on the dive angle
	if player.marioprevstate == SPINSTATE or directional
		player.drawangle = R_PointToAngle2(0,0,mo.momx,mo.momy)
	elseif groundpound
		player.drawangle = MarioInputAngle(player)
	end
	local diff = player.drawangle - MarioInputAngle(player)
		
	if (abs(diff) <= ANGLE_112h and abs(diff) >= 0) or P_IsObjectOnGround(mo)
		player.drawangle = $ - diff/2
	else
		speed = $/2
	end
	player.mariojumpdir = player.drawangle
	P_Thrust(mo, player.drawangle, speed)
	
	//Vertical momentum
	local verticalthrust = P_GetMobjGravity(mo)
	if mo.eflags & MFE_GOOWATER then verticalthrust = -gravity*P_MobjFlip(mo) end
	if abs(verticalthrust) > mo.scale/2 then verticalthrust = mo.scale/-2*P_MobjFlip(mo) end
	if P_IsObjectOnGround(mo)
		mo.z = $ + P_MobjFlip(mo)
		mo.momz = -8 * verticalthrust
	elseif mo.momz*P_MobjFlip(mo) < 0
		mo.momz = -4 * verticalthrust
	else
		mo.momz = $ + -4 * verticalthrust
	end
end

//Activate Mario's Hammer
local function DoHammer(player, nomomz, fire)
	if MarioPowerupAbility(player, fire) or player.marioshoe then return end
	
	local mo = player.mo
	
	mo.state = S_PLAY_MELEE
	player.panim = PA_DASH
	
	player.drawangle = MarioInputAngle(player)
	
	player.powers[pw_noautobrake] = max($, 2)
	S_StartSound(mo, sfx_mrhmr1)
	
	if not nomomz
	and not P_IsObjectOnGround(mo)
		local momz = mo.momz
		player.pflags = $ &~ PF_JUMPED | PF_THOKKED
		P_DoJump(player, false)
		mo.momz = $*4/5
		if momz*P_MobjFlip(mo) > mo.momz*P_MobjFlip(mo)
			mo.momz = momz
		end
	elseif P_IsObjectOnGround(mo)
		player.pflags = $ | PF_JUMPSTASIS
	end
end

//Activate Mario's Spin
DoSpin = function(player, luigi)
	local mo = player.mo
	if not player.mariodidcircle then player.mariodidcircle = 1 end
	
	player.mariospinjump = player.mariodidcircle * GROUNDSPINTIME
	mo.state = SPINSTATE
	
	if player.skidtime
		player.skidtime = 0
		player.marioskid = nil
		S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
	end
	
	player.luigispin = luigi and 1 or 0
	if not luigi
		MarioPlayVoice(player, "spin")
	end
	
	if player.powers[pw_shield] == SH_CAPE then S_StartSound(mo, sfx_mrspin) end
	
	if P_IsObjectOnGround(mo)
		player.mariospinsfx = 1
	else
		if not luigi
			player.mariospinsfx = -22
		end
			
		if not (player.pflags & PF_THOKKED)
			if mo.momz*P_MobjFlip(mo) < -4*P_GetMobjGravity(mo)*P_MobjFlip(mo)
				mo.momz = -4*P_GetMobjGravity(mo)
			end
			player.pflags = $ | PF_THOKKED
		end
		
		if mo.momz*P_MobjFlip(mo) > -4*P_GetMobjGravity(mo)*P_MobjFlip(mo)
			if player.mariowalljump and player.pflags & PF_STARTJUMP
				mo.momz = $/2
				player.pflags = $ &~ PF_STARTJUMP
			else
				mo.momz = $*4/5
			end
		end
	end
end

//Activate Mario's Ground Pound Hit
rawset(_G, "DoMarioGroundPoundHit", function(player, nodamage)
	local mo = player.mo
	
	S_StartSound(mo, player.mariosize < 0 and sfx_mrmin6 or sfx_mrgnd2)
	if player.marioshoe or player.mariosize > 0
		S_StartSoundAtVolume(mo, sfx_thwomp, 128)
	end
	
	local scale = mo.scale + FixedMul(player.mariosize, mo.scale)
	
	if player.marioshoe
		MarioCircle(mo, MT_MARIO_GHOST, 6,
		FixedMul(mobjinfo[MT_MARIO_GHOST].speed, scale),
		0, scale, 0, nil, false, player.drawangle)
	else
		MarioCircle(mo, MT_GHOST, 8, 32*scale, 0, scale,
		0, nil, false, player.drawangle)
		
		MarioCircle(mo, MT_MARIO_DUST, 8, 12*scale,
		-8*FRACUNIT, scale/2, 0, nil, false, player.drawangle)
	end
	
	if player.mariosize > 0
		local dist = RING_DIST*3
		if not nodamage
			P_RadiusAttack(mo, mo, dist, DMG_NUKE, false)
		end
		
		if player == displayplayer or player == secondarydisplayplayer
		or (displayplayer and displayplayer.valid
		and displayplayer.mo and displayplayer.mo.valid
		and P_AproxDistance(mo.x - displayplayer.mo.x,
		mo.y - displayplayer.mo.y) < dist)
			P_StartQuake(32*mo.scale, 5)
		end
		
		if not nodamage and not (leveltime % 3)
			local mobjtype = MT_FLINGCOIN //P_RandomChance(FRACUNIT/3) and MT_GOOMBA or MT_FLINGCOIN
			
			for i = 1, 4
				local coin = P_SpawnMobjFromMobj(mo,
				P_SignedRandom()*FRACUNIT, P_SignedRandom()*FRACUNIT,
				0, mobjtype)
				
				coin.z = P_CeilingzAtPos(coin.x, coin.y,
				coin.z, coin.height) - coin.height
				P_InstaThrust(coin, R_PointToAngle2(mo.x, mo.y, coin.x, coin.y),
				mo.scale)
				
				if mobjtype == MT_FLINGCOIN
					coin.fuse = MARIO_DESPAWNTIME
					coin.mariomonitorcoin = true
					coin.target = mo
					coin.flags = $ | MF_FLOAT
				end
			end
		end
	elseif player == displayplayer or player == secondarydisplayplayer
		P_StartQuake(8*mo.scale, 5)
	end
end)

//Activate Mario's Hammer Hit
rawset(_G, "DoMarioHammerHit", function(player, hit)
	local mo = player.mo
	mo.state = S_PLAY_MELEE_LANDING
	mo.tics = HAMMERTICS
	player.pflags = $ &~ PF_THOKKED
	
	local dist = HAMMERDIST*FRACUNIT
	dist = $ + FixedMul(player.mariosize, $)
	
	local dust = P_SpawnMobjFromMobj(mo,
	P_ReturnThrustX(nil, player.drawangle, dist),
	P_ReturnThrustY(nil, player.drawangle, dist),
	0, MT_DUST)
	
	dust.state = S_INVISIBLE
	dust.tics = 1
	
	dust.radius = mo.radius
	dust.height = mo.radius
	dust.flags = $ &~ MF_NOCLIPHEIGHT
	dust.z = $
	
	if hit
		MarioDoStars(dust)
		MarioDoHit(dust, true)
		S_StartSound(mo, sfx_mrhhit)
	end
	
	if P_IsObjectOnGround(dust) or not P_IsObjectOnGround(mo)
	or (mo.eflags & MFE_JUSTHITFLOOR and mo.eflags & MFE_ONGROUND)
		S_StartSound(mo, sfx_mrhmr2)
		
		MarioCircle(dust, MT_GHOST, 8, 16*mo.scale, 0,
		mo.scale/2, 0, nil, false, player.drawangle)
		
		if player == displayplayer or player == secondarydisplayplayer
			P_StartQuake(8*mo.scale, 2)
		end
		
		if dust.subsector and dust.subsector.valid
		and dust.subsector.sector and dust.subsector.sector.valid
			for fof in dust.subsector.sector.ffloors()
				if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
				if not (fof.flags & FF_BUSTUP) -- Is it bustable?
				and not (fof.flags & FF_MARIO) continue end
				
				if dust.z + mo.momz + mo.height < fof.bottomheight continue end -- Are we too low?
				if dust.z + mo.momz > fof.topheight continue end -- Are we too high?  
				
				if fof.flags & FF_MARIO
				and not (fof.flags & (FF_BUSTUP|FF_SHATTERBOTTOM))
					MarioGroundPoundBlock(mo, fof)
				else
					EV_CrumbleChain(fof) -- Crumble
					if fof.master.flags & ML_EFFECT5
						P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, mo, fof.target)
					end
				end
			end
		end
		
		if not P_IsObjectOnGround(mo)
			return
		end
		
	//elseif P_IsObjectOnGround(mo)
	//	P_RemoveMobj(dust)
	end
	
	mo.momx = $/2
	mo.momy = $/2
end)

//Make the attacks damage
addHook("PlayerCanDamage", function(player, mobj)
	local mo = player.mo
	if not (mo and mo.valid and IsMario(mo)) then return end
	if player.mariosize > 0 and not player.marioshoe
		return true
		
	elseif (player.rmomx or player.rmomy or not P_IsObjectOnGround(mo))
	and (player.powers[pw_shield] == SH_GOLDFLOWER
	or player.powers[pw_shield] == SH_METAL)
	and not player.marioshoe
		return true
		
	elseif player.marioshoe and not P_IsObjectOnGround(mo)
	and mo.momz * P_MobjFlip(mo) < 0
		return true
		
	elseif (player.powers[pw_shield] == SH_CAPE and player.mariospinjump)
	or mo.state == CAPESPINSTATE
		return true
		
	elseif mo.state == S_PLAY_MELEE or mo.state == S_PLAY_MELEE_FINISH
	or mo.state == S_PLAY_MELEE_LANDING
	or (player.marioshoe and not P_IsObjectOnGround(mo))
	or (mo.state == S_PLAY_FIRE and (mobj and mobj.valid and mobj.flags & MF_MONITOR))
		local angle = player.drawangle - R_PointToAngle2(player.mo.x-player.mo.momx,
		player.mo.y-player.mo.momy, mobj.x, mobj.y)
		if angle > abs(ANGLE_180)
			angle = $+ANGLE_180*2
		elseif angle < ANGLE_180
			angle = $-ANGLE_180*2
		end
		if abs(angle) < ANGLE_90
			return true
		end
		
	elseif player.mariosize < 0 and not player.mariogroundpound
	and mobj and mobj.valid and mobj.flags & MF_SOLID
		return false
		
	elseif mo.state == DIVERECOVERYSTATE
		return true
		
	elseif mo.mariohithead and mobj and mobj.valid and mobj.flags & MF_SOLID
	and ((P_MobjFlip(mo) > 0 and mo.z + mo.height < mobj.z)
	or (P_MobjFlip(mo) < 0 and mo.z > mobj.z + mobj.height))
		return true
	end
end)

//Ground Pound/Hammer Damage plus Boss Coins
addHook("MobjDamage", function(mobj, inf, mo, dmg, dmgtype)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if not player then return end
	
	//Boss Coins
	if mobj and mobj.valid and mobj.flags & MF_BOSS and player.mariohealth < 3
		local coin
		coin = P_SpawnMobjFromMobj(mobj, 0,0,mobj.info.height/2, MT_FLINGCOIN)
		P_SetObjectMomZ(coin, 8*FRACUNIT, false)
		P_Thrust(coin, R_PointToAngle2(mo.x, mo.y, mobj.x, mobj.y), 4*FRACUNIT)
		if mobj.type == MT_EGGMOBILE4 then coin.momx = $*6 coin.momy = $*6 end
		coin.fuse = MARIO_DESPAWNTIME
		coin.mariomonitorcoin = true
		
		local ov = P_SpawnMobjFromMobj(coin, 0,0,0, MT_OVERLAY)
		ov.target = coin
		ov.state = coin.info.deathstate
		
		S_StartSound(coin, sfx_mrcnap)
	end
	
	//Ground Pound/Hammer Damage
	if inf != mo or MarioCantMove(player) then return end
	if player.mariogroundpound and player.mariogroundpound <= GROUNDPOUNDSTART
		local dumbgravity = 24*abs(P_GetMobjGravity(mo))
		mo.momz = min(max($, -dumbgravity), dumbgravity)
		
		if mobj.info.spawnhealth > 1
			DoMarioGroundPoundHit(player, true)
		end
		
		if mobj and mobj.valid and mobj.health > 1 and not dmgtype and dmg <= 1
		and player.mariosize >= 0
			P_DamageMobj(mobj, mo, src, dmg*2, dmgtype)
			return true
		end
	elseif mo.state == S_PLAY_MELEE or mo.state == S_PLAY_MELEE_FINISH
	or mo.state == S_PLAY_MELEE_LANDING
		MarioDoStars(mobj)
		MarioDoHit(mobj, true)
		S_StartSound(mobj, sfx_mrhhit)
		
		if not P_IsObjectOnGround(mo)
			DoMarioHammerHit(player)
			
			local momz = mo.momz
			player.pflags = $ &~ PF_JUMPED
			P_DoJump(player)
			player.pflags = $ &~ PF_STARTJUMP &~ PF_THOKKED
			if momz*P_MobjFlip(mo) > mo.momz*P_MobjFlip(mo)
				mo.momz = momz
			end
		end
	elseif not player.marioshoe and not (player.pflags & PF_SPINNING)
	and not dmgtype and dmg <= 1 and mobj.info.spawnhealth > 1
	and (player.powers[pw_shield] == SH_GOLDFLOWER
	or player.powers[pw_shield] == SH_METAL)
		local part = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_POPPARTICLEMAR)
		P_TeleportMove(part, mobj.x/2 + mo.x/2, mobj.y/2 + mo.y/2, mobj.z/2 + mo.z/2)
		part.state = S_MARIO_HIT2
		part.color = SKINCOLOR_SUPERGOLD5
		
		mo.state = S_PLAY_SKID
		mo.tics = TICRATE/4
		S_StartSound(mobj, sfx_mrmeht)
	end
end)

//Ground Pound spike destroyer
local GroundPoundSpike = function(mobj, mo)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if (not player or MarioCantMove(player))
	and mo.type != MT_MARIO_CAPE and mo.type != MT_MARIO_FIREBALL then return end
	
	if not (mobj and mobj.valid
	and mobj.z - mobj.scale <= mo.z + mo.height + mo.momz
	and mobj.z + mobj.height + mobj.scale >= mo.z + mo.momz)
		return
	end
	
	if not player
		P_KillMobj(mobj, mo, mo.target or mo)
		return true
	
	elseif player.mariogroundpound == GROUNDPOUNDSTART
		P_KillMobj(mobj, mo, mo)
		mo.momz = 0
		
		DoMarioGroundPoundHit(player)
		return true
		
	elseif (not P_IsObjectOnGround(mo) and (mo.state == S_PLAY_MELEE
	or mo.state == S_PLAY_MELEE_FINISH)) or mo.state == S_PLAY_MELEE_LANDING
	or (player.powers[pw_shield] == SH_BALLOON and player.pflags & PF_SPINNING)
	or (player.powers[pw_shield] == SH_CAPE and player.mariospinjump)
	or mo.state == CAPESPINSTATE or player.mariosize > 0 or player.powers[pw_super]
		P_KillMobj(mobj, mo, mo)
		return true
	end
end
addHook("MobjCollide", GroundPoundSpike, MT_SPIKE)
addHook("MobjCollide", GroundPoundSpike, MT_WALLSPIKE)

//Custom jumping function for Mario's Jumps
addHook("JumpSpecial", function(player)
	local mo = player.mo
	if MarioCantMove(player)
	or not IsMario(mo)
	or (srb2p and not player.normalspeed)
	or player.pflags & PF_JUMPSTASIS
	or player.pflags & PF_JUMPDOWN
	or (player.powers[pw_shield] == SH_BALLOON and player.pflags & PF_THOKKED)
	return end
	
	if ((P_IsObjectOnGround(mo) or mo.eflags&MFE_JUSTHITFLOOR)
	and not (player.pflags&PF_JUMPED))
	or player.mariowallslide or player.mariolongbonk or player.climbing
	or player.powers[pw_shield] == SH_BALLOON
	or (player.powers[pw_super] and not (player.pflags & PF_THOKKED))
	or ((mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_GOOWATER) and not player.marioshoe
	and not (player.pflags & PF_SPINNING) and (not player.glidetime or player.climbing)))
		player.pflags = $ &~ PF_SPINNING
		
		if S_SoundPlaying(mo, sfx_mrrol2)
			S_StopSoundByID(mo, sfx_mrrol2)
		end
		
		local wasinlongjump = player.mariolongjump
		if wasinlongjump
			player.normalspeed = skins[player.skin].normalspeed
			player.mariolongjump = 0
		end
		
		local climbing = player.climbing
		if player.climbing
			player.climbing = 0
			player.charability = skins[player.skin].ability
			player.camerascale = skins[player.skin].camerascale
			player.mariotriple = 0
			
			//We're gonna do what's called a pro gamer move
			player.pflags = $ &~ PF_JUMPED
			mo.state = PULLUPSTATE
			player.panim = PA_DASH
			player.powers[pw_nocontrol] = 8
			
			local ghost = P_SpawnGhostMobj(mo)
			mo.z = $ + mo.height*P_MobjFlip(mo)/2
			P_TryMove(mo, mo.x + P_ReturnThrustX(nil, mo.angle, mo.radius),
			mo.y + P_ReturnThrustY(nil, mo.angle, mo.radius), true)
			
			S_StartSound(mo, sfx_mrlout)
			MarioPlayVoice(player, "pull")
			return true
		end
		
		if player.mariocrouch
		and not (player.mariopressspin or player.mariopressdown
		or mo.ceilingz - mo.floorz < P_GetPlayerHeight(player))
			player.mariocrouch = 0
			player.mariocharge = 0
			player.marioslidesfx = 0
		end
		
		//Get rid of slope physics
		if mo.standingslope
			local oldflags = mo.flags
			local oldmomz = mo.momz
			
			mo.flags = $ | MF_NOCLIPHEIGHT
			mo.momz = 0
			
			P_ZMovement(mo)
			
			mo.flags = oldflags
			mo.momz = oldmomz
		end
			
		//Set the triple to the next one in the list
		if player.mariotriple <= 1
		or player.speed >= mo.scale*RUNSPEED*2/3 and player.mariotriple < 3
			player.mariotriple = $+1
		else
			player.mariotriple = 1
		end
		
		//We need this for the side flip check...how painful
		local diff = MarioInputAngle(player)
		if player.skidtime
			if player.pflags * PF_DIRECTIONCHAR
				diff = player.drawangle - $
			else
				local pmx = player.rmomx + player.cmomx
				local pmy = player.rmomy + player.cmomy
				diff = R_PointToAngle2(0,0, pmx, pmy) - $
			end
			diff = InvAngle(InvAngle($))
		end
		
		//Higher jump
		local jumpfactor = player.jumpfactor
		if player.mariosize < 0
		and not (player.mariowallslide or player.mariolongbonk)
			player.jumpfactor = $*5/4
		end
		
		if player.mariowaterswitch
		and not (player.mariowallslide or player.mariolongbonk)
			player.jumpfactor = FixedDiv($, FixedDiv(117*FRACUNIT, 200*FRACUNIT))
		elseif player.powers[pw_shield] == SH_GOLDFLOWER
		or player.powers[pw_shield] == SH_METAL
			player.jumpfactor = $*9/8
		end
		
		//Balloon jump
		if player.powers[pw_shield] == SH_BALLOON
			player.pflags = $ &~ PF_JUMPED
			P_DoJump(player, false)
			player.pflags = $ | PF_THOKKED
			
			MarioCircle(mo, MT_MARIO_DUST, 8, 12*mo.scale,
			-8*FRACUNIT, mo.scale/2, 0, nil, false, player.drawangle)
			
			S_StartSound(mo, sfx_mrpbfl)
			player.mariosquish = -SQUISHTIME*2
		
		//Wall jump
		elseif player.mariowallslide or player.mariolongbonk
			if player.mariolongbonk
				local line = lines[player.lastlinehit]
				local side = sides[player.lastsidehit]
				if line and line.valid and side and side.valid
					local pointx, pointy = P_ClosestPointOnLine(mo.x, mo.y, line)
					player.drawangle = R_PointToAngle2(pointx, pointy, mo.x, mo.y)
				else
					player.drawangle = $ - ANGLE_180
				end
				
				player.mariobonkland = 0
				player.mariolongbonk = 0
			end
			
			P_InstaThrust(mo, player.drawangle, max(WALLJUMPTHRUST*mo.scale, player.speed*3/2))
			
			player.pflags = $ &~ PF_JUMPED
			player.jumpfactor = $*9/8
			P_DoJump(player, false)
			
			MarioCircle(mo, MT_MINECARTSPARK, 6, 4*mo.scale, FixedDiv(mo.height, mo.scale)*2/3,
			mo.scale/4, 2, 0, true, player.drawangle)
			
			player.mariowallslide = 0
			player.marioslidesfx = 0
			player.mariojumpdir = player.drawangle
			player.mariowalljump = 75*player.jumpfactor/FRACUNIT
			player.mariowallangle = player.drawangle
			
			mo.state = WALLJUMPSTATE
			mo.tics = -1
			S_StopSoundByID(mo, sfx_mrsld1)
			S_StopSoundByID(mo, sfx_mrsld2)
			S_StartSound(mo, sfx_mrwljm)
			
			MarioPlayVoice(player, "jump")
			
		//Swimming / Super jump
		elseif not (P_IsObjectOnGround(mo) or mo.eflags&MFE_JUSTHITFLOOR)
		and not climbing
			if player.powers[pw_shield] == SH_GOLDFLOWER
			or player.powers[pw_shield] == SH_METAL
				if mo.momz * P_MobjFlip(mo) < mo.scale/2
					mo.momz = mo.scale/2 * P_MobjFlip(mo)
				end
			else
				player.pflags = $ &~ PF_JUMPED
				P_DoJump(player, false)
			end
				
			//Super jump
			if player.powers[pw_super] and not (player.pflags & PF_THOKKED)
				mo.momz = $*3/2
				player.pflags = $ | PF_THOKKED &~ PF_NOJUMPDAMAGE
				
				MarioCircle(mo, MT_MARIO_DUST, 8, 12*mo.scale,
				-8*FRACUNIT, mo.scale/2, 0, nil, false, player.drawangle)
				
				S_StartSound(mo, sfx_mrpbfl)
				mo.state = ROLLSTATE
				
			//Swimming
			else
				mo.state = SWIMSTATE
				player.panim = PA_ABILITY2
				
				//Reset enemy combo counter
				if not player.powers[pw_invulnerability]
					player.scoreadd = 0
				end
				
				if leveltime % 2
					S_StartSound(mo, sfx_mrswm1)
				else
					S_StartSound(mo, sfx_mrswm2)
				end
			end
			
		//Spin jump
		elseif (player.mariospinjump or player.mariodidcircle)
		and not (mo.eflags&MFE_JUSTHITFLOOR) and not player.mariocrouch
			if player.mariodidcircle
				player.mariospinjump = player.mariodidcircle * GROUNDSPINTIME
			else
				player.mariospinjump = min(max($*GROUNDSPINTIME, -GROUNDSPINTIME), GROUNDSPINTIME)
			end
			mo.state = SPINSTATE
			
			player.mariospinsfx = -1
			player.jumpfactor = jumpfactor*13/12
			P_DoJump(player, false)
			player.pflags = $ | PF_THOKKED
			
			MarioPlayVoice(player, "spinjump")
			
		//Dive Recovery
		elseif mo.state == DIVESTATE
			//Landing cut
			if mo.eflags & MFE_JUSTHITFLOOR and not player.powers[pw_justsprung]
				mo.momx = $*2/3
				mo.momy = $*2/3
			end
			local verticalthrust = P_GetMobjGravity(mo)
			if mo.eflags & MFE_GOOWATER then verticalthrust = -gravity*P_MobjFlip(mo) end
			if abs(verticalthrust) > mo.scale/2 then verticalthrust = mo.scale/-2*P_MobjFlip(mo) end
			mo.momz = -8 * verticalthrust
			P_Thrust(mo, MarioInputAngle(player), 12*mo.scale)
			
			player.glidetime = GLIDETIMETIME //This says we can't dive again!
			player.mariotriple = 0
			
			mo.state = DIVERECOVERYSTATE
			S_StartSound(mo, sfx_mrlout)
			
			MarioPlayVoice(player, "recovery")
			
		//Ground Pound Jump
		elseif player.mariogroundpound
		and player.mariogroundpound < GROUNDPOUNDEND
			player.jumpfactor = $*11/8
			P_DoJump(player, false)
			
			mo.state = S_PLAY_JUMP
			if not player.mariosize
				if player.powers[pw_shield] != SH_SUPERBALL
					S_StartSound(mo, sfx_mrgpjm)
				else
					S_StartSound(mo, sfx_mrbal4)
				end
			elseif player.mariosize > 0
				S_StartSound(mo, sfx_mrmgjm)
			else
				S_StartSound(mo, sfx_mrmin4)
			end
			
			player.mariogroundpound = 0
			player.mariogroundpoundtime = 0
			player.mariogpjump = 16
			
			MarioPlayVoice(player, "triple")
			
		//Backflip
		elseif player.mariocrouch and player.mariocharge >= CHARGETIME
		and mo.ceilingz - mo.floorz >= P_GetPlayerHeight(player)
			player.jumpfactor = $*7/4
			P_DoJump(player, false)
			player.pflags = $ &~ PF_STARTJUMP &~ PF_NOJUMPDAMAGE &~ PF_STARTJUMP
			
			P_Thrust(mo, player.drawangle, -BACKFLIPSPEED*mo.scale)
			player.marioflipmove = 1
			
			player.mariotriple = 0
			player.mariocrouch = 0
			player.mariocharge = 0
			player.marioslidesfx = 0
			
			mo.state = BACKFLIPSTATE
			if not player.mariosize
				if player.powers[pw_shield] != SH_SUPERBALL
					S_StartSound(mo, sfx_mrgpjm)
				else
					S_StartSound(mo, sfx_mrbal4)
				end
			elseif player.mariosize > 0
				S_StartSound(mo, sfx_mrmgjm)
			else
				S_StartSound(mo, sfx_mrmin4)
			end
			
			MarioPlayVoice(player, "jump")
		
		//Long Jump
		elseif player.mariopressspin
		and (wasinlongjump or (player.mariocrouch < TICRATE/2
		and player.speed >= RUNSPEED*mo.scale/4 and not player.marioshoe))
			player.jumpfactor = $*4/5
			P_DoJump(player, false)
			player.pflags = $ &~ PF_STARTJUMP
			
			local speed = FixedHypot(mo.momx, mo.momy)
			local angle = R_PointToAngle2(0,0, mo.momx, mo.momy)
			if not wasinlongjump
				player.mariojumpdir = angle
			elseif player.marioblj and speed < player.marioblj
				speed = player.marioblj
				P_InstaThrust(mo, angle, speed)
			end
			
			player.mariolongjump = 1
			player.mariocrouch = 0
			player.mariocharge = 0
			player.marioslidesfx = 0
			
			//The intial thrust!
			//This is supposed to multiply your forward momentum ...
			local diff = angle - player.mariojumpdir
			diff = abs(InvAngle(InvAngle($)))
			if player.mariolongjump == 1
				if diff < ANGLE_90
					P_Thrust(mo, player.mariojumpdir, speed/2)
				else
					P_Thrust(mo, player.mariojumpdir, -speed/2)
				end
			end
			
			mo.state = LONGJUMPSTATE
			
			S_StartSound(mo, sfx_mrlong)
			MarioPlayVoice(player, "long")
			
		//Side flip
		elseif abs(diff) > ANGLE_22h and player.skidtime
		and (player.cmd.forwardmove or player.cmd.sidemove)
			player.jumpfactor = $*4/3
			P_DoJump(player, false)
			player.pflags = $ &~ PF_STARTJUMP
			
			player.mariotriple = 1
			
			player.drawangle = MarioInputAngle(player) - ANGLE_180
			player.mariojumpdir = player.drawangle
			P_InstaThrust(mo, player.drawangle, -SIDEFLIPSPEED*mo.scale)
			player.marioflipmove = 1
			
			if player.skidtime
				player.skidtime = 0
				player.marioskid = nil
				S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
			end
			
			mo.state = SIDEFLIPSTATE
			mo.state = SIDEFLIPSTATE
			mo.state = SIDEFLIPSTATE
			S_StartSound(mo, sfx_mrflip)
			
			MarioPlayVoice(player, "spin")
			
		//Hammer jump
		elseif (player.mariopresscustom == 1 or player.mariopressfire == 1)
		and not player.mariocrouch
			DoHammer(player, true, player.mariopressfire == 1)
			
			//Vertical momentum
			local verticalthrust = P_GetMobjGravity(mo)
			if mo.eflags & MFE_GOOWATER then verticalthrust = -gravity*P_MobjFlip(mo) end
			if abs(verticalthrust) > mo.scale/2 then verticalthrust = mo.scale/-2*P_MobjFlip(mo) end
			mo.z = $ + P_MobjFlip(mo)
			mo.momz = -6 * verticalthrust
			
			player.mariotriple = 1
			
		//Normal Jump
		elseif player.mariotriple <= 1 or player.mariocrouch
			if not player.mariosize and player.powers[pw_shield] != SH_SUPERBALL
				P_DoJump(player, true)
			else
				P_DoJump(player, false)
				mo.state = S_PLAY_JUMP
				if not player.mariosize
					S_StartSound(mo, sfx_mrbal1)
				elseif player.mariosize > 0
					S_StartSound(mo, sfx_mrmgjm)
				else
					S_StartSound(mo, sfx_mrmin1)
				end
			end
			
			if player.mariopspeed >= PSPEEDTIME
				player.mariocapeflight = 1
			elseif player.mariocapeflight
				player.mariocapeflight = 0
			end
			
			MarioPlayVoice(player, "jump")
			
		//Double Jump
		elseif player.mariotriple == 2
			player.jumpfactor = $*5/4
			P_DoJump(player, false)
			
			mo.state = DOUBLEJUMPSTATE
			
			if not player.mariosize
				if player.powers[pw_shield] != SH_SUPERBALL
					S_StartSound(mo, sfx_mrjmp2)
				else
					S_StartSound(mo, sfx_mrbal2)
				end
			elseif player.mariosize > 0
				S_StartSound(mo, sfx_mrmgjm)
			else
				S_StartSound(mo, sfx_mrmin2)
			end
			
			if player.mariopspeed >= PSPEEDTIME
				player.mariocapeflight = 1
			elseif player.mariocapeflight
				player.mariocapeflight = 0
			end
			
			MarioPlayVoice(player, "double")
			
		//Triple Jump
		elseif player.mariotriple >= 3
			player.jumpfactor = $*7/4
			P_DoJump(player, false)
			player.pflags = $ &~ PF_STARTJUMP
			
			if player.powers[pw_shield] == SH_CAPE
				mo.state = DOUBLEJUMPSTATE
				player.mariopspeed = PSPEEDTIME
				player.mariocapeflight = 1
			else
				mo.state = TRIPLEJUMPSTATE
				player.pflags = $ &~ PF_NOJUMPDAMAGE
			end
			
			if not player.mariosize
				if player.powers[pw_shield] != SH_SUPERBALL
					S_StartSound(mo, sfx_mrjmp3)
				else
					S_StartSound(mo, sfx_mrbal3)
				end
			elseif player.mariosize > 0
				S_StartSound(mo, sfx_mrmgjm)
			else
				S_StartSound(mo, sfx_mrmin3)
			end
			
			MarioPlayVoice(player, "triple")
		end
		
		player.jumpfactor = jumpfactor
		player.mariotripletime = TRIPLETIME
		
		//Oh, guess we didn't actually jump
		if P_IsObjectOnGround(mo)
			player.pflags = $ &~ PF_JUMPED
		end
		return true
	end
end)

//Super transformation with a shield
addHook("JumpSpinSpecial", function(player)
	local mo = player.mo
	if MarioCantMove(player)
	or not IsMario(mo)
	or player.pflags & PF_SPINDOWN
	or not (player.mariocrouch or not (player.mariocrouchbutton == nil
	or player.mariocrouchbutton == BT_SPIN or player.marioattackbutton == BT_SPIN
	or player.marioabilitybutton == BT_SPIN or player.mariospinbutton == BT_SPIN))
	or player.powers[pw_super]
	or player.powers[pw_invulnerability]
	or player.powers[pw_tailsfly]
	or not (player.charflags & SF_SUPER)
	or not (player.pflags & PF_JUMPED)
	or not (player.powers[pw_shield] & SH_NOSTACK)
	or (maptol & TOL_NIGHTS)
	or not All7Emeralds(emeralds)
	or player.rings < 50
	return end
	
	P_DoSuperTransformation(player, false)
	return true
end)

//Make crouching/crawling use your spin height
addHook("PlayerHeight", function(player)
	local mo = player.mo
	if mo and mo.valid and IsMario(mo)
		local height = P_GetPlayerHeight(player)
		if player.climbing
			height = $*2
		elseif player.mariocrouch or player.mariogroundpound
			height = P_GetPlayerSpinHeight(player)
			
		//Go into crouch, not spin! Well, except when in a shoe
		elseif mo.ceilingz - mo.floorz < mo.height
		and (not P_PlayerShouldUseSpinHeight(player) or player.mariopain)
			ResetMario(player)
			
			if player.marioshoe or player.powers[pw_shield] == SH_BALLOON
				mo.state = S_PLAY_ROLL
				player.pflags = $ | PF_SPINNING
			else
				player.mariocrouch = 1
				player.mariocharge = 0
				
				mo.state = CROUCHSTATE
				mo.tics = 2
				player.pflags = $ &~ PF_SPINNING
			end
			
			if player.speed < mo.radius and not P_TryMove(mo, mo.x, mo.y, true)
				P_InstaThrust(mo, MarioInputAngle(player), mo.radius)
			end
			
			height = P_GetPlayerSpinHeight(player)
		end
		
		if player.mariosize
			if P_PlayerShouldUseSpinHeight(player) and not player.mariopain
				height = P_GetPlayerSpinHeight(player)
			end
			height = $ + FixedMul(player.mariosize, $)
		end
		
		if height != P_GetPlayerHeight(player)
			return height
		end
	end
end)
addHook("PlayerCanEnterSpinGaps", function(player)
	local mo = player.mo
	if mo and mo.valid and IsMario(mo)
		return true
	end
end)

//Boots of Steel
addHook("ShouldDamage", function(mo, mobj, src, dmg, dmgtype)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	local mobjvalid = mobj and mobj.valid
	if srb2p or not player then return end
	if player.mariopain then return false end
	
	//Prevent being crushed while mini
	//This is meant as a bug fix for ground pounding monitors
	//Erm, I mean this is a 100% intended feature!!
	if dmgtype == DMG_CRUSHED and player.powers[pw_shield] == SH_MINI
		return false
	elseif dmgtype & DMG_DEATHMASK
		return
	end
	
	//Don't get Amy's shield when you already have a powerup!
	if mobj and mobj.valid and mobj.type == MT_LHRT
	and player.powers[pw_shield] and not (player.powers[pw_shield] & SH_NOSTACK)
		return false
	end
	
	//Extra elemental protection for Metal
	if player.mariosize > 0 then return false end
	if dmgtype == DMG_ELECTRIC and (player.powers[pw_shield] == SH_METAL
	or player.powers[pw_shield] == SH_GOLDFLOWER) then return false end
	
	//Cape projectile destruction
	if mobjvalid and mobj.flags & MF_MISSILE and mobj.type != MT_MARIO_CAPE
	and not (mobj.target and mobj.target.valid and mobj.target.player)
	and ((player.powers[pw_shield] == SH_CAPE and player.mariospinjump)
	or mo.state == CAPESPINSTATE)
		P_KillMobj(mobj)
		return false
	end
	
	if not (player.mariospinjump or mo.state == SPINSTATE or player.marioshoe)
	or (not mobjvalid and not (player.marioshoe and P_IsObjectOnGround(mo)) and dmgtype != DMG_SPIKE)
	or (mobjvalid and ((not player.marioshoe and P_IsObjectOnGround(mo))
	or mobj.player or mobj.type == MT_WALLSPIKE
	or (mobj.type == MT_SPIKE and mo.eflags & MFE_VERTICALFLIP != mobj.eflags & MFE_VERTICALFLIP)))
		return
	end
	
	if mobjvalid and not (mobj.flags & MF_SOLID)
		if P_IsObjectOnGround(mo) or not P_PlayerCanDamage(player, mobj)
			return
		elseif mo.eflags & MFE_VERTICALFLIP
			mo.flags = $ | MF_NOCLIPTHING
			mo.z = mobj.z - mo.height
			mo.flags = $ &~ MF_NOCLIPTHING
		else
			mo.flags = $ | MF_NOCLIPTHING
			mo.z = mobj.z + mobj.height
			mo.flags = $ &~ MF_NOCLIPTHING
		end
	end
	
	player.mariogroundpound = 0
	player.mariogroundpoundtime = 0
	
	if not player.marioshoe
		player.mariospinjump = $ or 1
		mo.state = SPINSTATE
	elseif not mobjvalid or (mobj.flags & MF_SOLID and not (mobj.flags & MF_PAIN))
		return false
	end
	player.pflags = $ &~ PF_JUMPED &~ PF_NOJUMPDAMAGE
	P_DoJump(player, false)
	
	local part = P_SpawnMobjFromMobj(mo, 0, 0, -16*FRACUNIT, MT_POPPARTICLEMAR)
	part.destscale = 1
	part.scalespeed = FixedMul($, part.scale)
	S_StartSound(mo, sfx_mrstmp)
	
	return false
end, MT_PLAYER)

//Wall Collision
addHook("MobjMoveBlocked", function(mo, mobj, collideline)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if not player or MarioCantMove(player) or player.mariocapeflight
	or (player.glidetime and P_IsObjectInGoop(mo)) return end
	
	//Balloon bounce
	if player.powers[pw_shield] == SH_BALLOON
		if player.speed > 4*mo.scale	
			P_BounceMove(mo)
			player.pflags = $ &~ PF_THOKKED
			
			S_StartSound(mo, sfx_mrpbmp)
			player.mariosquish = -SQUISHTIME
		end
		
	//Ignore walls during a blj
	elseif player.mariolongjump and player.marioblj
		//Allow teleporting past walls
		local x = mo.x + mo.momx
		local y = mo.y + mo.momy
		if player.speed > ANGLE_11hh
		and P_CheckPosition(mo, x, y)
		and P_CeilingzAtPos(x, y, mo.z, mo.height)
		- P_FloorzAtPos(x, y, mo.z, mo.height)
		>= mo.height
			//Only teleport if this sector has sectors attached to it
			local subsector = R_PointInSubsectorOrNil(x, y)
			local sector = subsector and subsector.sector
			if sector
				for i = 0, #sector.lines
					local line = sector.lines[i]
					if not (line and line.valid) then continue end
					if (line.frontsector and line.frontsector != sector)
					or (line.backsector and line.backsector != sector)
						//Teleport with noclip to prevent crushing
						local oldflag = mo.flags
						mo.flags = $ | MF_NOCLIP &~ MF_NOCLIPHEIGHT
						P_TeleportMove(mo, x, y, mo.z)
						mo.flags = $ &~ MF_NOCLIP
					
						break
					end
				end
			end
		end
		
		return true
	
	//Diving bonk
	elseif mo.state == DIVESTATE or player.mariolongjump
		local oldmomx = mo.momx
		local oldmomy = mo.momy
		P_BounceMove(mo)
		player.mariobouncedir = R_PointToAngle2(0,0, mo.momx, mo.momy)
		player.mariobouncespeed = player.speed
		mo.momx = oldmomx
		mo.momy = oldmomy
	
	//Lava bounce
	elseif mo.state == LAVASTATE and player.powers[pw_flashing]
		P_BounceMove(mo)
		P_InstaThrust(mo, R_PointToAngle2(0,0,mo.momx,mo.momy), 4*mo.scale)
		
	//Wall Sliding and Ledge Grabbing
	elseif not P_IsObjectOnGround(mo) and not player.mariocrouch
	and (player.cmd.forwardmove or player.cmd.sidemove) and not player.climbing
	and ( (player.pflags&PF_STARTJUMP and mo.momz*P_MobjFlip(mo) <= 0)
	or (not (player.pflags&PF_STARTJUMP) and mo.momz*P_MobjFlip(mo) <= -4*P_GetMobjGravity(mo))
	or player.mariowalljump)
	and not (player.pflags & (PF_JUMPSTASIS|PF_SPINNING))
		local prevline = player.lastlinehit
		local prevside = player.lastsidehit
		
		player.lastlinehit = -1
		player.lastsidehit = -1
		
		P_SlideMove(mo)
		player.powers[pw_pushing] = 3
		
		local line = lines[player.lastlinehit]
		if not line and collideline
			line = collideline
			player.lastlinehit = #line
			player.lastsidehit = line.sidenum[P_PointOnLineSide(mo.x, mo.y, line)]
		end
		local side = sides[player.lastsidehit]
		if line and line.valid and side and side.valid
			local pointx, pointy = P_ClosestPointOnLine(mo.x, mo.y, line)
			local angle = R_PointToAngle2(mo.x, mo.y, pointx, pointy)
			
			local diff = MarioInputAngle(player) - angle
			diff = InvAngle(InvAngle($))
			
			local diff2 = player.mariowallangle - angle
			diff2 = InvAngle(InvAngle($))
			
			if abs(diff) < ANGLE_90
			and not (line and line.backsector
			and line.backsector.ceilingpic == "F_SKY1"
			and line.frontsector
			and line.frontsector.ceilingpic == "F_SKY1"
			and (mo.z >= line.frontsector.ceilingheight
			or mo.z >= line.backsector.ceilingheight))
			and line.special != 41
				//Can we ledge grab?
				local grabbable = false
				if not player.mariowallslide
					//Get the side we're NOT on
					local grabside = side
					if line.frontside == side
						grabside = line.backside
					else
						grabside = line.frontside
					end
					
					//Get the sector we're grabbing onto...
					local sector = grabside and grabside.valid and grabside.sector
					if sector and sector.valid
					and sector.ceilingheight - sector.floorheight >= mo.height
					//and not (GetSecSpecial(sector.special, 1) and GetSecSpecial(sector.special, 1) <= 8)
						local thok = P_SpawnMobjFromMobj(mo,0,0,FixedDiv(mo.height, mo.scale)/2,MT_THOK)
						if thok and thok.valid
							thok.flags = $ &~ MF_NOCLIP &~ MF_NOCLIPTHING &~ MF_NOCLIPHEIGHT
							thok.tics = 1
							P_TeleportMove(thok, thok.x + P_ReturnThrustX(nil, angle, mo.scale),
							thok.y + P_ReturnThrustY(nil, angle, mo.scale), thok.z)
							
							local floor = thok.floorz
							local ceiling = thok.ceilingz
							
							//Are we in grab range?
							if ceiling - floor < mo.height
								grabbable = false
							elseif mo.eflags & MFE_VERTICALFLIP
								if mo.z - mo.height/2 < ceiling
								and mo.z + mo.height > ceiling
								and mo.z + mo.height < mo.ceilingz - mo.height*2/3
								and mo.z - mo.height > mo.floorz
								and not sector.c_slope
									mo.z = ceiling
									grabbable = true
								end
							elseif mo.z < floor
							and mo.z + mo.height*3/2 > floor
							and mo.z > mo.floorz + mo.height*2/3
							and mo.z + mo.height*2 < mo.ceilingz
							and not sector.f_slope
								mo.z = floor - mo.height
								grabbable = true
							end
							
							//Set mo.angle without changing the player's camera
							if grabbable
								local target = mo.target
								mo.target = thok
								A_FaceTarget(mo)
								mo.target = target
							end
						end
					end
				end
				
				//Refresh
				player.glidetime = 0
				player.mariogroundpound = 0
				player.mariogroundpoundtime = 0
				if player.powers[pw_super] then player.pflags = $ &~ PF_THOKKED end
				
				//Ledge Grab
				if grabbable
					player.drawangle = angle
					player.climbing = 5
					
					mo.momx = 0
					mo.momy = 0
					mo.momz = 0
					
					player.charability = CA_GLIDEANDCLIMB
					player.camerascale = skins[player.skin].camerascale/2
					player.pflags = $|PF_FULLSTASIS
					
					if player.powers[pw_shield] != SH_SUPERBALL and mariovoice.value
						MarioPlayVoice(player, "grab")
					else
						S_StartSound(mo, sfx_mrcwl1)
					end
					
				//Wall Slide
				elseif not player.mariowalljump or abs(diff2) < ANGLE_67h
					player.drawangle = angle - ANGLE_180
					player.mariowallslide = 3
					
					player.mariospinjump = 0
					player.mariospinsfx = 0

					if player.mariolastside == nil
						player.mariolastside = side
						player.mariosidexscroll = side.textureoffset
						player.mariosideyscroll = side.rowoffset
					end
				end
			end
		else
			player.lastlinehit = prevline
			player.lastsidehit = prevside
		end
		return true
	end
end, MT_PLAYER)

//I guess since vanilla does this for drowning, we can do it for the other damage types too?
addHook("MobjDeath", function(mo, inf, src, dmgtype)
	if mo and mo.valid and IsMario(mo) then mo.movedir = dmgtype end
end, MT_PLAYER)

//No getting bubbles with infinite air
addHook("TouchSpecial", function(mobj, mo)
	local player = mo and mo.valid and IsMario(mo) and mo.player
	if player and (player.mariosize > 0 or player.powers[pw_shield] == SH_BALLOON
	or player.powers[pw_shield] == SH_SUPERBALL or player.powers[pw_super])
		return true
	end
end, MT_EXTRALARGEBUBBLE)

//Restore things from before pain
rawset(_G, "RestoreMarioPain", function(player, mo, state, flags, nullout)
	if not player.mariopain then return end
	mo.momx = player.mariopain.momx
	mo.momy = player.mariopain.momy
	mo.momz = player.mariopain.momz
	
	if player.mariopain.state == S_PLAY_SKID
		mo.state = S_PLAY_WALK
	elseif state
		mo.state = player.mariopain.state
		mo.frame = player.mariopain.frame
		mo.sprite2 = player.mariopain.sprite2
		mo.tics = player.mariopain.tics
	end
	
	if flags
		mo.flags = player.mariopain.flags
		player.mariogroundpound = player.mariopain.groundpound
		player.mariospinjump = player.mariopain.spinjump
		player.mariospinsfx = player.mariopain.spinsfx
		player.mariogpjump = player.mariopain.gpjump
		player.mariocrouch = player.mariopain.crouch
		player.marioflipmove = player.mariopain.flipmove
		player.mariocapeflight = player.mariopain.capeflight
		
		if not player.mariopain.powerup and not player.mariopain.shield
		and player.powers[pw_shield]
			player.powers[pw_shield] = 0
		end
		
		if player.mariopain.powerup
		and not (player.mariopain.mega and not player.mariomega)
			player.powers[pw_flashing] = 0
		end
		
		local keepflags = PF_FLIPCAM|PF_ANALOGMODE|PF_DIRECTIONCHAR|PF_AUTOBRAKE
		|PF_GODMODE|PF_NOCLIP|PF_ATTACKDOWN|PF_SPINDOWN|PF_JUMPDOWN|PF_WPNDOWN|PF_FINISHED
		player.pflags = ($ & keepflags) | (player.mariopain.pflags &~ keepflags)
		
		SpawnMarioShieldOrb(player)
	end
	
	if nullout then player.mariopain = nil end
end)

//Get inputs
addHook("PreThinkFrame", do for player in players.iterate
	//Not Mario!
	local mo = player.mo
	if not (mo and mo.valid and IsMario(mo))
	or player.mariomenu or mo.luiginumberone
	or (gametyperules & GTR_RACE and leveltime < 4*TICRATE) //Race countdown
	//or (srb2p and not player.normalspeed)
		player.mariopressjump = 0
		player.mariopressspin = 0
		player.mariopressdown = 0
		player.mariopresscustom = 0
		player.mariopressfire = 0
		player.mariopressfirenormal = 0
		player.mariopressdirection = 0
		player.mariocustomspin = 0
		player.mariodidcircle = 0
		player.marioprevinput = {}
		player.marioinputangle = player.drawangle
		continue
	end
	
	//Input angle
	if (player.pflags&PF_DIRECTIONCHAR and player.pflags&PF_ANALOGMODE)
	or player.mo.flags2 & MF2_TWOD or twodlevel
		player.marioinputangle = player.mo.angle
	elseif not (player.cmd.forwardmove or player.cmd.sidemove)
		player.marioinputangle = player.drawangle
	else 
		player.marioinputangle = player.cmd.angleturn<<16 + R_PointToAngle2(0, 0, player.cmd.forwardmove*FRACUNIT, -player.cmd.sidemove*FRACUNIT)
	end
	
	//Detect jump inputs
	if player.mariopressjump == nil
	or not (player.cmd.buttons & BT_JUMP)
		player.mariopressjump = 0
	else
		player.mariopressjump = $ + 1
	end
	
	//Detect spin inputs
	if player.mariopressspin == nil
	//or player.mariocrouchbutton == 0
	or not (player.cmd.buttons & (player.mariocrouchbutton or BT_SPIN))
	or (srb2p and (not player.mariocrouchbutton or player.mariocrouchbutton == BT_SPIN))
		player.mariopressspin = 0
	else
		player.mariopressspin = $ + 1
	end
	if player.climbing then player.cmd.buttons = $ &~ BT_SPIN end
	
	//Detect downward inputs
	if player.mariopressdown == nil or not (mo.flags2 & MF2_TWOD or twodlevel)
	or player.mariospinenabled == 0 or player.cmd.forwardmove > -35
		player.mariopressdown = 0
	else
		player.mariopressdown = $ + 1
		if abs(player.cmd.sidemove) <= 10
			player.cmd.sidemove = 0
		end
	end
	
	//Detect custom 1 inputs
	if player.mariopresscustom == nil
	//or player.marioattackbutton == 0
	or not (player.cmd.buttons & (player.marioattackbutton or BT_CUSTOM1))
	or (srb2p and player.marioattackbutton == BT_SPIN)
		player.mariopresscustom = 0
		if player.mariocustomspin and player.mariocustomspin > 0
			player.mariocustomspin = 0
		end
	else
		player.mariopresscustom = $ + 1
	end
	
	//Detect fire inputs
	if player.mariopressfire == nil
	//or player.marioabilitybutton == 0
	or player.marioattackbutton == player.marioabilitybutton
	or not (player.cmd.buttons & (player.marioabilitybutton or BT_ATTACK))
		player.mariopressfire = 0
	else
		player.mariopressfire = $ + 1
	end
	
	//Detect fire normal inputs
	if player.mariopressfirenormal == nil
	//or player.mariospinbutton == 0
	or not (player.cmd.buttons & (player.mariospinbutton or BT_FIRENORMAL))
		player.mariopressfirenormal = 0
		if player.mariocustomspin and player.mariocustomspin < 0
			player.mariocustomspin = 0
		end
	else
		player.mariopressfirenormal = $ + 1
	end
	
	//Detect direction input
	if ((abs(player.cmd.forwardmove) >= 35 and not (mo.flags2 & MF2_TWOD or twodlevel))
	or abs(player.cmd.sidemove) >= 35)
	and player.mariopressdirection != nil
		player.mariopressdirection = $ + 1
	else
		player.mariopressdirection = 0
	end
	
	//Prevent crouch movement
	if player.mariocrouch and P_IsObjectOnGround(mo)
	and (player.rmomx or player.rmomy)
		player.cmd.forwardmove = 0
		player.cmd.sidemove = 0
		player.powers[pw_noautobrake] = max($, 2)
	end
	
	//Detect circular input
	player.mariodidcircle = 0
	//No input!
	if MarioCantMove(player) or mo.eflags & MFE_SPRUNG
	or player.mariocapeflight or player.glidetime or player.climbing
	or player.pflags & PF_SPINNING or not player.marioprevinput
		player.marioprevinput = {}
		
	//Custom input
	elseif player.mariocustomspin
		player.mariodidcircle = 1
		
	//Not in 2D bud...or with the spin disabled either!
	elseif mo.flags2 & MF2_TWOD or twodlevel
	or player.mariospinenabled == 0
		player.marioprevinput = {}
		
	//The actual turning
	elseif player.mariopressdirection
		local currentangle = R_PointToAngle2(0, 0, player.cmd.forwardmove*FRACUNIT, -player.cmd.sidemove*FRACUNIT)
		local lastangle = currentangle
		local circledir = 0
		local didfourth = 0
		
		for id, angle in pairs(player.marioprevinput)
			local diff = lastangle - angle
			diff = InvAngle(InvAngle($))
			
			lastangle = angle
			
			if not diff then continue end
			
			if circledir
				if (diff > 0 and circledir > 0)
				or (diff < 0 and circledir < 0)
					diff = currentangle - angle
					if abs(diff) >= abs(ANGLE_180)
						diff = currentangle + angle
					end
					
					if didfourth > 1
						if abs(diff) > ANGLE_22h
							player.mariodidcircle = circledir
							break
						end
					elseif didfourth
						if abs(diff) < ANGLE_45
							didfourth = $ + 1
						end
					elseif abs(diff) > ANGLE_135
						didfourth = 1
					end
				else
					break
				end
			else
				if diff > 0
					circledir = 1
				else
					circledir = -1
				end
			end
		end
		
		table.insert(player.marioprevinput, 0, currentangle)
		
		if #player.marioprevinput > SPININPUTTIME
			table.remove(player.marioprevinput)
		end
	elseif #player.marioprevinput
		player.marioprevinput = {}
	end
end end)

//Main thinker
addHook("PlayerThink", function(player)
	local mo = player.mo
	
	//No valid object...
	if not (mo and mo.valid)
		ResetMario(player)
		player.mariopain = nil
		player.marioprevskin = nil
		return
	end
	
	//Winning by doing absolutely nothing
	if mo.luigiwins == nil
		if mo.skin == "luigi" and gametyperules & (GTR_FRIENDLY|GTR_ALLOWEXIT)
		and not srb2p
			mo.luigiwins = 2 + TICRATE*5*60
		else
			mo.luigiwins = 0
		end
	elseif mo.luigiwins
		mo.luigiwins = $ - 1
		if mo.skin != "luigi" or player.cmd.forwardmove or player.cmd.sidemove
		or player.cmd.buttons &~ (BT_CAMLEFT|BT_CAMRIGHT)
		or player.pflags & PF_FINISHED or player.exiting
			mo.luigiwins = 0
		elseif not mo.luigiwins
			P_DoPlayerFinish(player)
			MarioPlayVoice(player, "nothing")
			S_ChangeMusic("LUWINS", true, player)
			mo.luiginumberone = TICRATE*12
		end
	elseif mo.luiginumberone and not player.exiting
		mo.luiginumberone = $ - 1
		player.pflags = $ | PF_FULLSTASIS
		if not mo.luiginumberone and S_MusicName() == "LUWINS"
			P_RestoreMusic(player)
		end
	end
	
	//Star = Sneakers!
	if player.powers[pw_invulnerability] == invulntics
	and (IsMario(mo) or mariomode)
		player.powers[pw_sneakers] = max($,player.powers[pw_invulnerability])
	end
	
	//Mega and mini size changing
	if not mo.initmariomega
		player.mariomega = 0
		player.mariosize = 0
		player.mariomegacombo = 0
		player.marioballoon = 0
		
		mo.initmariomega = true
	end
	
	//Mega timer
	if player.mariomega and mo.health
		player.mariomega = max(min($ - 1, player.powers[pw_sneakers] - TICRATE), 0)
		if not player.mariomega
			S_StartSound(mo, sfx_mrmgls)
			
			local totalcombo = player.mariomegacombo/21
			if totalcombo and not MarioScoreGametype()
				local spawnstate = mobjinfo[MT_1UP_ICON].spawnstate
				local nextstate = spawnstate and states[spawnstate].nextstate
				mobjinfo[MT_MARIO_POWERUP].seestate = nextstate
				mobjinfo[MT_MARIO_POWERUP].attacksound = mobjinfo[MT_1UP_ICON].seesound
				
				mo.angle = player.cmd.angleturn<<16
				local angle = mo.angle - ANGLE_90
				
				for i = 1, totalcombo
					local thrust = (i - 3) * 32*FRACUNIT
					
					P_SpawnMobjFromMobj(mo, P_ReturnThrustX(mo, angle, thrust),
					P_ReturnThrustY(mo, angle, thrust),
					128*FRACUNIT, MT_MARIO_POWERUP).angle = mo.angle
				end
			end
			
			if not player.mariopain
				player.mariopain = {
					momx = mo.momx,
					momy = mo.momy,
					momz = mo.momz,
					state = mo.state,
					frame = mo.frame,
					sprite2 = mo.sprite2,
					tics = mo.tics,
					flags = mo.flags,
					pflags = player.pflags,
					drawangle = player.drawangle,
					flashing = player.powers[pw_flashing],
					shield = player.powers[pw_shield],
					powerup = true,
					mega = true,
					groundpound = player.mariogroundpound,
					spinjump = player.mariospinjump,
					spinsfx = player.mariospinsfx,
					gpjump = player.mariogpjump,
					crouch = player.mariocrouch,
					flipmove = player.marioflipmove,
					capeflight = player.mariocapeflight,
				}
				
				if not player.powers[pw_carry]
					P_ResetPlayer(player)
				end
				
				mo.flags = MF_NOCLIP|MF_NOGRAVITY
				player.pflags = $|PF_FULLSTASIS|PF_INVIS
				mo.momx = 0
				mo.momy = 0
				mo.momz = 0
				if P_IsObjectOnGround(mo) then mo.z = $ + P_MobjFlip(mo) end
				
				player.powers[pw_flashing] = flashingtics
				mo.state = S_PLAY_PAIN
				player.panim = PA_DASH
			end
		end
	end
	
	//Size changing
	if (player.mariomega or player.powers[pw_shield] == SH_MINI
	or player.mariosize) and mo.health
		local revert = not IsMario(mo)
		or (not player.mariomega and player.powers[pw_shield] != SH_MINI)
		local oldsize = player.mariosize
		
		local change = abs(sin(FixedAngle(leveltime*12*FRACUNIT)))/12
		if revert
			if player.mariosize > 0
				change = -$
			else
				change = $/2
			end
		elseif not player.mariomega
			change = -$
			if player.mariosize <= 0
				change = $/2
			end
		end
		player.mariosize = max(min($ + change, FRACUNIT*2), -FRACUNIT/2)
		
		//Prevent growing where you'd get stuck
		local oldheight = P_GetPlayerSpinHeight(player)
		oldheight = $ + FixedMul(oldsize, $)
		local height = P_GetPlayerSpinHeight(player)
		height = $ + FixedMul(player.mariosize, $)
		if oldheight <= mo.ceilingz - mo.floorz
		and height > mo.ceilingz - mo.floorz
			player.mariosize = oldsize
			
		//Properly revert instead of missing the mark
		elseif revert and ((oldsize <= 0 and player.mariosize >= 0)
		or (oldsize >= 0 and player.mariosize <= 0))
			player.mariosize = 0
		end
		
		mo.spritexscale = FRACUNIT + player.mariosize
		mo.spriteyscale = mo.spritexscale
		
		local radius = FixedMul(skins[player.skin].radius, mo.scale)
		if player.mariosize >= 0
			mo.shadowscale = mo.spritexscale
			mo.radius = radius
			P_RadiusAttack(mo, mo, FixedMul(skins[player.skin].radius*2, mo.spritexscale))
		else
			mo.radius = radius + FixedMul(player.mariosize, radius)
			mo.shadowscale = FRACUNIT
		end
		
		if player.mariosize
		and (player.powers[pw_carry] == CR_GENERIC
		or player.powers[pw_carry] == CR_PLAYER
		or player.powers[pw_carry] == CR_ROPEHANG)
			mo.mariohangoffseted = true
			local diff = FixedMul(player.mariosize, player.height*2)
			if player.mariosize > 0
				diff = $/6
			end
			mo.spriteyoffset = FRACUNIT - diff
		elseif mo.mariohangoffseted
			mo.spriteyoffset = 0
			mo.mariohangoffseted = nil
		end
	elseif mo.mariohangoffseted
		mo.spriteyoffset = 0
		mo.mariohangoffseted = nil
	end
	
	//Global junk
	local prevmomz = player.marioprevmomz
	player.marioprevz = mo.z
	player.marioprevmomz = mo.momz
	
	//Reset shoe stuff
	if player.marioshoe and not (player.marioshoe.valid and player.marioshoe.health)
		player.marioshoe = nil
	end
	if player.mariowasinshoe and not player.marioshoe
		if mo.state == S_TAILSOVERLAY_0DEGREES
			mo.state = S_PLAY_FALL
		end
		mo.spriteyoffset = 0
	end
	player.mariowasinshoe = player.marioshoe and true or false
	
	//Say hello!
	if player.marioprevskin != mo.skin
	and not (IsMario(mo) and not (player.followmobj and leveltime > 2))
		if IsMario(mo)
			MarioPlayVoice(player, "hello")
		end
		player.marioprevskin = mo.skin
	end
	
	//You're not Mario!
	if not IsMario(mo)	
		ResetMario(player)
		RestoreMarioPain(player, mo, true, true, true)
		player.mariolives = player.lives
		player.mariokeepshoe = false
		player.mariosuperprevent = nil
		
		//Reset Balloon
		if player.marioballoon
			mo.rollangle = 0
			player.mariosquish = 0
			if not player.mariosize
				mo.spritexscale = FRACUNIT
				mo.spriteyscale = FRACUNIT
			end
			
			if mo.state == BALLOONSTATE
				if player.pflags & PF_SPINNING
					mo.state = S_PLAY_ROLL
				else
					mo.state = S_PLAY_FALL
				end
			end
				
			player.marioballoon = 0
		end
		
		//Reset underwater junk
		if player.mariowaterswitch
			player.normalspeed = skins[player.skin].normalspeed
			player.acceleration = skins[player.skin].acceleration
			player.mariowaterswitch = false
		end
		
		//Restore the proper powerups
		if player.mariowassuper
			player.powers[pw_shield] = player.marioshield
			player.mariowassuper = false
			P_SpawnShieldOrb(player)
		end
		player.marioshield = -1
		if player.powers[pw_shield] == SH_MINI
			player.powers[pw_shield] = SH_FAKEMINI
			P_SpawnShieldOrb(player)
		elseif player.powers[pw_shield] == SH_METAL
			player.powers[pw_shield] = SH_FAKEMETAL
			mo.colorized = false
			P_SpawnShieldOrb(player)
		elseif player.powers[pw_shield] == SH_GOLDFLOWER
			player.powers[pw_shield] = SH_FAKEGOLD
			mo.colorized = false
			P_SpawnShieldOrb(player)
		elseif player.powers[pw_shield] == SH_SUPERBALL and player.wasmario
			mo.colorized = false
			P_SpawnShieldOrb(player)
			P_RestoreMusic(player)
		end
		
		//Revent invincibility colorization
		if player.wasmario and player.wasmario > 1
			mo.colorized = false
		end
		
		player.wasmario = 0
		return
	end
	player.wasmario = player.powers[pw_invulnerability] + 1
	
	//Cancel skidding
	if player.skidtime
		if (mo.state == S_PLAY_SKID or mo.state == S_PLAY_WALK)
		and player.skidtime <= SKIDCANCEL
		and P_IsObjectOnGround(mo) and P_GetPlayerControlDirection(player) == 1
		--and (mo.momx != pmx or mo.momy != pmy)
			player.skidtime = 0
			player.marioskid = nil
			S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
			mo.state = S_PLAY_WALK
		elseif not (mo.state == S_PLAY_SKID or mo.state == S_PLAY_WALK)
			player.skidtime = 0
			player.marioskid = nil
			S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
		end
	end
	
	//Classic.wad support I guess
	mo.classic_spinfix = false
	mo.classic_canshield = false
	mo.classic_statchanges = false
	
	//Our normalspeed
	local normalspeed = FixedMul(player.normalspeed, mo.scale)
	if player.powers[pw_super] or player.powers[pw_sneakers]
		normalspeed = $*2
	end
	
	//Faster walk
	if mo.state == S_PLAY_WALK and P_IsObjectOnGround(mo)
		//At high speeds
		if (mo.frame & FF_FRAMEMASK) % 2
		and player.speed >= RUNSPEED*mo.scale
		and player.mariosize <= 0
			mo.tics = min($, 1)
			//SpawnMarioAfterimage(mo, 2)
			
		//At low speeds
		elseif player.speed < 12*mo.scale and not player.powers[pw_pushing]
		and (abs(player.cmd.forwardmove) > 35 or abs(player.cmd.sidemove) > 35)
			if (mo.frame & FF_FRAMEMASK) % 2 and player.speed < 9*mo.scale
				mo.tics = min($, 1)
				//SpawnMarioAfterimage(mo, 2)
			else
				mo.tics = min($, 2)
			end
			
			/*if player.speed <= 5*mo.scale
				local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
				P_Thrust(dust, FixedAngle(P_RandomRange(360, 0)*FRACUNIT), 3*mo.scale)
				P_Thrust(dust, R_PointToAngle2(0,0, mo.momx, mo.momy), -3*mo.scale)
				dust.scale = mo.scale
			end*/
			
			//Turn faster
			/*if player.pflags&PF_DIRECTIONCHAR
				local diff = (MarioInputAngle(player) - player.drawangle)
				player.drawangle = $ + InvAngle(InvAngle(diff)/2)
			end*/
		end
	end
	
	//Well this is silly
	if mo.state != SCUTTLESTATE 
		if srb2p
			if S_SoundPlaying(player.followmobj or mo, sfx_luscut)
				S_StopSoundByID(player.followmobj or mo, sfx_luscut)
			end
		elseif S_SoundPlaying(player.followmobj or mo, sfx_luscut)
		or S_SoundPlaying(player.followmobj or mo, sfx_l2scut)
		or S_SoundPlaying(player.followmobj or mo, sfx_l3scut)
			S_StopSoundByID(player.followmobj or mo, sfx_luscut)
			S_StopSoundByID(player.followmobj or mo, sfx_l2scut)
			S_StopSoundByID(player.followmobj or mo, sfx_l3scut)
		end
		
		if player.luigiscuttle and mo.momz*P_MobjFlip(mo) >= 0
			player.luigiscuttle = false
		end
	end
	
	//Only turn into super when duck jumping when you have the BT_SPIN mapped
	//Maybe it's stupid, but it prevents accidental transformation
	if P_SuperReady(player) and not player.powers[pw_super]
	and (player.mariocrouchbutton == nil or player.mariocrouchbutton == BT_SPIN
	or player.marioattackbutton == BT_SPIN or player.marioabilitybutton == BT_SPIN
	or player.mariospinbutton == BT_SPIN) and not P_IsObjectOnGround(mo)
	and not player.climbing and not player.powers[pw_carry]
	and not (player.pflags & (PF_GLIDING|PF_SLIDING|PF_SHIELDABILITY))
	and not player.mariocrouch //The important part
		player.pflags = $ | PF_SHIELDABILITY
	end
	
	//Override fireballing with the fire button
	if MARIOFIREPOWERS[player.powers[pw_shield]] or player.powers[pw_super]
		player.weapondelay = max($, 2)
	end
	
	//Fireball delay
	if player.mariofiredelay
		if player.mariofiredelay > 0
			player.mariofiredelay = $ - 1
		else
			player.mariofiredelay = $ + 1
		end
	end
	
	//Metal/Mega Mario Mario stepping
	local shield = player.powers[pw_shield]
	
	local frames = 1
	if player.panim == PA_WALK or player.panim == PA_RUN
		frames = skins[player.skin].sprites[mo.sprite2].numframes or $
	end
	
	if (shield == SH_GOLDFLOWER or shield == SH_METAL or player.mariosize > 0)
	and not player.mariopain and not player.marioshoe
	and ((mo.eflags & MFE_JUSTHITFLOOR and mo.eflags & MFE_ONGROUND and not player.powers[pw_carry])
	or ((P_IsObjectOnGround(mo) or player.powers[pw_carry] == CR_ROLLOUT)
	and (player.panim == PA_WALK or player.panim == PA_RUN))
	and (((mo.frame & FF_FRAMEMASK) + 1) % max(frames/2, 1)) == 0 and mo.tics == 1)
		local thok = P_SpawnMobjFromMobj(mo,0,0,0,MT_OVERLAY)
		thok.target = mo
		thok.state = S_INVISIBLE
		thok.tics = TICRATE
		if shield == SH_GOLDFLOWER or shield == SH_METAL
			S_StartSound(thok, sfx_mrstep)
		end
		if player.mariosize > 0
			S_StartSound(thok, sfx_mrmgst)
			if player == displayplayer or player == secondarydisplayplayer
				P_StartQuake(8*mo.scale, 2)
			end
		end
	end
	
	//Spin effect only when spinning
	if mo.state == DIVESTATE
		//A hack job for landing with the dive
		if P_IsObjectOnGround(mo)
			player.charability2 = CA2_NONE
		else
			player.charability2 = CA2_SPINDASH
			player.cmd.buttons = $ | BT_SPIN
			if srb2p
				mo.P_inputs[BT_SPIN] = 2
			end
		end
		
		//Transition to rolling if we hit a spring (for "scripted" level elements)
		if mo.eflags & MFE_SPRUNG
			MarioCircle(mo, MT_MINECARTSPARK, 12, 8*mo.scale, 0,
			mo.scale/4, 2, 0, true, player.drawangle)
			
			S_StartSound(mo, sfx_zoom)
			mo.state = ROLLSTATE
		end
		
	elseif player.charability2 == CA2_SPINDASH
		player.charability2 = CA2_NONE
	end
	
	//Mini water running/lack of Mega running
	//And also Luigi water running
	if (((player.mariosize < 0
	or (player.mo.skin == "luigi" and ((player.speed >= mo.scale*RUNSPEED and not P_IsObjectOnGround(mo)) or player.charflags & SF_NOSPINDASHDUST)))
	and mo.eflags & MFE_TOUCHWATER) or player.mariosize > 0
	or (player.mariopspeed and player.mariopspeed >= PSPEEDTIME))
	and not player.marioshoe
		player.charflags = $ | SF_NOSPINDASHDUST
		if player.mariosize < 0
			player.charflags = $ | SF_RUNONWATER
			player.runspeed = 16*FRACUNIT
		elseif player.mariopspeed >= PSPEEDTIME
		or (player.mo.skin == "luigi" and (player.powers[pw_shield] & SH_NOSTACK) == SH_CAPE)
			player.runspeed = 16*FRACUNIT
			if player.mo.skin == "luigi"
				player.charflags = $ | SF_RUNONWATER
			else
				player.charflags = $ &~ SF_RUNONWATER
			end
		elseif player.mariosize > 0 
			player.runspeed = INT32_MAX/2
		elseif player.mo.skin == "luigi"
			if not (player.charflags & SF_RUNONWATER) or player.luigirun == nil
				player.charflags = $ | SF_RUNONWATER
				player.runspeed = 4*FRACUNIT
				player.luigirun = TICRATE*2
			end
			if player.mariocrouch or player.mariospinjump
				player.runspeed = INT32_MAX/2
				player.luigirun = 0
			elseif P_IsObjectOnGround(mo) and mo.eflags & MFE_TOUCHWATER
			and not (mo.eflags & MFE_UNDERWATER) and mo.state == S_PLAY_RUN
				if player.luigirun > TICRATE
					mo.momx = $*19/20
					mo.momy = $*19/20
				else
					mo.momx = $*9/10
					mo.momy = $*9/10
				end
				player.luigirun = $ - 1
				if FixedHypot(mo.momx, mo.momy) < 4*mo.scale
				or player.luigirun <= 0
					player.runspeed = INT32_MAX/2
				end
			end
		end
	elseif player.charflags & SF_NOSPINDASHDUST
		player.charflags = $ &~ SF_RUNONWATER &~ SF_NOSPINDASHDUST
		player.runspeed = skins[player.skin].runspeed
		player.luigirun = 0
	end
	
	//Hammer busting
	if mo.state == S_PLAY_MELEE or mo.state == S_PLAY_MELEE_FINISH
	or mo.state == S_PLAY_MELEE_LANDING
	or (player.powers[pw_shield] == SH_BALLOON and player.pflags & PF_SPINNING)
	or (player.powers[pw_shield] == SH_CAPE and player.mariospinjump)
	or mo.state == CAPESPINSTATE or player.mariosize > 0 or player.powers[pw_super]
		player.charflags = $ | SF_CANBUSTWALLS
	elseif player.charflags & SF_CANBUSTWALLS
		player.charflags = $ &~ SF_CANBUSTWALLS
	end
	
	//Ledge grab climbing
	if player.climbing and mo.health
		player.charability = CA_GLIDEANDCLIMB
		player.camerascale = skins[player.skin].camerascale/2
		player.glidetime = max($, 2)
		
		mo.state = LEDGEGRABSTATE
		
		player.pflags = $ | PF_STASIS
		if player.climbing != 1
			player.pflags = $ | PF_JUMPSTASIS
		end
		
		if player.mariopressspin == 1 or player.mariopresscustom == 1
			player.climbing = 0
			mo.state = DOUBLEFALLSTATE
			MarioPlayVoice(player, "bump")
			
			player.charability = skins[player.skin].ability
			player.camerascale = skins[player.skin].camerascale
			
			P_Thrust(mo, mo.angle, -4*mo.scale)
		end
	elseif player.charability == CA_GLIDEANDCLIMB
		player.charability = skins[player.skin].ability
		player.camerascale = skins[player.skin].camerascale
	end
	
	//Slow ledge grab camera
	if (mo.state == PULLUPSTATE or mo.state == PULLUPSTATE2 or player.climbing)
	and camera and (player == displayplayer or player == secondarydisplayplayer)
		local cam = player == displayplayer and camera or camera2
		
		cam.momx = $/2
		cam.momy = $/2
		cam.momz = $/2
	end
	
	//Fall scream
	if P_IsObjectOnGround(mo) or mo.mariofallz == nil or player.mariocapeflight
		mo.mariofallz = mo.z
	elseif type(mo.mariofallz) == "number"
	and mo.z*P_MobjFlip(mo) < mo.mariofallz*P_MobjFlip(mo)
	and abs(mo.z - mo.mariofallz) >= mo.scale*1920
		mo.mariofallz = "screamed"
		MarioPlayVoice(player, "fall")
	end
	
	//Change your thrustfactor
	if player.charflags & SF_NOSKID
		player.charflags = $ &~ SF_NOSKID
	end
	if player.mariocapeflight and mo.state == CAPEFLYSTATE
		player.thrustfactor = 0
	elseif not (player.pflags & PF_SPINNING) and not MarioCantMove(player)
	and P_IsObjectOnGround(mo) and player.skidtime
	and (mo.state == S_PLAY_SKID or mo.state == S_PLAY_WALK)
		if mo.state == S_PLAY_WALK
			player.charflags = $ | SF_NOSKID
		end
		
		if not player.speed
			player.skidtime = 0
			player.marioskid = nil
			S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
			mo.state = S_PLAY_STND
		elseif FixedHypot(mo.momx, mo.momy) < SKIDSPEED*mo.scale
		and P_GetPlayerControlDirection(player) != 1
			mo.momx = $*2/3
			mo.momy = $*2/3
			
			--player.skidtime = 0
			--mo.state = S_PLAY_WALK
		elseif P_GetPlayerControlDirection(player) == 1
			player.thrustfactor = skins[player.skin].thrustfactor
		else
			player.thrustfactor = skins[player.skin].thrustfactor/4
		end
		
	elseif (player.marioflipmove and mo.state == BACKFLIPSTATE)
	or (P_GetPlayerControlDirection(player) != 1 and player.mariowalljump)
		player.thrustfactor = skins[player.skin].thrustfactor*2/3
		
	elseif player.mariospinjump and player.luigispin
		player.thrustfactor = skins[player.skin].thrustfactor*2
		
	elseif P_IsObjectOnGround(mo)
	or (player.mariospinjump and not player.powers[pw_super] and not player.luigispin)
	or (player.pflags & PF_SPINNING and mo.state != DIVESTATE)
	or mo.state == LAVASTATE or player.marioflipmove
		player.thrustfactor = skins[player.skin].thrustfactor
		
	elseif mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_GOOWATER)
	and not player.mariowaterswitch
		player.thrustfactor = skins[player.skin].thrustfactor*2
		
	else
		player.thrustfactor = skins[player.skin].thrustfactor*3/2
	end

	//Higher speed shoe accel
	if (player.powers[pw_sneakers] or player.powers[pw_super])
	and player.speed >= RUNSPEED*mo.scale/3
		player.thrustfactor = $*3/2
	end

	//Slower starting accel
	if FixedHypot(player.rmomx + player.cmomx, player.rmomy + player.cmomy) < SKIDSPEED*mo.scale
	and P_GetPlayerControlDirection(player) == 1
		player.thrustfactor = mo.skin == "luigi" and $*4/5 or $*2/3
	end
	
	//Luigi slipperiness (and bonk slide)
	if ((mo.skin == "luigi" and P_IsObjectOnGround(mo) and player.speed
	and not (player.pflags & PF_SPINNING) and not player.mariocrouch)
	or player.mariobonkland)
	and mo.friction < FRACUNIT 
		mo.friction = ($ - FRACUNIT)*2/3 + FRACUNIT
	end
	
	//Healing coins
	if player.mariohealth and player.mariocoins != nil
		if player.prevmariocoins == nil
			//Nothing lol
		elseif player.mariohealth >= 3
			player.prevmariocoins = 0
			player.mariocoins = 0
		elseif player.rings > player.prevmariocoins
			player.mariocoins = $ + (player.rings - player.prevmariocoins)
			
			if player.mariocoins > 3
				player.mariohealth = $ + 1
				player.mariocoins = 0
				S_StartSound(mo, sfx_mrheal)
			end
		end
	end
	player.prevmariocoins = player.rings
	
	//Don't let the Super Ball invinciblity theme cut out
	if player == consoleplayer and player.powers[pw_shield] == SH_SUPERBALL
	and player.powers[pw_invulnerability] and S_GetMusicPosition() >= 16*MUSICRATE
		P_PlayJingle(player, mariomode and JT_MINV or JT_INV)
	end
	
	//Make the Super Ball 1up theme shorter
	if player.powers[pw_shield] == SH_SUPERBALL and player.powers[pw_extralife] > 1
	and player.powers[pw_extralife] <= extralifetics - TICRATE*5/4
		player.powers[pw_extralife] = 1
	end
	
	if player.powers[pw_carry] != CR_NIGHTSMODE
	and not player.powers[pw_nights_linkfreeze] and leveltime > 2
		if player.mariopowerupstay == nil
			player.mariopowerupstay = 2
		elseif player.mariopowerupstay
			player.mariopowerupstay = $ - 1
		end
	elseif player.mariopowerupstay
		player.mariopowerupstay = nil
	end
	
	//Powerup stuff
	if player.powers[pw_super] and not G_IsSpecialStage()
	and mo.state != S_PLAY_SUPER_TRANS1 and mo.state != S_PLAY_SUPER_TRANS2
		if player.marioshield == -1
			player.marioshield = player.powers[pw_shield]
		end
		player.mariowassuper = true
		player.powers[pw_shield] = SH_CAPE
	else
		if not G_IsSpecialStage() and (player.mariowassuper
		or not (player.exiting or player.mariopowerupstay))
			if player.marioshield != player.powers[pw_shield]
				if player.marioshield == SH_SUPERBALL
				or (player.powers[pw_shield] & SH_NOSTACK) == SH_SUPERBALL
					if (leveltime <= 2 and (not multiplayer or gametyperules&GTR_FRIENDLY))
					or (player.powers[pw_shield] & SH_NOSTACK) == SH_SUPERBALL
						S_ChangeMusic("mland1", true, player)
					else
						P_RestoreMusic(player)
					end
				end
				
				//Carry over from being super
				if player.mariowassuper
					player.powers[pw_shield] = player.marioshield
					SpawnMarioShieldOrb(player)
					player.mariowassuper = false
					
					if player.marioshield == SH_SUPERBALL
						S_ChangeMusic("mland1", true, player)
					end
					
				//Transfer from another skin properly
				elseif player.marioshield == -1
					SpawnMarioShieldOrb(player)
				
				//Carry over from the previous level
				elseif player.marioshield != nil and not player.powers[pw_shield]
				and leveltime <= 2 //and player.mariolives == player.lives
				and (not multiplayer or gametyperules&GTR_FRIENDLY)
					player.powers[pw_shield] = player.marioshield
					SpawnMarioShieldOrb(player)
					
					if player.marioshield == SH_MINI
						player.mariosize = -FRACUNIT/2
					end
					
				//Get stun
				elseif not MarioCantMove(player) and player.powers[pw_shield]
				and not player.powers[pw_super] and not player.mariopain
					if not player.shieldswap_cooldown
						player.mariopain = {
							momx = mo.momx,
							momy = mo.momy,
							momz = mo.momz,
							state = mo.state,
							frame = mo.frame,
							sprite2 = mo.sprite2,
							tics = mo.tics,
							flags = mo.flags,
							pflags = player.pflags,
							drawangle = player.drawangle,
							flashing = player.powers[pw_flashing],
							shield = player.powers[pw_shield],
							powerup = true,
							groundpound = player.mariogroundpound,
							spinjump = player.mariospinjump,
							spinsfx = player.mariospinsfx,
							gpjump = player.mariogpjump,
							crouch = player.mariocrouch,
							flipmove = player.marioflipmove,
							capeflight = player.mariocapeflight,
						}
						
						if not player.powers[pw_carry]
							P_ResetPlayer(player)
						end
						
						mo.flags = MF_NOCLIP|MF_NOGRAVITY
						player.pflags = $|PF_FULLSTASIS|PF_INVIS
						mo.momx = 0
						mo.momy = 0
						mo.momz = 0
						if P_IsObjectOnGround(mo) then mo.z = $ + P_MobjFlip(mo) end
						
						player.powers[pw_flashing] = flashingtics
						mo.state = S_PLAY_PAIN
						player.panim = PA_DASH
					end
					
					//No stacking!
					if player.powers[pw_shield] & SH_STACK
					and player.powers[pw_shield] & SH_NOSTACK
						if player.marioshield & SH_NOSTACK
							player.powers[pw_shield] = $ &~ SH_NOSTACK
						else
							player.powers[pw_shield] = $ &~ SH_STACK
						end
					end
					
					SpawnMarioShieldOrb(player)
				end
			end
			player.marioshield = player.powers[pw_shield]
			
			if player.powers[pw_carry] != CR_NIGHTSMODE
				player.powers[pw_nights_linkfreeze] = max($, 2)
			end
		end
		player.mariowassuper = false
	end
	player.mariolives = player.lives
	
	//Keep shoe between levels
	if not G_IsSpecialStage()
		if player.mariokeepshoe and not player.marioshoe and leveltime <= 2
			player.marioshoe = P_SpawnMobjFromMobj(mo, 0,0,0, MT_MARIO_SHOE)
			player.marioshoe.tracer = mo
		end
		player.mariokeepshoe = player.marioshoe and true or false
	end
	
	//Infinite air with certain powerups (the metals do this automatically)
	if player.mariosize > 0 or player.powers[pw_shield] == SH_BALLOON
	or player.powers[pw_shield] == SH_SUPERBALL or player.powers[pw_super]
		if player.powers[pw_underwater]
			player.powers[pw_underwater] = underwatertics
		end
		if player.powers[pw_spacetime]
			player.powers[pw_spacetime] = spacetimetics
		end
	end
	
	//P Balloon
	if player.powers[pw_shield] == SH_BALLOON
		player.marioballoon = $ + 1
		
		//Leave shoe if in one
		if player.marioshoe and player.marioshoe.valid
			player.marioshoe.tracer = nil
			player.marioshoe = nil
		end
		
		//Rotate
		if player.pflags & PF_SPINNING
		or (player.pflags & PF_JUMPED and not (player.pflags & PF_NOJUMPDAMAGE))
		or (player.mariogroundpound and player.mariogroundpound < GROUNDPOUNDSTART)
			mo.rollangle = $ + ANGLE_45
			
		elseif mo.marioexplode or MarioCantMove(player)
		or player.mariogroundpound >= GROUNDPOUNDSTART
		or P_IsObjectOnGround(mo)
			mo.rollangle = $/2
			
		else
			mo.rollangle = INT16_MAX/6*sin(FixedAngle(player.marioballoon*8*FRACUNIT))
		end
			
		//Start Ground Pound
		if not MarioCantMove(player) and not player.mariogroundpound
		and not P_IsObjectOnGround(mo) and not player.glidetime
		and (player.mariopressspin == 1 or player.mariopressdown == 1)
			ResetMario(player)
			
			if player.mariopressspin == 1 then player.mariopressspin = $ + 1 end
			if player.mariopressdown == 1 then player.mariopressdown = $ + 1 end
			
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			
			S_StartSound(mo, sfx_mrgnd1)
			MarioPlayVoice(player, "groundpound")
			
			player.pflags = $ | PF_FULLSTASIS
			mo.state = S_PLAY_STND
			mo.state = GROUNDPOUNDSTARTSTATE
			
			player.mariogroundpound = 1
		end
		
		//Bounce off floor
		if player.marioballoon > 2 and not (player.pflags & PF_SPINNING)
		and (not player.mariogroundpound or player.mariogroundpound > GROUNDPOUNDSTART)
		and P_IsObjectOnGround(mo) and not MarioCantMove(player)
			P_SetObjectMomZ(mo, 4*FRACUNIT, false)
			mo.z = $ + P_MobjFlip(mo)
			
			S_StartSound(mo, sfx_mrpbmp)
			player.mariosquish = SQUISHTIME
		end
		
		//Prepare to explode
		if (player.mariopresscustom or player.mariopressfire) and not player.mariogroundpound
		and not (player.pflags & PF_SPINNING) and not MarioCantMove(player)
			if not mo.marioexplode
				S_StartSound(mo, sfx_mrpbch)
				mo.marioexplode = 1
			elseif mo.marioexplode >= TICRATE*2
				if player == displayplayer or player == secondarydisplayplayer
					P_StartQuake(8*mo.scale, 10)
				end
				P_BlackOw(player)
				mo.marioexplode = 0
			else
				mo.marioexplode = $ + 1
			end
			
			mo.momx = $*8/9
			mo.momy = $*8/9
			
			if not player.mariosize
				mo.spritexscale = FRACUNIT
			end
			mo.spritexscale = $ + FixedMul($,
			sin(FixedAngle(mo.marioexplode*16*FRACUNIT))/4
			+ mo.marioexplode*UINT8_MAX*3)
			mo.spriteyscale = mo.spritexscale
			
			player.mariosquish = 1
		else
			mo.marioexplode = 0
			if S_SoundPlaying(mo, sfx_mrpbch)
				S_StopSoundByID(mo, sfx_mrpbch)
			end
			
			if player.mariogroundpound == GROUNDPOUNDSTART
				player.mariosquish = -SQUISHTIME
			end
			
			//Squishy
			if player.mariosquish
				if player.mariosquish > 0
					player.mariosquish = $ - 1
				else
					player.mariosquish = $ + 1
				end
				
				local factor = player.mariosquish*FRACUNIT/SQUISHFACTOR
				if not player.mariosize
					mo.spritexscale = FRACUNIT
					mo.spriteyscale = FRACUNIT
				end
				mo.spritexscale = $ + FixedMul($, factor)
				mo.spriteyscale = $ - FixedMul($, factor)
			end
		end
		
	//Reset values
	elseif player.marioballoon
		mo.rollangle = 0
		player.mariosquish = 0
		if not player.mariosize
			mo.spritexscale = FRACUNIT
			mo.spriteyscale = FRACUNIT
		end
		
		if mo.state == BALLOONSTATE
			if player.pflags & PF_SPINNING
				mo.state = S_PLAY_ROLL
			else
				mo.state = S_PLAY_FALL
			end
		end
			
		player.marioballoon = 0
	end

	//Prevent the rest of the script under certain circumstances
	if MarioCantMove(player)
		ResetMario(player)
		
		//Reset underwater junk
		if player.mariowaterswitch
			player.normalspeed = skins[player.skin].normalspeed
			player.acceleration = skins[player.skin].acceleration
			player.mariowaterswitch = false
		end
		
		// Mario death animation
		if not mo.health
			player.powers[pw_shield] = 0
			
			//Lengthen the animation
			if mo.fuse > TICRATE*2
			or (mo.fuse > TICRATE and not player.lives and G_GametypeUsesLives())
				player.deadtimer = 0
			end
			
			// The pause...
			if mo.fuse > 5*TICRATE/2
				if mo.fuse == 3*TICRATE - 1
				and not splitscreen and (mariodeathjingle.value or not player.lives)
					S_StopMusic(player)
				end
				if mo.movedir == DMG_DROWNED
					P_SetObjectMomZ(mo, FRACUNIT/2, false) // slower fall from drowning
					mo.state = S_PLAY_DEAD
					if S_SoundPlaying(mo, sfx_drown)
						S_StopSoundByID(mo, sfx_drown)
						P_PlayDeathSound(mo)
					end
				else
					P_SetObjectMomZ(mo, 2*FRACUNIT/3, false)
				end
				mo.momx = 0
				mo.momy = 0
			
			//Then the music and fling upwards!
			elseif mo.fuse == 5*TICRATE/2
				if mo.movedir != DMG_SPIKE and mo.movedir != DMG_DROWNED
					P_SetObjectMomZ(mo, 16*FRACUNIT, false)
				end
				
				//Play the music
				if not splitscreen and (mariodeathjingle.value or not player.lives)
					P_PlayJingleMusic(player, "MRDEAD", MUSIC_RELOADRESET, false, JT_OTHER)
					S_StartMusicCaption("Mario Death", 2*TICRATE, player)
				end
				
				if mo.movedir == DMG_DROWNED
					S_StartSound(mo, sfx_drown)
					MarioPlayVoice(player, "drown")
					mo.state = S_PLAY_DRWN
				else
					MarioPlayVoice(player, "die")
				end
			end
			
			//Game over music
			if not splitscreen and player.deadtimer == 2*TICRATE - TICRATE/25
			and not player.lives and G_GametypeUsesLives()
				P_PlayJingle(player, JT_GOVER)
			end
			
			RestoreMarioPain(player, mo, false, false, true)
			
		//Pain state
		elseif P_PlayerInPain(player) and mo.state == mo.info.painstate
		and player.powers[pw_flashing] and player.mariopain
			//Always set the angle
			player.drawangle = player.mariopain.drawangle
			
			//Prevent crushing
			if not (mo.flags & MF_NOCLIPHEIGHT) and player.mariosize
				if mo.z + mo.height > mo.ceilingz
					mo.z = mo.ceilingz - mo.height
				end
			end
			
			if mo.tics == 1 or player.mariopain.lava
				RestoreMarioPain(player, mo, true, true, false)
				
				if mo.state == LAVASTATE or mo.state == LEDGEGRABSTATE
				or mo.state == BALLOONSTATE
					if player.pflags & PF_SPINNING
						mo.state = S_PLAY_ROLL
					else
						mo.state = S_PLAY_FALL
					end
				end
				
				if player.mariopain.lava
					mo.z = $ + P_MobjFlip(mo)
					P_SetObjectMomZ(mo, 16*FRACUNIT, false)
					
					player.powers[pw_flashing] = TICRATE/2
					mo.state = LAVASTATE
					player.panim = PA_DASH
				elseif player.mariopain.mega and not player.mariomega
					player.powers[pw_flashing] = min($, TICRATE)
				elseif player.mariopain.powerup
					player.powers[pw_flashing] = 0
				else
					player.powers[pw_flashing] = min($, TICRATE*2)
				end
				
				if player.mariopain.flashing
					player.powers[pw_flashing] = max($, player.mariopain.flashing)
				end
				
				player.mariopain = nil
			else
				if player.mariopain.mega
					if (player.mariosize <= 0 and not player.mariomega)
					or (player.mariosize >= FRACUNIT*2 and player.mariomega)
						mo.tics = min($, 2)
						
					//Mega stars
					elseif not (mo.tics % 5)
						local scale = mo.scale + FixedMul(player.mariosize, mo.scale)
						local color = SKINCOLOR_SUPERSILVER5 + 5*((mo.tics/5) % 7)
						
						MarioCircle(mo, MT_GHOST, 8, 8*scale, 0, scale/2,
						0, color, false, player.drawangle)
					end
				elseif mo.tics > TICRATE*2/3
					mo.tics = TICRATE*2/3
					
					if player.mariopain.shield == SH_SUPERBALL
					and not player.mariopain.powerup
						S_StopSound(mo)
						S_StartSound(mo, sfx_mrbal6)
					elseif player.mariopain.shield == SH_CAPE
					and player.mariopain.powerup and not player.mariopain.mega
						P_SpawnMobjFromMobj(mo, 0,0,0, MT_MARIO_DUST)
					end
				end
				
				player.panim = PA_DASH
				if player.mariopain.powerup
					mo.frame = player.mariopain.frame
					if player.mariopain.shield == SH_CAPE
					and not player.mariopain.mega
						mo.flags2 = $ | MF2_DONTDRAW
					elseif player.mariopain.shield == SH_BALLOON
					and ((mo.tics/4) % 2 or player.mariopain.mega)
						mo.frame = $ &~ FF_FRAMEMASK
						mo.sprite2 = SPR2_FLY_
					elseif player.mariopain.sprite2 == SPR2_FLY_
					and (player.mariopain.shield == SH_BALLOON or (mo.tics/4) % 2)
						if player.mariopain.pflags & PF_SPINNING
							mo.sprite2 = SPR2_ROLL
						else
							mo.sprite2 = SPR2_FALL
						end
					else
						mo.sprite2 = player.mariopain.sprite2
					end
				end
				
				//Powerup flashing
				if player.mariopain.shield and not (mo.tics % 4)
				and not player.mariopain.mega
					local oldshield = player.powers[pw_shield]
					if (mo.tics/4) % 2
						player.shieldscale = skins[player.skin].shieldscale
						player.powers[pw_shield] = player.mariopain.shield
						if SpawnMarioShieldOrb(player) and not player.mariopain.powerup
							//I hate this as much as you do
							for mobj in mobjs.iterate()
								if mobj.flags2 & MF2_SHIELD and mobj.target == mo
								and mobj.flags & MF_SCENERY
									mobj.flags = $ &~ MF_SCENERY
									mobj.fuse = 4
									mobj.scale = FixedMul(mobj.scale, player.shieldscale)
									break
								end
							end
						end
					else
						player.shieldscale = 0
						player.powers[pw_shield] = SH_NONE
					end
					
					player.powers[pw_shield] = oldshield
				end
				
				mo.flags = MF_NOCLIP|MF_NOGRAVITY
				player.pflags = $|PF_FULLSTASIS
				mo.momx = 0
				mo.momy = 0
				mo.momz = 0
				if P_IsObjectOnGround(mo) then mo.z = $ + P_MobjFlip(mo) end
			end
		else
			RestoreMarioPain(player, mo, true, true, true)
		end
		
		//This is dumb, but oh well
		if player.powers[pw_carry] == CR_ROPEHANG
		and not P_IsObjectOnGround(mo)
			player.glidetime = max($, 1)
		end
		
		return
	end
	RestoreMarioPain(player, mo, false, false, true)
	
	//End wall jump after a certain period of time
	//Lets you jump off 90 degree angles again
	if player.mariowalljump
		if player.mariospinjump
			player.mariowalljump = $ - 1
		else
			player.mariowalljump = $ - 2
		end
	end
	
	//Extra gravity specifically when jumping or performing an aerial move
	if not (mo.flags & MF_NOGRAVITY) and not P_IsObjectOnGround(mo)
	and not player.mariogroundpound
		if player.mariocapeflight and mo.state == CAPEFLYSTATE
			mo.momz = $ + P_GetMobjGravity(mo)/4
		elseif player.powers[pw_shield] == SH_BALLOON
			if mo.marioexplode
				mo.momz = $/2
			elseif not P_IsObjectInGoop(mo)
				mo.momz = $ - P_GetMobjGravity(mo)*2/3
			end
			
		elseif (player.mariosize >= 0 or (player.luigispin and player.mariospinjump))
		and ((player.pflags & PF_JUMPED and (player.pflags & PF_NOJUMPDAMAGE
		or mo.state == TRIPLEJUMPSTATE or mo.state == BACKFLIPSTATE)
		and not (player.pflags & PF_SPINNING)) or mo.state == DIVESTATE)
			//Extra falling speed
			if mo.momz*P_MobjFlip(mo) < 0
				mo.momz = $ + P_GetMobjGravity(mo)/4
			end
			
			if (player.powers[pw_shield] == SH_GOLDFLOWER
			or player.powers[pw_shield] == SH_METAL)
			and not (player.mariospinjump
			and (mo.momz*P_MobjFlip(mo) > 0 or player.luigispin))
				mo.momz = $ + P_GetMobjGravity(mo)/2
				
			elseif not player.mariospinjump
				mo.momz = $ + P_GetMobjGravity(mo)/4
				
			elseif player.luigispin and player.mariospinjump
			and mo.momz*P_MobjFlip(mo) < 4*P_GetMobjGravity(mo)*P_MobjFlip(mo)
				mo.momz = 4*P_GetMobjGravity(mo)
			end
		end
	end
	
	//Change stats underwater to go against the water physics
	if (player.powers[pw_shield] == SH_GOLDFLOWER
	or player.powers[pw_shield] == SH_METAL)
	and mo.eflags & MFE_UNDERWATER and not P_IsObjectInGoop(mo)
	and not player.gotflagdebuff
		if not player.mariowaterswitch
			player.normalspeed = skins[player.skin].normalspeed*2
			player.acceleration = skins[player.skin].acceleration*3/2
			player.mariowaterswitch = true
		else
			if player.normalspeed >= skins[player.skin].normalspeed
				player.normalspeed = max($, skins[player.skin].normalspeed*2)
			end
			if player.acceleration >= skins[player.skin].acceleration
				player.acceleration = max($, skins[player.skin].acceleration*3/2)
			end
		end
		if mo.eflags & MFE_SPRUNG
			mo.z = $ + P_MobjFlip(mo)
			mo.momz = $*11/8
		end
		if P_GetMobjGravity(mo) > mo.scale/-2
		and P_GetMobjGravity(mo) < 0
		and not P_IsObjectOnGround(mo)
			P_SetObjectMomZ(mo, -gravity + FRACUNIT/3, true)
		end
		
	elseif player.mariowaterswitch
		player.normalspeed = skins[player.skin].normalspeed
		player.acceleration = skins[player.skin].acceleration
		player.mariowaterswitch = false
		if mo.momz * P_MobjFlip(mo) > 0 and player.pflags & PF_JUMPED
		and not (mo.eflags&MFE_UNDERWATER) and not P_IsObjectInGoop(mo)
		and not player.gotflagdebuff
			mo.momz = $/2
		end
	end
	
	// Ring attraction
	if (shield == SH_GOLDFLOWER or player.powers[pw_super])
	and leveltime & 1
		local maxdist = FixedMul(RING_DIST, mo.scale)
		local span = ANGLE_45
		local thokmobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
		thokmobj.radius = maxdist

		local rings = {
			MT_RING, MT_FLINGRING, MT_COIN, MT_FLINGCOIN,
			MT_BLUESPHERE, MT_FLINGBLUESPHERE
		}

		searchBlockmap("objects", function(refmobj, object)
			if object == mo
			or object.flags & (MF_NOCLIPTHING|MF_MISSILE)
			or object.flags2 & MF2_FRET
			or object.health <= 0
			or (object.mariomonitorcoin and object.fuse > MARIO_DESPAWNTIME - TICRATE)
			or not P_CheckSight(mo, object)
			or ((mo.flags2 & MF2_TWOD or twodlevel) and abs(mo.y - object.y) > mo.radius)
				return
			end

			local isRing = false // Is a ring?

			// Iterate ring types
			for i, v in ipairs(rings) do
				if object.type == v // We're a ring type?
					isRing = true // Set isRing and break
					break
				end
			end

			// Not a ring
			if not isRing
				return
			end

			local zdist = (mo.z + mo.height/2) - (object.z + mo.height/2)
			local dist = FixedHypot(mo.x - object.x, mo.y - object.y)

			if R_PointToAngle2(0, 0, dist, zdist) + span > span*2
			and R_PointToAngle2(0, 0, dist, -1*zdist) + span > span*2
				return // Don't home outside of desired angle!
			end
			
			dist = FixedHypot(dist, zdist*2)

			if (dist > maxdist)
				return // out of range
			end
			
			if object.info.reactiontime
			and object.type != object.info.reactiontime
				local newring = P_SpawnMobjFromMobj(object, 0, 0, 0, object.info.reactiontime)
				P_RemoveMobj(object)
				object = newring
			end
			
			local speed = object.info.speed
			object.info.speed = 80*FRACUNIT

			P_HomingAttack(object, mo)
			object.info.speed = speed
		end, thokmobj)

		P_RemoveMobj(thokmobj)
	end
	
	//Cape P Speed
	if player.powers[pw_shield] != SH_CAPE
	or mo.state == CAPEFLYSTATE or mo.eflags & MFE_UNDERWATER
		player.mariopspeed = 0
	elseif player.speed >= FixedMul(skins[player.skin].runspeed, mo.scale)
	and not player.mariocapeflight
		player.mariopspeed = PSPEEDTIME
	elseif P_IsObjectOnGround(mo) and not (player.pflags & (PF_SPINNING|PF_STASIS))
	and not player.mariocapeflight
	and player.speed >= mo.scale*(twodlevel and RUNSPEED/3 or RUNSPEED)
		player.mariopspeed = min($ + 1, PSPEEDTIME)
	elseif (P_IsObjectOnGround(mo) or player.pflags & (PF_SPINNING|PF_STASIS))
	and player.mariopspeed
		player.mariopspeed = $ - 1
	end
	
	//Wall bonking
	if player.mariobouncedir != nil and player.mariobouncespeed != nil
	and player.mariobouncespeed - player.speed > player.mariobouncespeed/2
	and player.powers[pw_pushing]
		if player.powers[pw_justlaunched] and not player.mariolongjump
			player.pflags = $ | PF_SPINNING
			if mo.state != DIVESTATE
				mo.state = DIVESTATE
			end
		else
			player.drawangle = player.mariobouncedir - ANGLE_180
			if P_IsObjectOnGround(mo)
				mo.z = $ + P_MobjFlip(mo)
				P_InstaThrust(mo, player.mariobouncedir, 12*mo.scale)
			else
				P_InstaThrust(mo, player.mariobouncedir, 8*mo.scale)
			end
			
			player.pflags = $ &~ PF_SPINNING
			mo.state = BONKSTATE
			
			player.mariobonkland = BONKLANDTICS
			
			S_StartSound(mo,sfx_mrbonk)
			MarioPlayVoice(player, "bonk")
			
			MarioCircle(mo, MT_GHOST, 8, 8*mo.scale, FixedDiv(mo.height, mo.scale)/2,
			mo.scale/2, 0, nil, true, player.drawangle)
			
			if player.mariolongjump
				player.normalspeed = skins[player.skin].normalspeed
				player.mariolongjump = 0
				
				player.mariolongbonk = 7
			end
		end
	end
	player.mariobouncedir = nil
	player.mariobouncespeed = nil
	
	//This is dumb but oh well
	if not player.mariobonkland
		player.mariolongbonk = 0
		player.mariobonktimer = 0
	end
	
	if not player.mariobonkland and mo.state == BONKSTATE
		player.mariobonkland = mo.tics
	end
	
	//Cape flight
	if player.mariocapeflight
		player.mariocapeflight = $ + 1
		if mo.mariocapefall
			mo.mariocapefall = $ - 1
		end
		
		local forwardmove = player.cmd.forwardmove
		if player.mo.flags2 & MF2_TWOD or twodlevel
			forwardmove = player.cmd.sidemove
			if player.mariojumpdir != nil and player.mariojumpdir < 0
				forwardmove = -$
			end
		end
		
		if mo.state == CAPEFLYSTATE or mo.state == S_TAILSOVERLAY_FLY
			player.panim = PA_DASH
		end
		
		local capegrav = P_GetMobjGravity(mo)
		if P_IsObjectInGoop(mo)
			capegrav = FixedMul(gravity/3, mo.scale) * -P_MobjFlip(mo)
		end
		
		//Cancel with state/shield
		if (mo.state != CAPEFLYSTATE and mo.state != S_TAILSOVERLAY_FLY
		and player.panim != PA_JUMP and player.panim != PA_FALL)
		or player.powers[pw_shield] != SH_CAPE
			player.mariocapeflight = 0
			player.mariocapemom = 0
			mo.mariocapefall = 0
			
			if mo.state == CAPEFLYSTATE
				mo.state = S_PLAY_FALL
			end
		
		//Ground nuke
		elseif P_IsObjectOnGround(mo) and forwardmove >= 35
		and prevmomz * P_MobjFlip(mo) < -16*mo.scale
			player.mariocapeflight = 0
			player.mariocapemom = 0
			mo.mariocapefall = 0
			
			local dist = RING_DIST*2
			P_RadiusAttack(mo, mo, dist, DMG_NUKE, false)
			MarioCircle(mo, MT_SUPERSPARK, 16, 60*mo.scale,
			0, mo.scale, 0, nil, false, player.drawangle)
			
			if player == displayplayer or player == secondarydisplayplayer
			or (displayplayer and displayplayer.valid
			and displayplayer.mo and displayplayer.mo.valid
			and P_AproxDistance(mo.x - displayplayer.mo.x,
			mo.y - displayplayer.mo.y) < FixedMul(dist, mo.scale))
				P_StartQuake(32*mo.scale, 5)
			end
			
			S_StartSound(mo, sfx_cannon)
			
			mo.state = S_PLAY_RUN
			player.mariopspeed = PSPEEDTIME
			
		//Sliding
		elseif P_IsObjectOnGround(mo)
			if mo.state != CAPEFLYSTATE
				mo.state = CAPEFLYSTATE
			end
			mo.frame = ($ &~ FF_FRAMEMASK) + D
			mo.tics = -1
			
			player.pflags = $ | PF_SPINNING
			player.panim = PA_ROLL
			
			player.mariojumpdir = player.drawangle
			player.mariocapemom = 0
			mo.mariocapefall = 0
			
		//Jump upwards
		elseif mo.state != CAPEFLYSTATE
			player.mariocapemom = 0
			mo.mariocapefall = 0
			
			if player.cmd.buttons & BT_JUMP
			and player.mariocapeflight < CAPEJUMPTIME
				player.pflags = $ &~ PF_JUMPED
				P_DoJump(player, false)
				mo.momz = $*3/2
			elseif mo.momz * P_MobjFlip(mo) < 0
				mo.state = CAPEFLYSTATE	
				if player.pflags & PF_ANALOGMODE
					player.mariojumpdir = player.drawangle
				else
					player.mariojumpdir = mo.angle
				end
			end
			
		//Flight
		else
			local speed = FixedHypot(mo.momx, mo.momy)
			player.pflags = $ &~ PF_APPLYAUTOBRAKE &~ PF_STARTJUMP &~ PF_SPINNING | PF_NOJUMPDAMAGE
			player.powers[pw_noautobrake] = max($, 2)
			
			//Turning
			if player.mariojumpdir == nil
				if player.pflags & PF_ANALOGMODE
					player.mariojumpdir = player.drawangle
				else
					player.mariojumpdir = mo.angle
				end
			elseif not (mo.flags2 & MF2_TWOD) and not twodlevel
				local sidemove = player.cmd.sidemove
				local auto = not (player.pflags & PF_ANALOGMODE) and not sidemove
				local olddir = player.mariojumpdir
				local diff = player.mariojumpdir - mo.angle
				diff = InvAngle(InvAngle($))
				if auto and player.mariojumpdir != mo.angle
					if diff > 0
						sidemove = 50
					else
						sidemove = -50
					end
				end
				
				local change = CAPETURNFACTOR*sidemove*FRACUNIT
				player.mariojumpdir = $ - change
				
				if auto
					local newdiff = player.mariojumpdir - mo.angle
					newdiff = InvAngle(InvAngle($))
					if (diff < 0 and newdiff > 0)
					or (diff > 0 and newdiff < 0)
						player.mariojumpdir = mo.angle
					end
				end
					
				local oldmomx = mo.momx
				local oldmomy = mo.momy
					
				P_InstaThrust(mo, player.mariojumpdir, speed)
					
				mo.momx = $/2 + oldmomx/2
				mo.momy = $/2 + oldmomy/2
			end
			player.drawangle = player.mariojumpdir
			//mo.angle = player.mariojumpdir
			
			//Movement
			mo.frame = $ &~ FF_FRAMEMASK
			
			//Flying Upwards
			if player.mariocapemom
				player.mariocapemom = $ - 1
				if not player.mariocapemom
					player.mariocapemom = -1
				end
				
				mo.mariocapefall = CAPEFALLTIME
				if forwardmove > -35 and mo.momz * P_MobjFlip(mo) < 0
					mo.frame = $ + B
				end
				
				if player.mariocapemom > 0
					mo.momz = $ - CAPEMOMFACTOR*capegrav
					mo.momz = $*31/32
				elseif forwardmove > -35 and mo.momz * P_MobjFlip(mo) < 0
					player.mariocapemom = 0
				end
				
			//Rocketing Downwards
			elseif forwardmove >= 35
				mo.mariocapedive = true
				mo.frame = $ + C
				mo.momz = $ + capegrav*2
				if speed < LONGJUMPSPEED*mo.scale
					P_Thrust(mo, player.mariojumpdir, mo.scale*3/2)
					speed = FixedHypot(mo.momx, mo.momy)
					if speed > LONGJUMPSPEED*mo.scale
						P_InstaThrust(mo,
						R_PointToAngle2(0,0, mo.momx, mo.momy), LONGJUMPSPEED*mo.scale)
					end
				end
				
			//Start upwards
			elseif forwardmove <= -35
				if mo.mariocapefall
					mo.mariocapefall = CAPEFALLTIME
					mo.momz = $*31/32
				elseif mo.momz * P_MobjFlip(mo) > 0 or player.powers[pw_pushing]
					mo.mariocapefall = max($, 2)
					mo.momz = $*31/32
				else
					player.mariocapemom = max(min(abs(mo.momz/CAPEMOMDIV/mo.scale), MAXCAPEMOM), MINCAPEMOM)
					S_StartSound(mo, sfx_mrfly)
					
					mo.momz = $ + CAPEMOMFACTOR*capegrav
					mo.frame = $ + B
				end
				
			//Neutral
			else
				mo.frame = $ + B
				if mo.mariocapedive
					mo.mariocapedive = nil
				else
					mo.momz = $*31/32
				end
			end
		end
	end
	
	//P Balloon (again to prevent the ability if else tree)
	if player.powers[pw_shield] == SH_BALLOON and not player.mariogroundpound
		ResetMario(player)
	
	//Bonk Landing
	elseif player.mariobonkland
		ResetMario(player, true)
		
		if player.mariolongbonk
			player.pflags = $ | PF_STASIS
			player.mariolongbonk = $ - 1
		else
			player.pflags = $ | PF_FULLSTASIS
		end
		
		if S_SoundPlaying(mo, sfx_spin)
			S_StopSoundByID(mo, sfx_spin)
		end
		
		player.mariobonktimer = $ + 1
		if player.mariobonktimer >= BONKTIMETICS
			player.mariobonkland = 0
			mo.state = states[BONKSTATE].nextstate
		elseif P_IsObjectOnGround(mo)
			mo.state = BONKSTATE
			mo.sprite2 = SPR2_TALA
			mo.tics = player.mariobonkland
			player.pflags = $ | PF_JUMPSTASIS
			
			if player.mariobonkland == BONKLANDTICS
				if player.powers[pw_shield] != SH_SUPERBALL and mariovoice.value
					MarioPlayVoice(player, "butt")
				else
					S_StartSound(mo, sfx_mario1)
				end
			end
			player.mariobonkland = $ - 1
		elseif mo.state == BONKSTATE
			player.mariobonkland = BONKLANDTICS
			mo.tics = player.mariobonkland
			if mo.sprite2 == SPR2_TALA
				mo.sprite2 = states[mo.state].frame
			end
		else
			player.mariobonkland = 0
		end
	
	//Wall sliding
	elseif player.mariowallslide
		local line = lines[player.lastlinehit]
		local side = sides[player.lastsidehit]
		if not P_IsObjectOnGround(mo)
		and line and line.valid and side and side.valid
			local pointx, pointy = P_ClosestPointOnLine(mo.x, mo.y, line)
			player.drawangle = R_PointToAngle2(pointx, pointy, mo.x, mo.y)
			
			//Slide dust
			local dust = P_SpawnMobjFromMobj(mo, 0,0, FixedDiv(mo.height, mo.scale)*2/3, MT_MARIO_DUST)
			local dist = R_PointToDist2(pointx, pointy, mo.x, mo.y)
			local angle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
			if line.frontside == side
				angle = $-ANGLE_180
			end
			dust.scale = $/2
			P_TeleportMove(dust,
			P_ReturnThrustX(dust, player.drawangle, dist/-2)
			+ P_ReturnThrustX(dust, angle, P_SignedRandom()*dust.scale/4) + dust.x,
			P_ReturnThrustY(dust, player.drawangle, dist/-2)
			+ P_ReturnThrustY(dust, angle, P_SignedRandom()*dust.scale/4) + dust.y,
			dust.z)
			
			//Slide momentum
			mo.momx = ($ - player.cmomx)*5/6 + player.cmomx
			mo.momy = ($ - player.cmomy)*5/6 + player.cmomy
			
			if mo.momz*P_MobjFlip(mo) < 0
				mo.momz = min(max($, -WALLJUMPFALL*mo.scale), WALLJUMPFALL*mo.scale)
			end
			
			//Stuff
			mo.state = WALLSLIDESTATE
			player.mariowallslide = $ - 1
			player.marioslidesfx = max($, 1)

			//Move with conveyors
			if side == player.mariolastside
				//Move with conveyors
				if side.textureoffset != player.mariosidexscroll
				or side.rowoffset != player.mariosideyscroll
					//Horizontal movement
					P_TryMove(mo,
					mo.x+FixedMul(side.textureoffset-player.mariosidexscroll, cos(angle)),
					mo.y+FixedMul(side.textureoffset-player.mariosidexscroll, sin(angle)),
					true)
					
					//Verical Movement
					mo.z = $ + side.rowoffset-player.mariosideyscroll
					mo.momz = 0
				end
			else
				player.mariolastside = side
			end
			
			//Update variables
			player.mariosidexscroll = side.textureoffset
			player.mariosideyscroll = side.rowoffset
		else
			player.mariowallslide = 0
		end
		
		player.mariotriple = 0
		player.mariospinsfx = 0
		player.marioflipmove = 0
		player.mariojumpdir = nil
		
		if player.mariolongjump
			player.normalspeed = skins[player.skin].normalspeed
			player.mariolongjump = 0
		end
		
		if not player.mariowallslide
			mo.state = S_PLAY_FALL
			player.marioslidesfx = 0
			player.mariolastside = nil
		end
	
	//Swimming bubbles
	elseif mo.state == SWIMSTATE
	and player.mariopressspin != 1 and player.mariopressdown != 1
	and player.mariopresscustom != 1 and player.mariopressfire != 1
		ResetMario(player)
		
		if not (mo.frame & FF_FRAMEMASK) and player.panim != PA_ABILITY2
			mo.tics = -1
		end
		
		if not (mo.eflags & MFE_UNDERWATER) or mo.eflags & MFE_GOOWATER
			if mo.momz * P_MobjFlip(mo) < 0 or mo.eflags & MFE_GOOWATER
				mo.state = S_PLAY_JUMP
			else
				player.pflags = $ &~ PF_JUMPED
				P_DoJump(player, true)
			end
			
			S_StopSoundByID(mo, sfx_mrswm1)
			S_StopSoundByID(mo, sfx_mrswm2)
			
		elseif mo.tics >= 0
			local bubble = P_SpawnMobjFromMobj(mo, 0,0, FixedDiv(mo.height, mo.scale)/3, MT_MEDIUMBUBBLE)
			local dist = mo.radius*3/2 + player.speed
			if mo.skin == "luigi" then dist = player.speed end //Luigi's hands aren't in front
			local angle = player.drawangle + ANGLE_90
			
			P_TeleportMove(bubble,
			P_ReturnThrustX(bubble, player.drawangle, dist)
			+ P_ReturnThrustX(bubble, angle, P_SignedRandom()*bubble.scale/4) + bubble.x,
			P_ReturnThrustY(bubble, player.drawangle, dist)
			+ P_ReturnThrustY(bubble, angle, P_SignedRandom()*bubble.scale/4) + bubble.y,
			bubble.z)
			
			bubble.momx = mo.momx
			bubble.momy = mo.momy
			
			P_Thrust(bubble, player.drawangle, -6*bubble.scale)
		end
		
	//Lava jump
	elseif mo.state == LAVASTATE
		ResetMario(player)
		player.pflags = $ &~ (PF_JUMPED|PF_NOJUMPDAMAGE|PF_SPINNING)
		
		if mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_TOUCHLAVA)
		and not player.powers[pw_flashing]
			S_StartSound(mo, sfx_s3k4b)
			mo.state = S_PLAY_FALL
		else
			if P_IsObjectOnGround(mo)
				P_SetObjectMomZ(mo, 4*FRACUNIT, false)
				mo.momx = $/2
				mo.momy = $/2
				
				S_StartSound(mo, sfx_mario1)
				player.panim = PA_ETC
			end
		
			for i = 1, 2
				local smoke = P_SpawnMobjFromMobj(mo,32*(P_RandomFixed()-FRACUNIT/2),
				32*(P_RandomFixed()-FRACUNIT/2),
				6*FRACUNIT + FixedDiv(mo.momz*P_MobjFlip(mo), mo.scale), MT_SMOKE)
			end
		end
	
	//Ground Pound (Yes, this is mostly just Crash's Body Slam, which is mostly just Yoshi's Ground Pound)
	elseif player.mariogroundpound
		//Odyssey-style Dive/Roll
		if (player.mariopresscustom == 1 or player.mariopressspin == 1)
		and not (mo.eflags & MFE_SPRUNG)
		and (player.mariogroundpound != GROUNDPOUNDSTART
		or player.powers[pw_shield] == SH_BALLOON or player.powers[pw_super])
			if not P_IsObjectOnGround(mo) or player.powers[pw_shield] == SH_BALLOON
				DoDive(player, true)
			else
				if mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_GOOWATER)
				and not player.mariowaterswitch
					P_InstaThrust(mo, MarioInputAngle(player), LONGJUMPSPEED*mo.scale/2)
				else
					P_InstaThrust(mo, MarioInputAngle(player), LONGJUMPSPEED*mo.scale)
				end
				player.pflags = $ | PF_SPINNING
				
				MarioCircle(mo, MT_MINECARTSPARK, 12, 8*mo.scale, 0,
				mo.scale/4, 2, 0, true, player.drawangle)
				
				S_StartSound(mo, sfx_zoom)
				mo.state = S_PLAY_ROLL
			end
			
			player.mariogroundpound = 0
			player.mariogroundpoundtime = 0
			
		//Cancel
		elseif (mo.momz*P_MobjFlip(mo) > mo.scale*4
		or (P_IsObjectInGoop(mo) and player.mariogroundpound >= GROUNDPOUNDSTART)
		or ((player.mariogroundpound > GROUNDPOUNDSTART or player.mariogroundpoundtime > TICRATE*5)
		and not P_IsObjectOnGround(mo))) and not (mo.eflags & MFE_SPRUNG)
			if P_IsObjectInGoop(mo)
				mo.state = GROUNDPOUNDENDSTATE
				mo.tics = -1
				player.glidetime = -1
			elseif mo.state == GROUNDPOUNDSTARTSTATE
			or mo.state == GROUNDPOUNDENDSTATE
				mo.state = S_PLAY_SPRING
			end
			
			player.mariogroundpound = 0
			player.mariogroundpoundtime = 0
			
		//Landing
		elseif (P_IsObjectOnGround(mo) and player.mariogroundpound >= GROUNDPOUNDSTART)
		or mo.eflags & MFE_SPRUNG
			if player.mariogroundpound < GROUNDPOUNDEND
			or player.mariopressspin or player.mariopressdown
				if player.mariogroundpound == GROUNDPOUNDSTART
					DoMarioGroundPoundHit(player)
				end
				
				mo.state = GROUNDPOUNDENDSTATE
				mo.tics = 2
				
				player.mariogroundpound = $ + 1
			else
				player.mariogroundpound = 0
				player.mariogroundpoundtime = 0
			end
			
			if mo.eflags & MFE_SPRUNG
				mo.momx = $*5/4
				mo.momy = $*5/4
				mo.momz = $*5/4
				
				player.mariogroundpound = 0
				player.mariogroundpoundtime = 0
				S_StartSound(mo, sfx_sprong)
				
				mo.state = S_PLAY_JUMP
				player.mariogpjump = 16
				
				MarioPlayVoice(player, "triple")
			elseif player.mariogroundpound < GROUNDPOUNDEND
			and mo.subsector and mo.subsector.valid
			and mo.subsector.sector and mo.subsector.sector.valid
				for fof in mo.subsector.sector.ffloors()
					if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
					if not (fof.flags & FF_MARIO) continue end -- Is it bustable?
					
					if mo.z + mo.momz + mo.height < fof.bottomheight continue end -- Are we too low?
					if mo.z + mo.momz > fof.topheight continue end -- Are we too high?  
					
					if MarioGroundPoundBlock(mo, fof)
					and (player.mariopressspin or player.mariopressdown)
						player.mariogroundpound = GROUNDPOUNDSTART + 1
					end
				end
			end
			
		//Falling
		elseif player.mariogroundpound == GROUNDPOUNDSTART
			mo.momz = $ + 4*P_GetMobjGravity(mo)
			
			mo.state = GROUNDPOUNDENDSTATE
			mo.tics = -1
			player.panim = PA_DASH
			
			if mo.subsector and mo.subsector.valid
			and mo.subsector.sector and mo.subsector.sector.valid
				for fof in mo.subsector.sector.ffloors()
					if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
					if not (fof.flags & FF_BUSTUP) -- Is it bustable?
					and not (fof.flags & FF_MARIO and fof.flags & FF_SHATTERBOTTOM) continue end
					
					if mo.z + mo.momz + mo.height < fof.bottomheight continue end -- Are we too low?
					if mo.z + mo.momz > fof.topheight continue end -- Are we too high?  
					
					EV_CrumbleChain(fof) -- Crumble
					if fof.master.flags & ML_EFFECT5
						P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, mo, fof.target)
					end
					
					mo.momz = 0
					
					DoMarioGroundPoundHit(player)
				end
			end
		
		//Start up
		else
			if mo.state == GROUNDPOUNDSTARTSTATE
				P_SetObjectMomZ(mo, 4*FRACUNIT, false)
			else
				mo.momz = 0
			end
			
			player.mariogroundpound = $ + 1
			if player.mariogroundpound == GROUNDPOUNDSTART
				P_SetObjectMomZ(mo, -20*FRACUNIT, false)
			end
		end
		
		if player.mariogroundpound
			if mo.state != GROUNDPOUNDSTARTSTATE
			and mo.state != GROUNDPOUNDENDSTATE
				mo.state = GROUNDPOUNDENDSTATE
				mo.tics = -1
			end
			
			mo.momx = player.cmomx
			mo.momy = player.cmomy
			
			if player.mariogroundpound <= GROUNDPOUNDSTART
				player.pflags = $ | PF_JUMPSTASIS
				if player.cmd.buttons & BT_JUMP
					player.pflags = $ | PF_JUMPDOWN
				else
					player.pflags = $ &~ PF_JUMPDOWN
				end
			end
			
			//Face one direction
			if player.mariojumpdir == nil
				player.mariojumpdir = player.drawangle
			else
				player.drawangle = player.mariojumpdir
			end
			
			player.mariogroundpoundtime = $ + 1
		end
	
	//Diving
	elseif mo.state == DIVESTATE
		player.pflags = $ | PF_SPINNING &~ PF_JUMPED
		
		if S_SoundPlaying(mo, sfx_spin)
			S_StopSoundByID(mo, sfx_spin)
		end
		
		if player.mariojumpdir == nil or player.powers[pw_justsprung]
			player.mariojumpdir = player.drawangle
		end
		
		//Ground or air dive?
		if P_IsObjectOnGround(mo)
			player.mariojumpdir = nil
			
			//Hop
			if player.mariopressspin == 1 or player.mariopresscustom == 1
				local verticalthrust = P_GetMobjGravity(mo)
				if mo.eflags & MFE_GOOWATER then verticalthrust = -gravity*P_MobjFlip(mo) end
				if abs(verticalthrust) > mo.scale/2 then verticalthrust = mo.scale/-2*P_MobjFlip(mo) end
				mo.momz = -8 * verticalthrust
				P_Thrust(mo, MarioInputAngle(player), 10*mo.scale)
				
				mo.sprite2 = SPR2_SWIM
				mo.tics = 4
				
				S_StartSound(mo, sfx_mrdive)
				MarioPlayVoice(player, "hop")
				
				player.glidetime = GLIDETIMETIME
				
			//Spindash switch
			elseif mo.subsector and mo.subsector.valid
			and mo.subsector.sector and mo.subsector.sector.valid
				for fof in mo.subsector.sector.ffloors()
					if fof.valid and mo.z + mo.momz <= fof.topheight
					and mo.z + mo.momz + mo.height >= fof.bottomheight
					and fof.master and fof.master.valid and fof.master.flags&ML_NOCLIMB
					and (fof.master.special == 153 or fof.master.special == 180
					or (fof.master.special >= 150 and fof.master.special <= 152)
					or (fof.master.special >= 190 and fof.master.special <= 195)
					or fof.master.special == 176 or fof.master.special == 177)
						player.pflags = $|PF_STARTDASH
						P_InstaThrust(player.mo,0,0)
					end
				end
			end
		else
			player.drawangle = player.mariojumpdir
			player.marioslidesfx = 0
		end
		
		//Landing cut
		if mo.eflags & MFE_JUSTHITFLOOR and not player.powers[pw_justsprung]
			mo.momx = $*2/3
			mo.momy = $*2/3
		end
		
		//Extra spin control
		if not player.glidetime
			local oldmomx = mo.momx
			local oldmomy = mo.momy
			local oldspeed = R_PointToDist2(0,0, oldmomx, oldmomy)
			local factor = P_IsObjectOnGround(mo) and 128 or 64
			
			local angle = player.cmd.angleturn<<16
			if mo.flags2 & MF2_TWOD or twodlevel
				angle = ANGLE_90
			end
			
			P_Thrust(mo, angle, player.cmd.forwardmove*mo.scale/factor)
			P_Thrust(mo, angle - ANGLE_90, player.cmd.sidemove*mo.scale/factor)
			
			if R_PointToDist2(0,0, mo.momx, mo.momy) > oldspeed
				P_InstaThrust(mo, R_PointToAngle2(0,0, mo.momx, mo.momy), oldspeed)
			end
		end
		
	//Hammer time
	elseif mo.state == S_PLAY_MELEE or mo.state == S_PLAY_MELEE_FINISH
	or mo.state == S_PLAY_MELEE_LANDING
		player.powers[pw_noautobrake] = max($, 2)
		
		if mo.state != S_PLAY_MELEE_LANDING
			player.pflags = $ | PF_JUMPSTASIS
			if player.cmd.buttons & BT_JUMP
				player.pflags = $ | PF_JUMPDOWN
			else
				player.pflags = $ &~ PF_JUMPDOWN
			end
		end
		
		if mo.state == S_PLAY_MELEE and (mo.frame & FF_FRAMEMASK) == C
			mo.state = states[$].nextstate
		end
		
		if mo.state == S_PLAY_MELEE_FINISH then mo.tics = -1 end
		
		//Floor hit
		if P_IsObjectOnGround(mo) and mo.state == S_PLAY_MELEE_FINISH
			DoMarioHammerHit(player)
		end
		
		//Go forward to simulate the hitbox
		if mo.state == S_PLAY_MELEE_FINISH or mo.state == S_PLAY_MELEE_LANDING
			local dist = HAMMERDIST*mo.scale
			dist = $ + FixedMul(player.mariosize, $)
			
			local x = mo.x
			local y = mo.y
			local z = mo.z
			local momx = mo.momx
			local momy = mo.momy
			local momz = mo.momz
			local flags = mo.flags
			local godmode = player.pflags & PF_GODMODE
			local flashing = player.powers[pw_flashing]
			player.panim = PA_DASH
			mo.flags = $ | MF_NOCLIPTHING
			player.pflags = $ | PF_GODMODE
			player.powers[pw_flashing] = TICRATE
			mo.radius = $*2
			mo.flags = flags
			if srb2p then player.dashmode = 10*TICRATE end
			P_TryMove(mo, mo.x + P_ReturnThrustX(nil, player.drawangle, dist),
			mo.y + P_ReturnThrustY(nil, player.drawangle, dist), true)
			local grounded = P_IsObjectOnGround(mo)
			
			//Go back
			mo.flags = $ | MF_NOCLIPTHING
			mo.radius = $/2
			if srb2p then player.dashmode = 0 end
			P_TryMove(mo, x, y, true)
			mo.z = z
			if grounded and not P_IsObjectOnGround(mo)
			and not (mo.eflags & MFE_SPRUNG)
				mo.momx = momx
				mo.momy = momy
				mo.momz = momz
			end
			if player.panim == PA_DASH and mo.state != S_PLAY_MELEE_FINISH
				player.panim = PA_ABILITY2
			end
			mo.flags = flags
			player.pflags = ($ &~ PF_GODMODE) | godmode
			player.powers[pw_flashing] = flashing
			
			if mo.eflags & MFE_SPRUNG
				mo.momx = $*5/4
				mo.momy = $*5/4
				mo.momz = $*5/4

				S_StartSound(mo, sfx_sprong)
				DoMarioHammerHit(player, true)
				
				if mo.state == S_PLAY_MELEE_FINISH
				or mo.state == S_PLAY_MELEE_LANDING
					if player.pflags & PF_SPINNING
						mo.state = S_PLAY_ROLL
					else
						mo.state = S_PLAY_SPRING
					end
				end
			end
		elseif player.panim == PA_ABILITY2
			player.panim = PA_DASH
		end
		
		//Air state
		if mo.state == S_PLAY_MELEE_LANDING and mo.tics == 1
		and not P_IsObjectOnGround(mo)
			mo.state = DOUBLEJUMPSTATE
		end
		
		if player.mariojumpdir == nil or player.powers[pw_justsprung]
			player.mariojumpdir = player.drawangle
		else
			player.drawangle = player.mariojumpdir
		end
	
	//Crouching
	elseif player.mariocrouch
		local height = P_GetPlayerHeight(player)
		height = $ + FixedMul(player.mariosize, $)
		local canuncrouch = mo.ceilingz - mo.floorz >= height
		
		player.mariocrouch = $ + 1
		player.mariotriple = 0
		
		if player.skidtime
			player.skidtime = 0
			player.marioskid = nil
			S_StopSoundByID(mo, skins[player.skin].soundsid[SKSSKID])
		end
		
		//Cancel
		if (player.marioshoe and not player.mariolongjump) or
		((P_IsObjectOnGround(mo) and player.mariocrouch > 3
		and not player.mariopressspin and not player.mariopressdown)
		or (not P_IsObjectOnGround(mo) and not (player.pflags & PF_JUMPED))
		or player.mariopresscustom == 1 or player.mariopressfirenormal == 1)
		and (canuncrouch or (P_IsObjectOnGround(mo)
		and (player.mariopresscustom == 1 or player.mariopressfirenormal == 1)))
			player.mariocrouch = 0
			player.mariocharge = 0
			player.marioslidesfx = 0
			
			if player.mariolongjump
				player.normalspeed = skins[player.skin].normalspeed
				player.mariolongjump = 0
			end
			
			if mo.state == CROUCHSTATE
			or mo.state == CRAWLSTATE
				mo.state = S_PLAY_WALK
			end
			
			//Crouch custom moves
			if player.mariopresscustom == 1
				
				//Start roll
				if (player.mariopressdirection and P_IsObjectOnGround(mo))
				or not canuncrouch
					player.pflags = $ | PF_SPINNING
					
					S_StartSound(mo, sfx_mrrol2)
					mo.state = ROLLSTATE
					
					player.drawangle = MarioInputAngle(player)
					P_Thrust(mo, player.drawangle, 8*mo.scale)
					if player.speed < 16*mo.scale
						P_InstaThrust(mo, player.drawangle, 16*mo.scale)
					end
					
				//Start spin
				else
					player.mariocustomspin = 1
					DoSpin(player)
				end
				
			//Spin button
			elseif player.mariopressfirenormal == 1
				player.mariocustomspin = -1
				DoSpin(player)
			end
			
		//The actual code
		else
			//Animations
			if not (player.cmd.forwardmove or player.cmd.sidemove)
			or not P_IsObjectOnGround(mo)
			or (mo.eflags & MFE_JUSTHITFLOOR and mo.eflags & MFE_ONGROUND)
				mo.state = CROUCHSTATE
				mo.tics = 2
				
			elseif player.mariocrouch > 3
				if mo.state != CRAWLSTATE
					mo.state = CRAWLSTATE
				end
				if not ((mo.frame & FF_FRAMEMASK) % 2) and mo.tics == 1
					S_StartSoundAtVolume(mo, sfx_mrcwl1 + (((mo.frame & FF_FRAMEMASK)/2) % 2), 128)
				end
			end
			player.panim = PA_DASH
			
			//Long Jump Landing
			if player.mariolongjump
				if player.mariocrouch > 7
					player.normalspeed = skins[player.skin].normalspeed
					player.mariolongjump = 0
				else
					player.drawangle = player.mariojumpdir
				end
				mo.friction = FRACUNIT
			else
				mo.friction = min($*17/16, FRACUNIT)
			end
			
			//Sliding/Crawling
			if P_IsObjectOnGround(mo)
				if (player.rmomx or player.rmomy)
					if not player.mariolongjump
						player.drawangle = R_PointToAngle2(0,0, player.rmomx, player.rmomy)
					end
					player.mariojumpdir = player.drawangle
					
					player.marioslidesfx = max($, 1)
					P_SpawnSkidDust(player, mo.radius/2, false)
					
					player.mariocharge = 0
				else
					player.marioslidesfx = 0
					
					//Crawling
					if player.cmd.forwardmove or player.cmd.sidemove
						local factor = 12
			
						local angle = player.cmd.angleturn<<16
						if mo.flags2 & MF2_TWOD or twodlevel
							angle = ANGLE_90
						end
						
						player.mariojumpdir = player.drawangle
						
						local scale = max(mo.scale + FixedMul(player.mariosize, mo.scale), mo.scale)
						
						P_InstaThrust(mo, angle, player.cmd.forwardmove*scale/factor)
						P_Thrust(mo, angle - ANGLE_90, player.cmd.sidemove*scale/factor)
						
						P_XYMovement(mo)
						
						mo.momx = player.cmomx
						mo.momy = player.cmomy
					else
						//Face one direction
						if player.mariojumpdir == nil
							player.mariojumpdir = player.drawangle
						else
							player.drawangle = player.mariojumpdir
						end
					end
					
					if mo.state == CRAWLSTATE and player.mariocharge < CHARGETIME
						player.mariocharge = 0
					else
						player.mariocharge = $ + 1
						if player.mariocharge >= CHARGETIME
							if player.mariocharge == CHARGETIME
								S_StartSound(mo, sfx_mrchrg)
							end
							if not (player.mariocharge%8)
								local ghost = P_SpawnGhostMobj(mo)
								ghost.color = SKINCOLOR_SUPERPURPLE5
								ghost.colorized = true
								ghost.renderflags = $ | RF_FULLBRIGHT
								ghost.blendmode = AST_ADD
								ghost.momx = mo.momx/2
								ghost.momy = mo.momy/2
								ghost.spritexscale = mo.spritexscale
								ghost.spriteyscale = mo.spriteyscale
								ghost.fuse = 8
								ghost.destscale = $*16
								if ghost.tracer and ghost.tracer.valid
									P_RemoveMobj(ghost.tracer)
								end
							end
						end
					end
				end
			else
				player.mariocrouch = max($, TICRATE/2)
				player.marioslidesfx = 0
				player.mariocharge = 0
				
				//Face one direction
				if player.mariojumpdir == nil
				or player.cmd.forwardmove or player.cmd.sidemove
					player.mariojumpdir = player.drawangle
				else
					player.drawangle = player.mariojumpdir
				end
			end
		end
		
	//Long Jump
	elseif player.mariolongjump
		player.mariotriple = 0
		mo.friction = FRACUNIT
		
		//Cancel
		if mo.eflags & MFE_SPRUNG
			player.mariolongjump = 0
			player.normalspeed = skins[player.skin].normalspeed
		
		//Landing
		elseif P_IsObjectOnGround(mo)
			player.mariocrouch = 1
			player.mariocharge = 0
			player.normalspeed = skins[player.skin].normalspeed
			
			if not (player.cmd.forwardmove or player.cmd.sidemove)
				mo.momx = $/2
				mo.momy = $/2
			end
		else
			//Face one direction
			if player.mariojumpdir == nil
				player.mariojumpdir = player.drawangle
			else
				player.drawangle = player.mariojumpdir
			end
			
			//Make sideways movement less effective
			local diff = MarioInputAngle(player) - player.mariojumpdir
			diff = abs(InvAngle(InvAngle($)))
			
			local factor = AngleFixed(ANGLE_90 - abs(diff - ANGLE_90))
			player.normalspeed = FixedDiv(skins[player.skin].normalspeed, factor + FRACUNIT)
			if player.mariowaterswitch then player.normalspeed = $*2 end
			player.thrustfactor = $ / (factor/48/FRACUNIT + 1)
			
			//Speed cap (funny)
			local diff = R_PointToAngle2(0,0, mo.momx, mo.momy) - player.mariojumpdir
			diff = abs(InvAngle(InvAngle($)))
			
			local maxspeed = LONGJUMPSPEED*mo.scale
			local speed = FixedHypot(mo.momx, mo.momy)
			if mo.eflags & MFE_UNDERWATER and not (mo.eflags & MFE_GOOWATER)
			and not player.mariowaterswitch then maxspeed = $/2 end
			
			if diff >= ANGLE_90
				player.marioblj = speed
			else
				player.marioblj = 0
			end
			
			if player.marioblj and speed > maxspeed/4
				P_InstaThrust(mo, R_PointToAngle2(0,0, mo.momx, mo.momy),
				speed - mo.scale)
			elseif speed >= maxspeed
				P_InstaThrust(mo, R_PointToAngle2(0,0, mo.momx, mo.momy), maxspeed)
			end
			
			player.mariolongjump = $ + 1
		end
		
	//Other junk
	else
		//Face one direction while jumping
		//And do the spin jump!
		if player.mariocapeflight then //nothing
		elseif (player.panim == PA_JUMP or mo.state == S_PLAY_JUMP
		or mo.state == JUMPSTATE or mo.state == DOUBLEJUMPSTATE
		or mo.state == TRIPLEJUMPSTATE or mo.state == WALLJUMPSTATE
		or mo.state == FALLSTATE or mo.state == DOUBLEFALLSTATE
		or mo.state == SPINSTATE or mo.state == CAPESPINSTATE
		or mo.state == S_TAILSOVERLAY_PLUS60DEGREES or mo.state == SCUTTLESTATE
		or player.mariogpjump or player.marioflipmove) and not (mo.eflags & MFE_SPRUNG)
			if player.mariojumpdir == nil
				player.mariojumpdir = player.drawangle
			else
				player.drawangle = player.mariojumpdir
			end
			player.powers[pw_justsprung] = max($,2)
			
			//Prevent the camera from going haywire in Simple
			if (mo.state == SPINSTATE or mo.state == CAPESPINSTATE or player.mariogpjump)
			and not (player.cmd.forwardmove or player.cmd.sidemove)
			and player.pflags&PF_DIRECTIONCHAR and player.pflags&PF_ANALOGMODE
				mo.angle = player.cmd.angleturn<<16
			end
			
			//Spin jump
			if mo.state == SPINSTATE
				if player.mariospinjump
					//Do the spin!
					if player.luigispin
						player.luigispin = $ + 1
						player.mariojumpdir = $ + ANGLE_45*2/3
						if not (player.luigispin % 12)
							S_StartSoundAtVolume(mo, sfx_mrtrfl, 128)
						end
					else
						player.mariojumpdir = $ + ANGLE_45 * min(max(player.mariospinjump, -1), 1)
					end
					
					//No jump damage...
					if player.pflags & PF_JUMPED then player.pflags = $ | PF_NOJUMPDAMAGE end
					
					//Fall slower in the air
					if not P_IsObjectOnGround(mo)
						if mo.flags & MF_NOGRAVITY
							//You get nothing lol
						elseif mo.momz*P_MobjFlip(mo) < 0
							mo.momz = $ - P_GetMobjGravity(mo)*2/3
						else
							mo.momz = $ - P_GetMobjGravity(mo)/4
						end
						
						if not player.powers[pw_super]
							mo.momx = $*23/24
							mo.momy = $*23/24
						end
						
						if player.mariospinsfx > 0
							player.mariospinsfx = -22
						end
						
					//Slow down on the ground
					elseif player.mariospinjump < GROUNDSPINTIME - 1
						if not player.powers[pw_super]
							mo.momx = $*8/9
							mo.momy = $*8/9
						end
						
						P_SpawnSkidDust(player, mo.radius, false)
					end
					
					//Fireball frenzy!!
					if (MARIOFIREPOWERS[player.powers[pw_shield]] or player.powers[pw_super])
					and (player.mariofiredelay == 0 or player.mariofiredelay == MARIO_FIREDELAY - 5)
						if player.mariofiredelay
							player.mariofiredelay = min(-$, MARIO_FIREDELAY/-2)
						else
							player.mariofiredelay = MARIO_FIREDELAY
						end
						
						if player.powers[pw_shield] == SH_GOLDFLOWER
						or player.powers[pw_super]
							mobjinfo[MT_MARIO_FIREBALL].raisestate = mo.skin == "luigi" and 4 or 1
							S_StartSound(mo, sfx_mrglfr)
						elseif player.powers[pw_shield] == SH_ICEFLOWER
							mobjinfo[MT_MARIO_FIREBALL].raisestate = 2
							S_StartSound(mo, sfx_mriceb)
						elseif player.powers[pw_shield] == SH_SUPERBALL
							mobjinfo[MT_MARIO_FIREBALL].raisestate = 3
							S_StartSound(mo, sfx_mrbal5)
						else
							mobjinfo[MT_MARIO_FIREBALL].raisestate = mo.skin == "luigi" and 5 or 0
							S_StartSound(mo, sfx_mario7)
						end
						
						local angle = player.cmd.angleturn<<16 - ANGLE_90
						if mo.flags2 & MF2_TWOD or twodlevel
							angle = $ + ANGLE_90
						end
						if player.mariofiredelay < 0
							angle = $ - ANGLE_180
						end
						P_SPMAngle(mo, MT_MARIO_FIREBALL, angle, 0,0)
					end
					
					//Cancel it on the ground if we stop spinning
					if player.mariospinjump >= GROUNDSPINGRACE
						player.mariospinjump = $ - 1
					elseif player.mariospinjump <= -GROUNDSPINGRACE
						player.mariospinjump = $ + 1
					elseif P_IsObjectOnGround(mo)
					and not player.mariodidcircle
						if player.mariospinjump >= 1
							player.mariospinjump = $ - 1
						elseif player.mariospinjump <= -1
							player.mariospinjump = $ + 1
						end
						
						if not player.mariospinjump
							player.mariospinsfx = 0
							mo.state = S_PLAY_WALK
						end
					else
						player.mariospinjump = min(max($*GROUNDSPINGRACE, -GROUNDSPINGRACE), GROUNDSPINGRACE)
					end
				else
					player.mariospinsfx = 0
					mo.state = S_PLAY_FALL
				end
				
			//Ground Pound Jump
			elseif player.mariogpjump
				if P_IsObjectOnGround(mo)
					player.mariogpjump = 0
				else
					player.mariojumpdir = $ - ANGLE_22h
					player.mariogpjump = $ - 1
					
					P_SpawnSkidDust(player, mo.radius, false)
				end
				
			//Cape spin
			elseif mo.state == CAPESPINSTATE
				player.mariojumpdir = player.cmd.angleturn<<16 - ANGLE_22h - ANG60 * ((leveltime + #player) % 6)
				
				if mo.tics == 1 then player.drawangle = player.cmd.angleturn<<16 end
				
				if P_IsObjectOnGround(mo)
					P_SpawnSkidDust(player, mo.radius, false)
				end
				
			//Flip movie
			elseif player.marioflipmove
				player.powers[pw_noautobrake] = max($, 2)
				
				if P_IsObjectOnGround(mo) or not (player.pflags & PF_JUMPED)
				or (player.panim != PA_FALL and player.panim != PA_SPRING
				and mo.state != BACKFLIPSTATE and mo.state != SIDEFLIPSTATE)
					player.marioflipmove = 0
				end
				
				if mo.state == SIDEFLIPSTATE
					player.mariojumpdir = $ - ANGLE_180/states[SIDEFLIPSTATE].tics
				end
			end
		else
			player.mariojumpdir = nil
			player.mariowalljump = 0
			player.marioflipmove = 0
			
			//Fake skidding
			local pmx = player.rmomx + player.cmomx
			local pmy = player.rmomy + player.cmomy
			
			if not (player.skidtime and mo.state == S_PLAY_SKID) and player.speed
			and (mo.state == S_PLAY_WALK or mo.state == S_PLAY_RUN)
			and not (player.charflags & SF_NOSKID)
			and P_IsObjectOnGround(mo) and not (mo.eflags & MFE_GOOWATER)
			and not (player.pflags & (PF_JUMPED|PF_SPINNING|PF_SLIDING))
			and not player.mariodidcircle
			and FixedHypot(pmx, pmy) >= mo.scale // if you were moving faster than half your run speed last frame
			--and (mo.momx != pmx or mo.momy != pmy) // and you are moving differently this frame
				//Get the control angle
				local mang = R_PointToAngle2(0,0,pmx,pmy) // movement angle
				local diff = mang - MarioInputAngle(player)
				diff = InvAngle(InvAngle($))
				
				// check for skidding
				if abs(diff) > ANGLE_45 or not P_GetPlayerControlDirection(player)
					local pang = R_PointToAngle2(pmx,pmy,mo.momx,mo.momy) // push angle
					local dang = mang - pang // delta angle

					dang = InvAngle(dang) // Make delta angle always positive, invert it if it's negative.

					// If your push angle is more than this close to a full 180 degrees, trigger a skid.
					if abs(dang) > ANGLE_135 or not P_GetPlayerControlDirection(player)
						local speed = FixedHypot(mo.momx, mo.momy)
						if speed < SKIDSPEED*mo.scale
							mo.momx = $*2/3
							mo.momy = $*2/3
						else
							player.skidtime = SKIDTIME
							if speed > SKIDSPEED*mo.scale
							and (P_GetPlayerControlDirection(player)
							or speed >= RUNSPEED*mo.scale*2/3)
								S_StartSound(mo, sfx_skid)
								player.marioskid = true
								if mo.state != S_PLAY_SKID
									mo.state = S_PLAY_SKID
								end
								player.panim = PA_ETC
								mo.tics = player.skidtime
								player.drawangle = mang
							end
						end
					end
				end
			end
		end
		
		//Scuttle
		if mo.state == SCUTTLESTATE
			if player.cmd.buttons & BT_JUMP
				mo.momz = $ - P_GetMobjGravity(mo)/2
				if leveltime & 1
					mo.state = SCUTTLESTATE
				end
			else
				mo.state = FALLSTATE
			end
		end
		
		if player.glidetime or player.climbing or player.pflags & PF_SPINNING
		or (mo.state == SPINSTATE and P_IsObjectOnGround(mo))
			//lol you can't do anything
			
		//Start a scuttle!
		elseif player.cmd.buttons & BT_JUMP and mo.momz*P_MobjFlip(mo) < 0
		and mo.skin == "luigi" and mo.state != SCUTTLESTATE
		and not player.luigiscuttle
		and (player.panim == PA_JUMP or mo.state == S_PLAY_JUMP
		or mo.state == JUMPSTATE or mo.state == DOUBLEJUMPSTATE
		or mo.state == WALLJUMPSTATE or mo.state == FALLSTATE
		or mo.state == DOUBLEFALLSTATE or mo.state == S_TAILSOVERLAY_FLY)
			mo.state = SCUTTLESTATE
			player.luigiscuttle = true
			MarioPlayVoice(player, "scuttle")
			
		//Start a spin!
		elseif player.mariopressfirenormal == 1 and not player.mariospinjump
		and (not P_IsObjectOnGround(mo)
		or (player.mariospinbutton != player.marioattackbutton
		and player.mariospinbutton != player.mariocrouchbutton
		and player.mariospinbutton != player.marioabilitybutton))
			player.mariocustomspin = -1
			DoSpin(player)
			
		//Start a fire move!
		elseif player.mariopressfire == 1
		and (not player.marioshoe or MARIOFIREPOWERS[player.powers[pw_shield]]
		or player.powers[pw_shield] == SH_CAPE)
			ResetMario(player, false, true)
			DoHammer(player, false, true)
			
		//Start a custom 1 move!
		elseif player.mariopresscustom == 1
		and (not player.marioshoe or MARIOFIREPOWERS[player.powers[pw_shield]]
		or player.powers[pw_shield] == SH_CAPE)
			local wascapeflight = player.mariocapeflight and mo.state == CAPEFLYSTATE
			local directional = player.mariogpjump
			ResetMario(player, false, true)
			
			//Start Diving
			if player.mariopressdirection and not player.marioshoe
			and player.mariopressjump != 1 and player.mariodiveenabled != 0
			and mo.state != S_PLAY_FIRE and mo.state != S_PLAY_FIRE_FINISH
			and mo.state != CAPESPINSTATE and player.marioprevstate != CAPESPINSTATE
			and not wascapeflight
				player.mariotriple = 0
				player.mariotripletime = 0
				
				DoDive(player, false, directional)
				
			//Hammer time
			else
				DoHammer(player)
			end
		
		//Start a spin move!
		elseif (player.mariopressspin or player.mariopressdown)
		and ((P_IsObjectOnGround(mo) and not player.marioshoe)
		or player.mariopressspin == 1 or player.mariopressdown == 1)
			
			//Start Ground Pound
			if not P_IsObjectOnGround(mo)
				ResetMario(player)
					
				mo.momx = 0
				mo.momy = 0
				mo.momz = 0
				
				S_StartSound(mo, sfx_mrgnd1)
				MarioPlayVoice(player, "groundpound")
				
				player.pflags = $ | PF_FULLSTASIS
				mo.state = GROUNDPOUNDSTARTSTATE
				
				player.mariogroundpound = 1
				
			//Exit shoe
			elseif player.marioshoe and player.marioshoe.valid
				P_DoJump(player, true)
				player.pflags = $ &~ PF_STARTJUMP
				player.marioshoe.tracer = nil
				player.marioshoe = nil
				
			//Crouching
			else
				ResetMario(player)
				S_StartSound(mo, sfx_mrduck)
				
				player.mariocrouch = 1
				player.mariocharge = 0
			end
		
		//Start spin jump
		elseif not player.mariospinjump and player.mariodidcircle
		and not player.powers[pw_justsprung]
		and (P_IsObjectOnGround(mo) or mo.momz*P_MobjFlip(mo) <= -4*P_GetMobjGravity(mo))
			DoSpin(player)
		end
	end
	
	//Sliding effect
	if player.pflags & PF_SPINNING and P_IsObjectOnGround(mo)
	and mo.state != ROLLSTATE and mo.state != states[ROLLSTATE].nextstate
		player.marioslide = true
		player.marioslidesfx = max($, 1)
		
		if not (player.pflags & PF_STARTDASH)
			P_SpawnSkidDust(player, mo.radius, false)
		end
	elseif player.marioslide and not player.mariowallslide
		player.marioslide = nil
		player.marioslidesfx = 0
	end
	
	//Prevent dives
	if player.glidetime > 0
		player.glidetime = $ - 1
	end
	
	//Make sure we cancel the spin guys...
	if mo.state == SPINSTATE and not player.mariospinjump
		player.mariospinsfx = 0
		mo.state = S_PLAY_FALL
	elseif mo.state != SPINSTATE and player.mariospinjump
		player.mariospinjump = 0
		player.mariospinsfx = 0
	end
	
	//Triple jump grace period
	if P_IsObjectOnGround(mo)
		if player.mariotripletime
			player.mariotripletime = $-1
		else
			player.mariotriple = 0
		end
	end
end)