//
//The Lua for Mario's Powerup object
//

rawset(_G, "MARIO_DESPAWNTIME", TICRATE*16) // How long the powerup lasts for

//Power constants
local SHIELDPOWER = 1
local ONEUPPOWER = -1
local STARPOWER = -2
local MEGAPOWER = -3
local COINPOWER = -4
local POISONPOWER = -5

//Shield constants
rawset(_G, "SH_FAKEBALL", SH_PITY)
rawset(_G, "SH_SUPERBALL", SH_PINK)
rawset(_G, "SH_CAPE", SH_WHIRLWIND)
rawset(_G, "SH_BALLOON", SH_ARMAGEDDON)
rawset(_G, "SH_FAKEMETAL", SH_ELEMENTAL)
rawset(_G, "SH_METAL", SH_ELEMENTAL|SH_PROTECTSPIKE)
rawset(_G, "SH_FAKEMINI", SH_ATTRACT)
rawset(_G, "SH_MINI", SH_PITY|SH_PINK|SH_WHIRLWIND)
rawset(_G, "SH_SHOE", SH_FLAMEAURA)
rawset(_G, "SH_ICEFLOWER", SH_BUBBLEWRAP)
rawset(_G, "SH_FAKEGOLD", SH_THUNDERCOIN)
rawset(_G, "SH_GOLDFLOWER", SH_PINK|SH_METAL)
rawset(_G, "SH_FAKEFLOWER", SH_FORCE|1)

rawset(_G, "MARIOSHIELDS", {
	[SH_NONE] = true,
	[SH_SUPERBALL] = true,
	[SH_CAPE] = true,
	[SH_BALLOON] = true,
	[SH_FIREFLOWER] = true,
	[SH_FAKEGOLD] = true,
	[SH_GOLDFLOWER] = true,
	[SH_ICEFLOWER] = true,
	[SH_FAKEMINI] = true,
	[SH_MINI] = true,
	[SH_FAKEMETAL] = true,
	[SH_METAL] = true,
})

rawset(_G, "MARIOFIREPOWERS", {
	[SH_SUPERBALL] = true,
	[SH_FIREFLOWER] = true,
	[SH_GOLDFLOWER] = true,
	[SH_ICEFLOWER] = true,
})

rawset(_G, "SpawnMarioShieldOrb", function(player)
	//Only real powerups here!
	if player.powers[pw_shield] == SH_FAKEMINI
		player.powers[pw_shield] = SH_MINI
	elseif player.powers[pw_shield] == SH_FAKEMETAL
		player.powers[pw_shield] = SH_METAL
	elseif player.powers[pw_shield] == SH_FAKEGOLD
		player.powers[pw_shield] = SH_GOLDFLOWER
	end
			
	if MARIOSHIELDS[player.powers[pw_shield] &~ SH_FIREFLOWER]
	or player.powers[pw_shield] & SH_FIREFLOWER
		player.shieldscale = 0
		P_SpawnShieldOrb(player)
		return false
	else
		player.shieldscale = skins[player.skin].shieldscale
		P_SpawnShieldOrb(player)
		return true
	end
end)

//Dont spawn them in persona
if srb2p
	MARIOSHIELDS = {}
	MARIOFIREPOWERS = {}
	
	SH_FAKEBALL = INT8_MIN
	SH_SUPERBALL = INT8_MIN
	SH_CAPE = INT8_MIN
	SH_BALLOON = INT8_MIN
	SH_FAKEMETAL = INT8_MIN
	SH_METAL = INT8_MIN
	SH_FAKEMINI = INT8_MIN
	SH_MINI = INT8_MIN
	SH_SHOE = INT8_MIN
	SH_ICEFLOWER = INT8_MIN
	SH_FAKEGOLD = INT8_MIN
	SH_GOLDFLOWER = INT8_MIN
	SH_FAKEFLOWER = INT8_MIN
	
	return
end

//Frames for each
freeslot("SPR_MPWR")
states[freeslot("S_MARIO_ICEFLOWER")] = {SPR_MPWR, I|FF_ANIMATE, -1, nil, 3,  3, S_MARIO_ICEFLOWER}
states[freeslot("S_MARIO_GOLDFLOWER")] = {SPR_MPWR, M|FF_ANIMATE, -1, nil, 3,  3, S_MARIO_GOLDFLOWER}
states[freeslot("S_MARIO_BALLFLOWER")] = {SPR_MPWR, R|FF_ANIMATE, -1, nil, 3,  3, S_MARIO_BALLFLOWER}
local PowerupFrames = {
	[ONEUPPOWER] = 0,
	[STARPOWER] = 1,
	[MEGAPOWER] = 2,
	[POISONPOWER] = 16,
	[SH_FAKEMINI] = 3,
	[SH_MINI] = 3,
	[SH_FAKEMETAL] = 4,
	[SH_METAL] = 4,
	[SH_BALLOON] = 5,
	[SH_CAPE] = 6,
	
	//The states - Note these are negative!
	[SH_FAKEBALL] = -S_MARIO_BALLFLOWER,
	[SH_SUPERBALL] = -S_MARIO_BALLFLOWER,
	[SH_FAKEFLOWER] = -S_FIREFLOWER1,
	[SH_FIREFLOWER] = -S_FIREFLOWER1,
	[SH_ICEFLOWER] = -S_MARIO_ICEFLOWER,
	[SH_FAKEGOLD] = -S_MARIO_GOLDFLOWER,
	[SH_GOLDFLOWER] = -S_MARIO_GOLDFLOWER,
	[SH_SHOE] = -S_MARIO_SHOE,
}

//What actions we can turn into a powerup
local PowerupActions = {
	[A_ExtraLife] = ONEUPPOWER,
	[A_Invincibility] = STARPOWER,
	[A_SuperSneakers] = MEGAPOWER,
	[A_RingBox] = COINPOWER,
	[A_EggmanBox] = POISONPOWER,
	[A_GiveShield] = SHIELDPOWER,
}

//Custom 1up function to play the Mario sound!
local function A_MarioOneUp(mobj)
	local player = mobj.target and mobj.target.valid and mobj.target.player
	if player
		if (netgame or multiplayer) and G_GametypeUsesCoopLives()
			S_StartSound(nil, sfx_marioa)
		else
			S_StartSound(player.mo, sfx_marioa)
		end
		P_GiveCoopLives(player, 1, false)
	end
end

local function A_MarioPowerup(mobj)
	mobj.momx = 0
	mobj.momy = 0
	mobj.momz = 0
	
	if mobj.powerupvar == POISONPOWER
		mobj.color = SKINCOLOR_SUPERPURPLE5
		mobj.colorized = true
		mobj.state = S_MARIO_DUST
		mobj.fuse = 0
		return
	end
	
	mobj.fuse = mobj.tics
	mobj.type = MT_GHOST
	mobj.flags2 = $ | MF2_BOSSNOTRAP
	
	mobj.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
	mobj.blendmode = AST_ADD
end

//The object
states[freeslot("S_MARIO_POWERUP")] = {SPR_UPPB or freeslot("SPR_UPPB"), FF_ANIMATE, 18, A_MarioPowerup, 5,  3, S_NULL}
states[freeslot("S_MARIO_1UP_ICON")] = {SPR_TV1U, 2, 18, A_MarioOneUp,  0, 0, S_NULL}
mobjinfo[freeslot("MT_MARIO_POWERUP")] = {
	spawnstate = S_INVISIBLE,
	deathstate = S_MARIO_POWERUP,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_SPECIAL,
	reactiontime = 1,
}

//Set up the powerup
addHook("MobjSpawn", function(mobj)
	//The object's info was modified via the monitor pop action, which is how
	//we know what powerup we are!
	mobj.powerupseestate = mobj.info.seestate
	local state = states[mobj.powerupseestate]
	mobj.powerupvar = PowerupActions[state.action]
	mobj.powerupsound = mobj.info.attacksound
	
	//Coin spread!
	if mobj.powerupvar == COINPOWER
		for i = 1, 10
			local coin = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_FLINGCOIN)
			if i == 1
				S_StartSound(coin, sfx_mario9)
			elseif i == 6
				S_StartSound(coin, sfx_peww)
			end
			
			//Find out if we're supposed to be flipped or not
			if mobj.info.raisestate
				if mobj.info.raisestate > 2
					coin.flags2 = $ &~ MF2_OBJECTFLIP
					coin.eflags = $ &~ MFE_VERTICALFLIP
				else
					coin.flags2 = $ | MF2_OBJECTFLIP
					coin.eflags = $ | MFE_VERTICALFLIP
				end
			end

			coin.fuse = MARIO_DESPAWNTIME
			coin.mariomonitorcoin = true
			coin.target = mobj.target
			
			//Only on the 2D axis in 2D bud...
			if twodlevel
				coin.flags2 = $ | MF2_TWOD
			end

			local fa = (ANGLE_180/5)*i
			P_InstaThrust(coin, fa, coin.scale)
			P_SetObjectMomZ(coin, 8*FRACUNIT, false)
		end
		
		mobj.tics = 1
		return
	
	//Replace the 1up function
	elseif mobj.powerupvar == ONEUPPOWER
		//Poison in ultimate mode!
		if MarioUltimateMode()
			mobj.powerupseestate = S_EGGMAN_ICON2
			mobj.powerupvar = POISONPOWER
		else
			mobj.powerupseestate = S_MARIO_1UP_ICON
		end
		
	//Get our shield!
	elseif mobj.powerupvar == SHIELDPOWER
		mobj.powerupvar = state.var1
		
		//We want real powerups here!
		if mobj.powerupvar == SH_FAKEFLOWER
			//Ice flower instead of fire flower in snow levels
			if mapheaderinfo[gamemap].weather == PRECIP_SNOW
				mobj.powerupvar = SH_ICEFLOWER
			else
				mobj.powerupvar = SH_FIREFLOWER
			end
		elseif mobj.powerupvar == SH_SHOE //Shoes in Forest Fortress
		or (gamemap == 32 and mapheaderinfo[gamemap].lvlttl == "Forest Fortress")
			local shoe = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_MARIO_SHOE)
			S_StartSound(shoe, sfx_mario9)
			
			shoe.fuse = mobj.info.height/FRACUNIT
			shoe.flags = $ | MF_NOGRAVITY | MF_NOCLIPTHING
			shoe.momz = mobj.scale*P_MobjFlip(mobj)
			
			mobj.tics = 1
			return
		elseif mobj.powerupvar == SH_FAKEMINI
			mobj.powerupvar = SH_MINI
		elseif mobj.powerupvar == SH_FAKEMETAL
			mobj.powerupvar = SH_METAL
		elseif mobj.powerupvar == SH_FAKEGOLD
			mobj.powerupvar = SH_GOLDFLOWER
		elseif mobj.powerupvar == SH_FAKEBALL
			mobj.powerupvar = SH_SUPERBALL
			mobj.powerupsound = sfx_mrball
		end
		
	//More ice flowers in Frozen Hillside
	elseif mobj.powerupvar == POISONPOWER
	and gamemap == 30 and mapheaderinfo[gamemap].lvlttl == "Frozen Hillside"
		mobj.powerupseestate = S_BUBBLEWRAP_ICON2
		mobj.powerupvar = SH_ICEFLOWER
		
	//The heck did we do wrong???
	elseif not mobj.powerupvar
		print("Uh oh, Mario's Monitor Powerup system broke somehow.")
		mobj.tics = 1
		return
	end
	
	S_StartSound(mobj, mobj.powerupvar == MEGAPOWER and sfx_mrmgap or sfx_mario9)
	mobj.shadowscale = FRACUNIT*2/3
	
	local frame = PowerupFrames[mobj.powerupvar]
	if frame != nil and frame < 0
		mobj.state = abs(frame)
	else
		mobj.sprite = SPR_MPWR
		mobj.frame = PowerupFrames[mobj.powerupvar] or 0
	end
	
	mobj.fuse = mobj.info.height/FRACUNIT
	mobj.flags = $ | MF_NOGRAVITY | MF_NOCLIPTHING
	mobj.momz = mobj.scale*P_MobjFlip(mobj)
end, MT_MARIO_POWERUP)

//Powerup Thinker
addHook("MobjThinker", function(mobj)
	if not mobj.health then return end
	
	if not mariopowerup.value and mobj.powerupvar != COINPOWER
	and mobj.powerupvar != ONEUPPOWER and mobj.powerupvar != STARPOWER
	and mobj.powerupvar != POISONPOWER
		local thok = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_THOK)
		thok.state = S_INVISIBLE
		thok.tics = TICRATE
		
		S_StartSound(thok, sfx_mrpoof)
		MarioCircle(mobj, MT_MARIO_DUST, 8, 4*mobj.scale, 0,
		mobj.scale*2/3, 0, nil, true, mobj.angle)
		
		P_KillMobj(mobj)
	end
	
	//Despawn flash
	if mobj.fuse <= TICRATE*2 and not mobj.reactiontime
		mobj.flags2 = $ ^^ MF2_DONTDRAW
	end
	
	if mobj.fuse == mobj.info.height/FRACUNIT and mobj.reactiontime
		mobj.scalespeed = mobj.scale/mobj.fuse
		P_SetScale(mobj, mobj.scale/2)
		if mobj.powerupvar == MEGAPOWER
			mobj.scalespeed = $*2
			mobj.destscale = $*2
		elseif mobj.powerupvar == SH_BALLOON
			mobj.spritexscale = FRACUNIT*3/2
			mobj.spriteyscale = FRACUNIT*3/2
		end
	end
	
	//Find out if we're supposed to be flipped or not
	if mobj.info.raisestate
		if mobj.info.raisestate > 2
			mobj.flags2 = $ &~ MF2_OBJECTFLIP
			mobj.eflags = $ &~ MFE_VERTICALFLIP
		else
			mobj.flags2 = $ | MF2_OBJECTFLIP
			mobj.eflags = $ | MFE_VERTICALFLIP
		end
		mobj.momz = abs($)*P_MobjFlip(mobj)
		mobj.info.raisestate = 0
	end
	
	//Cut out the rising early if we hit a surface
	if mobj.powerupvar != SH_CAPE and mobj.reactiontime and mobj.fuse > 1
	and (not mobj.momz or mobj.z <= mobj.floorz or mobj.z + mobj.height >= mobj.ceilingz)
		mobj.fuse = 1
	end
	
	if mobj.reactiontime then return end
	
	//Movement behavior
	//Normal Movement
	if mobj.powerupvar == ONEUPPOWER
	or mobj.powerupvar == STARPOWER
	or mobj.powerupvar == MEGAPOWER
	or mobj.powerupvar == POISONPOWER
	or mobj.powerupvar == SH_MINI
	or mobj.powerupvar == SH_METAL
	or mobj.powerupvar == SH_BALLOON
		mobj.flags = $|MF_BOUNCE
		
		local newspeed = 4*mobj.scale
		local speed = FixedHypot(mobj.momx, mobj.momy)
		if speed
			mobj.angle = R_PointToAngle2(0,0, mobj.momx, mobj.momy)
		end
		
		//Star bounce
		if mobj.powerupvar == STARPOWER
			mobj.frame = $ | FF_FULLBRIGHT
			
			if P_IsObjectOnGround(mobj)
				mobj.z = $ + P_MobjFlip(mobj)
				P_SetObjectMomZ(mobj, 8*FRACUNIT, false)
				
				S_StartSound(mobj, sfx_mrstar)
			end
			
			A_SmokeTrailer(mobj, MT_BOXSPARKLE, 0)
			if leveltime % 2
				P_SpawnGhostMobj(mobj)
			end
			
		//Balloon float
		elseif mobj.powerupvar == SH_BALLOON
			mobj.momz = FixedMul(sin(FixedAngle(mobj.fuse*TICRATE*FRACUNIT/6) - ANG1*103), mobj.scale*4*P_MobjFlip(mobj))
		
		//Slower Metal Cap
		elseif mobj.powerupvar == SH_METAL
			newspeed = $/4
			
		//Slower Mega Mushroom and Bounce
		elseif mobj.powerupvar == MEGAPOWER
			newspeed = $/2
			
			if P_IsObjectOnGround(mobj)
				mobj.z = $ + P_MobjFlip(mobj)
				P_SetObjectMomZ(mobj, 4*FRACUNIT, false)
			end
			
		//Poison hunt
		elseif mobj.powerupvar == POISONPOWER
		and mobj.target and mobj.target.valid
			local diff = mobj.angle - R_PointToAngle2(mobj.x, mobj.y, mobj.target.x, mobj.target.y)
			diff = InvAngle(InvAngle($))
			mobj.angle = $ - diff/16
		end
		
		P_InstaThrust(mobj, mobj.angle, newspeed)
		
	//Feather float
	elseif mobj.powerupvar == SH_CAPE and mobj.momz * P_MobjFlip(mobj) < 0
		if not (mobj.flags & MF_NOCLIPHEIGHT)
			mobj.flags = $ | MF_NOCLIPHEIGHT | MF_SCENERY
			mobj.frame = $ + 1
		end
		
		if P_IsObjectOnGround(mobj)
			mobj.shadowscale = 0
		elseif not mobj.shadowscale
			mobj.shadowscale = FRACUNIT*2/3
		end
		
		local timer = FixedAngle(mobj.fuse*TICRATE*FRACUNIT/6) + ANG1*20 //Awesome magic number
		local xytimer = sin(timer)
		local ztimer = cos(timer)
		
		mobj.momz = FixedMul(-abs(ztimer), mobj.scale*2*P_MobjFlip(mobj))
		mobj.mirrored = xytimer > 0
		
		local speed = FixedMul(xytimer, mobj.scale*8)
		P_TeleportMove(mobj, mobj.x + P_ReturnThrustX(nil, mobj.angle, speed),
		mobj.y + P_ReturnThrustY(nil, mobj.angle, speed), mobj.z)
	end
end, MT_MARIO_POWERUP)

//Make the powerup jump out
addHook("MobjFuse", function(mobj)
	if not mobj.reactiontime then return end
	
	mobj.flags = $ &~ MF_NOGRAVITY &~ MF_NOCLIPTHING
	P_SetObjectMomZ(mobj, 8*FRACUNIT, false)
	
	if mobj.target and mobj.target.valid
		mobj.angle = R_PointToAngle2(mobj.target.x, mobj.target.y, mobj.x, mobj.y)
	end
	
	if mobj.powerupvar == SH_CAPE
		mobj.angle = $ - ANGLE_90
		mobj.momz = $*3/2
	elseif mobj.powerupvar == POISONPOWER
		mobj.angle = $ - ANGLE_180
	end
	
	mobj.reactiontime = 0
	mobj.fuse = MARIO_DESPAWNTIME
	
	return true
end, MT_MARIO_POWERUP)

//Prevent immediate coin collection
addHook("TouchSpecial", function(mobj, mo)
	if mobj.mariomonitorcoin and mobj.fuse > MARIO_DESPAWNTIME - TICRATE
		if mo and mo.valid and mo.flags2 & MF2_TWOD
			mobj.flags2 = $ | MF2_TWOD
			mobj.momy = 0
		end
		return true
	end
end, MT_FLINGCOIN)

/*function A_AttractChase(mobj, var1, var2)
	if mobj.mariomonitorcoin and mobj.fuse > MARIO_DESPAWNTIME - TICRATE
		mobj.flags2 = $ &~ MF2_DONTDRAW
	else
		super(mobj, var1, var2)
	end
end*/

//Collect the powerup
addHook("TouchSpecial", function(mobj, mo)
	local player = mo.valid and IsMario(mo) and mo.player
	if player and P_CanPickupItem(player, true)
	and mobj and mobj.valid and not mobj.reactiontime
		local oldshield = player.powers[pw_shield]
		local newshield = oldshield
		local oldmega = player.mariomega
		mobj.target = mo
		
		if player.powers[pw_super] and mobj.powerupvar > 0
			player.powers[pw_shield] = SH_CAPE
			oldshield = player.marioshield
			player.marioshield = mobj.powerupvar //Set powerup for after being super
			newshield = player.marioshield
		else
			if mobj.powerupvar == MEGAPOWER then player.mariomega = max($, 1) end
			states[mobj.powerupseestate].action(mobj, mobj.powerupvar) // The item get function
			newshield = player.powers[pw_shield]
		end
		
		//Grow time
		if mobj.powerupvar == MEGAPOWER
			if not oldmega
				player.mariomegacombo = 0
				
				if not player.mariopain
					player.mariopain = {
						momx = mo.momx,
						momy = mo.momy,
						momz = mo.momz,
						state = mo.state,
						frame = mo.frame,
						sprite2 = mo.sprite2,
						tics = mo.tics,
						flags = mo.flags,
						pflags = player.pflags,
						drawangle = player.drawangle,
						flashing = player.powers[pw_flashing],
						shield = player.powers[pw_shield],
						powerup = true,
						mega = true,
						groundpound = player.mariogroundpound,
						spinjump = player.mariospinjump,
						spinsfx = player.mariospinsfx,
						gpjump = player.mariogpjump,
						crouch = player.mariocrouch,
						flipmove = player.marioflipmove,
						capeflight = player.mariocapeflight,
					}
					
					if not player.powers[pw_carry]
						P_ResetPlayer(player)
					end
					
					mo.flags = MF_NOCLIPTHING|MF_NOGRAVITY
					player.pflags = $|PF_FULLSTASIS|PF_INVIS
					mo.momx = 0
					mo.momy = 0
					mo.momz = 0
					if P_IsObjectOnGround(mo) then mo.z = $ + P_MobjFlip(mo) end
					
					player.powers[pw_flashing] = flashingtics
					mo.state = S_PLAY_PAIN
					player.panim = PA_DASH
				end
			end
			
			player.mariomega = sneakertics - TICRATE
			
		//Star voice line
		elseif mobj.powerupvar == STARPOWER
			MarioPlayVoice(player, "star")
		end
		
		//Handle the fire flower
		if player.powers[pw_shield] & SH_FIREFLOWER
		and not player.powers[pw_super]
			if player.powers[pw_shield] & SH_NOSTACK
				if mobj.powerupvar == SH_FIREFLOWER
					player.powers[pw_shield] = $ &~ SH_NOSTACK
					mo.color = SKINCOLOR_WHITE
				else
					player.powers[pw_shield] = $ &~ SH_FIREFLOWER
					mo.color = player.skincolor
				end
			else
				mo.color = SKINCOLOR_WHITE
			end
		end
		
		//Score? Nuh uh!
		if MarioScoreGametype()
		and mobj.powerupvar != ONEUPPOWER and mobj.powerupvar != POISONPOWER
			S_StartSound(mo, mobj.powerupsound or sfx_mario3)
			if mobj.powerupvar > 0 and oldshield == newshield
				S_StartSound(mo, sfx_mrresv)
			end
			return
		end
		
		//Score? Yeparoony!
		local score = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_SCORE)
		if not (score and score.valid)
			return
		elseif mobj.powerupvar == ONEUPPOWER
			score.state = S_SCRK
		elseif mobj.powerupvar == POISONPOWER
			score.state = S_SCRK
			score.frame = M
		else
			score.state = S_SCRD
			P_AddPlayerScore(player, 1000)
			
			S_StartSound(mo, mobj.powerupsound or sfx_mario3)
			if mobj.powerupvar > 0 and oldshield == newshield
				S_StartSound(mo, sfx_mrresv)
			end
		end
		
		if score.scale != mo.scale
			score.momz = FixedMul(FixedDiv($, score.scale), mo.scale)
			score.scale = mo.scale
		end
	else return true end
end, MT_MARIO_POWERUP)

//Spawn the powerup
local Mario_MonitorPop = A_MonitorPop
A_MonitorPop = function(actor, var1, var2)
	local mo = actor.target
	local player = mo.valid and mo.player
	local olditem = actor.info.damage
	local random = false
	
	//Determine if the monitor has a valid powerup
	if player and olditem
		//Random monitor!
		if olditem == MT_UNKNOWN
			olditem = GetMarioRandomItem()
			actor.info.damage = olditem
			random = true
		end
		
		//Find the item...
		local spawnstate = mobjinfo[olditem].spawnstate
		local nextstate = spawnstate and states[spawnstate].nextstate
		local action = nextstate and states[nextstate].action
		local powerup = PowerupActions[action]
		
		//Apply it if it's valid
		if powerup and (IsMario(mo) or (mariocoins.value and powerup == COINPOWER))
		and (powerup != SHIELDPOWER or PowerupFrames[states[nextstate].var1])
		and (mariopowerup.value or powerup == COINPOWER or powerup == STARPOWER
		or powerup == ONEUPPOWER or powerup == POISONPOWER)
			//Set up the mobjinfo for super function
			mobjinfo[MT_MARIO_POWERUP].seestate = nextstate
			mobjinfo[MT_MARIO_POWERUP].attacksound = mobjinfo[olditem].seesound
			actor.info.damage = MT_MARIO_POWERUP
			
			//Match the powerup gravity to the player instead of the monitor
			if actor.target and P_MobjFlip(actor) != P_MobjFlip(actor.target)
				mobjinfo[MT_MARIO_POWERUP].raisestate = P_MobjFlip(actor.target) + 2
			else
				mobjinfo[MT_MARIO_POWERUP].raisestate = 0
			end
			
			//Unique powerup sound for Mega
			if powerup == MEGAPOWER
				mobjinfo[MT_MARIO_POWERUP].attacksound = sfx_mrmega
			end
			
			//Remove the old lives icon
			if olditem == MT_1UP_ICON
			and actor.tracer and actor.tracer.valid
				P_RemoveMobj(actor.tracer)
			end
		end
	end
	
	Mario_MonitorPop(actor, var1, var2)
	actor.info.damage = random and MT_UNKNOWN or olditem
end

//Do it again!! This is ineffecient but whatever
local Mario_GoldMonitorPop = A_GoldMonitorPop
A_GoldMonitorPop = function(actor, var1, var2)
	local mo = actor.target
	local player = mo.valid and mo.player
	local olditem = actor.info.damage
	local random = false
	
	//Determine if the monitor has a valid powerup
	if player and olditem
		//Random monitor!
		if olditem == MT_UNKNOWN
			olditem = GetMarioRandomItem()
			actor.info.damage = olditem
			random = true
		end
		
		//Find the item...
		local spawnstate = mobjinfo[olditem].spawnstate
		local nextstate = spawnstate and states[spawnstate].nextstate
		local action = nextstate and states[nextstate].action
		local powerup = PowerupActions[action]
		
		//Apply it if it's valid
		if powerup and (IsMario(mo) or (mariocoins.value and powerup == COINPOWER))
		and (powerup != SHIELDPOWER or PowerupFrames[states[nextstate].var1])
		and (mariopowerup.value or powerup == COINPOWER or powerup == STARPOWER
		or powerup == ONEUPPOWER or powerup == POISONPOWER)
			//Set up the mobjinfo for super function
			mobjinfo[MT_MARIO_POWERUP].seestate = nextstate
			mobjinfo[MT_MARIO_POWERUP].attacksound = mobjinfo[olditem].seesound
			actor.info.damage = MT_MARIO_POWERUP
			
			//Match the powerup gravity to the player instead of the monitor
			if actor.target and P_MobjFlip(actor) != P_MobjFlip(actor.target)
				mobjinfo[MT_MARIO_POWERUP].raisestate = P_MobjFlip(actor.target) + 2
			else
				mobjinfo[MT_MARIO_POWERUP].raisestate = 0
			end
			
			//Unique powerup sound for Mega
			if powerup == MEGAPOWER
				mobjinfo[MT_MARIO_POWERUP].attacksound = sfx_mrmega
			end
			
			//Remove the old lives icon
			if olditem == MT_1UP_ICON
			and actor.tracer and actor.tracer.valid
				P_RemoveMobj(actor.tracer)
			end
		end
	end
	
	Mario_GoldMonitorPop(actor, var1, var2)
	actor.info.damage = random and MT_UNKNOWN or olditem
end