//
//The Lua for the Kuribo Shoe
//

local CAPESPINSTATE = S_MARIO_CAPESPIN //cape.lua

//Ghost projectile
states[freeslot("S_MARIO_GHOST1")] = {freeslot("SPR_MRGH"), FF_FULLBRIGHT|FF_ANIMATE, 18, nil, 2, 6, freeslot("S_MARIO_GHOST2")}
states[S_MARIO_GHOST2] = {SPR_MRGH, D|FF_FULLBRIGHT|FF_ANIMATE, -1, nil, 1, 3, S_MARIO_GHOST2}
mobjinfo[freeslot("MT_MARIO_GHOST")] = {
	spawnstate = S_MARIO_GHOST1,
	spawnhealth = 1000,
	deathstate = S_XPLD1,
	speed = 32*FRACUNIT,
	radius = 24*FRACUNIT,
	height = 32*FRACUNIT,
	damage = 1,
	flags = MF_MISSILE|MF_GRENADEBOUNCE|MF_SLIDEME|MF_SCENERY,
}

addHook("MobjMoveCollide", function(mobj, mo)
	if mo.flags & MF_SOLID and mo != mobj.target
	and (mo.flags & MF_SHOOTABLE or mo.flags & MF_PUSHABLE)
	and mobj.z - mobj.scale <= mo.z + mo.height
	and mobj.z + mobj.height + mobj.scale >= mo.z
		P_SetObjectMomZ(mo, 8*FRACUNIT, false)
		return false
	end
end, MT_MARIO_GHOST)

addHook("MobjThinker", function(mobj)
	if not P_IsObjectOnGround(mobj) and leveltime & 1
		mobj.momx = FixedMul($, mobj.friction)
		mobj.momy = FixedMul($, mobj.friction)
	end
end, MT_MARIO_GHOST)

addHook("MobjFuse", function(mobj)
	P_KillMobj(mobj)
	A_ForceStop(mobj)
	
	mobj.shadowscale = 0
	mobj.spriteyoffset = -8*FRACUNIT
	return true
end, MT_MARIO_GHOST)

addHook("MobjSpawn", function(mobj)
	mobj.shadowscale = FRACUNIT*4/5
	mobj.fuse = TICRATE
end, MT_MARIO_GHOST)

//The shoe
states[freeslot("S_MARIO_SHOE")] = {freeslot("SPR_MRBT"), FF_ANIMATE, -1, nil, 3, 3, S_MARIO_SHOE}
mobjinfo[freeslot("MT_MARIO_SHOE")] = {
	spawnstate = S_MARIO_SHOE,
	spawnhealth = 1000,
	seesound = sfx_mrbtgt,
	attacksound = sfx_mrbtjm,
	deathsound = sfx_mrbtls,
	deathstate = S_MARIO_SHOE,
	radius = 16*FRACUNIT,
	height = 48*FRACUNIT,
	flags = MF_SPECIAL|MF_SLIDEME,
	reactiontime = 1,
	dispoffset = -1,
}

//Make the shoe jump out
addHook("MobjFuse", function(mobj)
	if not mobj.reactiontime then return end
	
	mobj.flags = $ &~ MF_NOGRAVITY &~ MF_NOCLIPTHING
	P_SetObjectMomZ(mobj, 8*FRACUNIT, false)
	
	mobj.reactiontime = 0
	mobj.fuse = MARIO_DESPAWNTIME
	
	return true
end, MT_MARIO_SHOE)

//Shoe Thinker
addHook("MobjThinker", function(mobj)
	if not mobj.health
		if mobj.threshold != 69
			mobj.threshold = 69
			A_Scream(mobj)
			
			mobj.flags = $ &~ MF_NOGRAVITY | MF_NOCLIP | MF_NOCLIPHEIGHT
			mobj.fuse = TICRATE*2
			
			P_SetObjectMomZ(mobj, 8*FRACUNIT, false)
			P_InstaThrust(mobj, mobj.angle + ANGLE_90, 2*mobj.scale)
		end
		
		mobj.rollangle = $ + ANGLE_11hh
		return
	end
	
	//Despawn flash
	if mobj.fuse and mobj.fuse <= TICRATE*2 and not mobj.reactiontime
		mobj.flags2 = $ ^^ MF2_DONTDRAW
	end
	
	//Somebody's in deh shoe
	local mo = mobj.tracer
	local player = mo and mo.valid and mo.player
	if not player
		if not mobj.fuse
			mobj.tracer = nil
			
			mobj.movedir = 0
			mobj.spritexscale = FRACUNIT
			mobj.spriteyscale = FRACUNIT
			mobj.spritexoffset = 0
			mobj.spriteyoffset = 0
			
			mobj.fuse = MARIO_DESPAWNTIME
			mobj.flags = $ &~ MF_NOCLIP &~ MF_NOCLIPHEIGHT &~ MF_NOCLIPTHING &~ MF_NOGRAVITY
		end
		
		mobj.shadowscale = FRACUNIT
		return
	end
	mobj.shadowscale = 0
	
	//Actually no they aren't
	if not IsMario(mo) or not mo.health
		player.marioshoe = nil
		mobj.tracer = nil
		
		mobj.movedir = 0
		mobj.spritexscale = FRACUNIT
		mobj.spriteyscale = FRACUNIT
		mobj.spritexoffset = 0
		mobj.spriteyoffset = 0
			
		mobj.fuse = MARIO_DESPAWNTIME
		mobj.flags = $ &~ MF_NOCLIP &~ MF_NOCLIPHEIGHT &~ MF_NOCLIPTHING &~ MF_NOGRAVITY
		return
	end
	mobj.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOCLIPTHING | MF_NOGRAVITY
	
	player.marioshoe = mobj
	mobj.fuse = 0
	mobj.reactiontime = 0
	
	mobj.scale = mo.scale
	mobj.rollangle = mo.rollangle
	mobj.mirrored = mo.mirrored
	mobj.spritexscale = mo.spritexscale
	mobj.spriteyscale = mo.spriteyscale
	mobj.spritexoffset = mo.spritexoffset
	mobj.renderflags = mo.renderflags
	mobj.blendmode = mo.blendmode
	mobj.flags2 = mo.flags2 & (MF2_DONTDRAW|MF2_SHADOW|MF2_SPLAT)
	
	A_CapeChase(mobj, 1,0)
	P_TeleportMove(mobj, mobj.x, mobj.y, mobj.z)
	
	//Bouncy bouncy
	local oldmovedir = mobj.movedir
	if P_IsObjectOnGround(mo)
	and (player.cmd.forwardmove or player.cmd.sidemove or mobj.movedir)
	and not (player.pflags & PF_SPINNING) and not player.skidtime
	and not player.mariospinjump and mo.state != CAPESPINSTATE
	and not player.mariogroundpound and not player.mariolongjump
	and not MarioCantMove(player)
		if not mobj.movedir
			S_StartSound(mo, mobj.info.attacksound)
		end
		mobj.movedir = $ + ANGLE_22h*4/5
		if mobj.movedir < 0
			mobj.movedir = 0
			
			//Sick source code splashes
			if ((not (mo.eflags & MFE_VERTICALFLIP)
			and mo.z + P_GetPlayerHeight(player) >= mo.watertop and mo.z <= mo.watertop)
			or (mo.eflags & MFE_VERTICALFLIP
			and mo.z + mo.height - P_GetPlayerHeight(player) <= mo.waterbottom and mo.z + mo.height >= mo.waterbottom))
				local splishtype = (mo.eflags & MFE_TOUCHLAVA) and MT_LAVASPLISH or MT_SPLISH
				local water = P_SpawnMobj(mo.x, mo.y,
					((mo.eflags & MFE_VERTICALFLIP) and mo.waterbottom - FixedMul(mobjinfo[splishtype].height, mo.scale) or mo.watertop), splishtype)
				if (mo.eflags & MFE_GOOWATER)
					S_StartSound(water, sfx_ghit)
				elseif (mo.eflags & MFE_TOUCHLAVA)
					S_StartSound(water, sfx_splash)
				else
					S_StartSound(water, sfx_wslap)
				end
				if (mo.eflags & MFE_VERTICALFLIP)
					water.flags2 = $ | MF2_OBJECTFLIP
					water.eflags = $ | MFE_VERTICALFLIP
				end
				water.scale = mo.scale
			end
		end
	elseif mobj.movedir >= ANGLE_90
		mobj.movedir = max($ + ANGLE_22h*4/5, 0)
	else
		mobj.movedir = max($ - ANGLE_22h*4/5, 0)
	end
	if oldmovedir or mobj.movedir
		mobj.spriteyoffset = 16 * abs(sin(mobj.movedir))
		mo.spriteyoffset = mobj.spriteyoffset
	else
		mobj.spriteyoffset = mo.spriteyoffset
	end
	
	if player.mariopain or player.mariospinjump or mo.state == CAPESPINSTATE
		mo.sprite2 = SPR2_TAL1
	else
		mo.state = S_TAILSOVERLAY_0DEGREES
		mo.tics = -1
		player.panim = PA_DASH
	end
end, MT_MARIO_SHOE)

//Get in shoe
addHook("TouchSpecial", function(mobj, mo)
	local player = mo.valid and IsMario(mo) and mo.player
	if player and P_CanPickupItem(player, true) and mo.momz * P_MobjFlip(mo) <= 0
	and not player.marioshoe and player.powers[pw_shield] != SH_BALLOON
	and mobj and mobj.valid and not mobj.tracer and not mobj.reactiontime
		player.marioshoe = mobj
		mobj.tracer = mo
		
		mobj.fuse = 0
		S_StartSound(mo, mobj.info.seesound)
	end
	return true
end, MT_MARIO_SHOE)